from decimal import Decimal
from Campus.Business import CampusManager
from Candidate.Entity.Candidate import Candidate
from User.Business.UserManager import UserManager
from User.Business.UserPermissionManager import UserPermissionManager

from lawrence.base import EntityManager
from User.Entity.UserPermission import UserPermission
from Candidate.Entity.CandidateNegotiation import CandidateNegotiation

class CandidateManager (EntityManager):
	__entity__ = Candidate

	def __init__(self, **kwargs):
		EntityManager.__init__(self, **kwargs)
		self.UserManager = UserManager(**kwargs)
		self.UserPermissionManager = UserPermissionManager(**kwargs)
		self.CampusManager = CampusManager(**kwargs)

	def GetById(self, candidateId):
		if isinstance(candidateId, str):
			candidateId = int(candidateId)
		if candidateId is None:
			return None
		try:
			candidate = self.DatabaseCursor.Single('''
				SELECT Candidate.Id AS Id,
					   Candidate.RegistrationDateTime,
					   Candidate.BirthDate,
					   Candidate.next_call_time,
					   Candidate.status,
					   Candidate.student_name,
					   SUBSTRING_INDEX(Candidate.CityName, '||', 1) AS DistrictName,
					   SUBSTRING_INDEX(Candidate.CityName, '||', -1) AS CityName,
					   Candidate.school,
					   Candidate.contact_name,
					   Candidate.contact_email,
					   Candidate.ExamDateTime,
					   CASE WHEN CandidateAppointment.AppointmentDateTime IS NULL
							  THEN 'unscheduled'
							WHEN DATE(CandidateAppointment.AppointmentDateTime) >= date(NOW()) AND CandidateAppointment.UserId IS NULL
							  THEN 'scheduled'
							WHEN CandidateAppointment.UserId IS NULL
							  THEN 'missed'
							WHEN DATE(CandidateAppointment.AppointmentDateTime) >= DATE_SUB(DATE(NOW()), INTERVAL 7 DAY)
							  THEN 'met_recently'
							ELSE 'met' END AS AppointmentStatus,
					   Course.Id AS `Course.Id`,
					   Course.Name AS `Course.Name`,
					   Course.IsExamEligible AS `Course.IsExamEligible`,
					   Course.IsENEMEligible AS `Course.IsENEMEligible`,
					   Campus.Id AS `Campus.Id`,
					   ExamCampus.Id AS `ExamCampus.Id`
				FROM Candidate
					 LEFT OUTER JOIN Course
					   ON Candidate.CourseId = Course.Id
					 LEFT OUTER JOIN Campus
					   ON Candidate.CampusId = Campus.Id
					 LEFT OUTER JOIN Campus AS ExamCampus
					   ON Candidate.ExamCampusId = ExamCampus.Id
					 LEFT OUTER JOIN CandidateAppointment
					   ON CandidateAppointment.CandidateId = Candidate.Id AND
						  COALESCE(CandidateAppointment.AppointmentDateTime >
								   Candidate.ExamDateTime, 1) AND
						  CandidateAppointment.AppointmentDateTime =
							(SELECT MAX(aux.AppointmentDateTime)
							 FROM CandidateAppointment AS aux
							 WHERE aux.CandidateId = CandidateAppointment.CandidateId)
				WHERE Candidate.Id = %s''',
				(candidateId, ))
		except ValueError:
			return None
		candidate = Candidate(**candidate)

		if candidate.Campus.Id is not None:
			candidate.Campus = self.CampusManager.GetById(candidate.Campus.Id)
		else:
			candidate.Campus = None
		if candidate.ExamCampus.Id is not None:
			candidate.ExamCampus = self.CampusManager.GetById(candidate.ExamCampus.Id)
		else:
			candidate.ExamCampus = None

		candidate.ContactPhones = [row.Phone for row in self.DatabaseCursor.Execute('''
			SELECT phone AS Phone
			FROM CandidatePhone
			WHERE CandidateId = %s''',
			(candidate.Id, ))]

		try:
			enemData = self.DatabaseCursor.Single('''
				SELECT name, code,
					   ch_score, cn_score, lc_score, mt_score, rd_score
				FROM CandidateENEM
				WHERE CandidateId = %s''',
				(candidate.Id, ))
			candidate.enem = enemData
			candidate.enem.average_score = Decimal('0.2') * (
				candidate.enem.ch_score + candidate.enem.cn_score +
				candidate.enem.lc_score + candidate.enem.mt_score +
				candidate.enem.rd_score)
		except ValueError:
			pass

		try:
			candidate.rating = self.DatabaseCursor.Single('''
				SELECT CandidateRating.rating AS rating
				FROM CandidateRating
				WHERE CandidateRating.CandidateId = %s
				ORDER BY CandidateRating.time DESC''',
				(candidate.Id, )).rating
		except ValueError:
			candidate.rating = None

		candidate.Negotiation = self.GetNegotiations(candidate)

		return candidate

	def GetNegotiations(self, candidate):
		assert isinstance(candidate, Candidate)
		return [CandidateNegotiation(**row) for row in self.DatabaseCursor.Execute('''
			SELECT SchoolClass.Id AS `SchoolClass.Id`,
				   SchoolClass.Name AS `SchoolClass.Name`,
				   ClassShift.Id AS `Shift.Id`,
				   ClassShift.name AS `Shift.Name`,
				   CandidateNegotiation.Id,
				   CandidateNegotiation.LowerBound,
				   CandidateNegotiation.UpperBound,
				   CandidateNegotiation.FullNumber,
				   CandidateNegotiation.NegotiationDate,
				   CandidateNegotiation.ExpirationDate
			FROM CandidateNegotiation
				 INNER JOIN SchoolClass
				   ON CandidateNegotiation.SchoolClassId = SchoolClass.Id
				 INNER JOIN ClassShift
				   ON SchoolClass.ShiftId = ClassShift.Id
			WHERE CandidateId = %s''',
			(candidate.Id, ))]

	def Search(self, candidateFilter, nextCallNow = False):
		return Candidate.Search(user = self.CurrentUser, candidateFilter = candidateFilter, nextCallNow = nextCallNow)

	def TimeSeries(self, candidateFilter):
		return Candidate.TimeSeries(user = self.CurrentUser, candidateFilter = candidateFilter)

	def ValidStatusTransitions(self, candidate):
		statuses = Candidate.STATUSES.copy()

		if candidate.status != 'pending_reenrollment':
			del statuses['pending_reenrollment']
		if candidate.grades:
			del statuses['unconfirmed']
			del statuses['confirmed']
			del statuses['absent']
		else:
			if candidate.status != 'present':
				del statuses['present']
			if candidate.status != 'absent':
				del statuses['absent']
			if candidate.status == 'confirmed':
				del statuses['unconfirmed']
			if candidate.status != 'failed':
				del statuses['failed']
			if candidate.status == 'absent':
				del statuses['confirmed']
				del statuses['unconfirmed']
			if candidate.status == 'pending_reenrollment':
				del statuses['confirmed']
				del statuses['unconfirmed']
				del statuses['cancelled']

		# Some transitions may be disallowed because of a lack of permission.
		if 'failed' in statuses and 'failed' != candidate.status and not self.UserPermissionManager.HasPermission(
				self.UserManager.GetById(self.CurrentUser.Id),
				UserPermission(Group = 'Candidate', Permission = 'ImportGrades')):
			del statuses['failed']
		if 'enrolled' in statuses and 'enrolled' != candidate.status and not self.UserPermissionManager.HasPermission(
				self.UserManager.GetById(self.CurrentUser.Id),
				UserPermission(Group = 'Candidate', Permission = 'Enroll')):
			del statuses['enrolled']
		if 'cancelled' in statuses and 'cancelled' != candidate.status and not self.UserPermissionManager.HasPermission(
				self.UserManager.GetById(self.CurrentUser.Id),
				UserPermission(Group = 'Candidate', Permission = 'Cancel')):
			del statuses['cancelled']

		assert candidate.status in statuses
		return statuses

