import datetime

import markupsafe
from Campus.Business import CampusManager

from Candidate.Business.CandidateManager import CandidateManager
from Candidate.Entity.Candidate import Candidate
from Candidate.Entity.CandidateFilter import CandidateFilter
from Candidate.Controller.CandidateQSE import validate, getToken
from Ozark.Typing import List
from User.Business.UserPermissionManager import UserPermissionManager
from lawrence.base import Handler
from lawrence.HTTPStatus import *
import lawrence.parse as parse
import lawrence.util as util
import lawrence.config as config
import lawrence.db as db

DBRow = db.DBRow
from User.Entity.UserPermission import UserPermission
import lawrence.mail

class Index (Handler):
	__method__ = 'GET'
	__path__ = '/[Cc]andidate/'

	def Handle(self, *,
			Course: List[int],
			Segment: List[int],
			Campus: List[int],
			Status: List[str],
			AppointmentStatus: List[str],
			ExamDate: List[datetime.date],
			ExamTime: List[datetime.time],
			Tag: List[int],
			Query: str = None,
			SortOrder: str = None,
			Page: int = 1,
			_Format: str = None):
		if not self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser,
				UserPermission(Group = 'Candidate', Permission = 'View')):
			raise Forbidden()

		candidateList = Candidate.Search(
			user = self.CurrentUser,
			candidateFilter = CandidateFilter(
				Campus = Campus or None,
				Course = Course or None,
				Segment = Segment or None,
				Status = Status or None,
				AppointmentStatus = AppointmentStatus or None,
				ExamDate = ExamDate or None,
				ExamTime = ExamTime or None,
				Tag = Tag or None,
				Query = Query,
				SortOrder = SortOrder,
				Page = Page,
			),
		)

		if _Format == 'AutoComplete':
			import json
			self.ContentType = 'application/json'
			return json.dumps([row.student_name for row in candidateList])

		self.Model.CanEMail = self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser,
				UserPermission(Group = 'Candidate', Permission = 'EMail'))
		self.Model.CanCreateQueue = self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser,
				UserPermission(Group = 'Candidate', Permission = 'CreateQueue'))
		self.Model.CanExport = self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser,
				UserPermission(Group = 'Candidate', Permission = 'Export'))
		self.Model.CanImportTags = self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser,
				UserPermission(Group = 'Candidate', Permission = 'ImportTags'))

		self.Model.CandidateList = candidateList
		
		return self.RenderTemplate('Candidate/View/Candidate/Index')

class Listing (Handler):
	__method__ = 'GET'
	__path__ = '/Candidate/_Listing/'

	def Handle(self, *,
			Course: List[int],
			Segment: List[int],
			Campus: List[int],
			Status: List[str],
			AppointmentStatus: List[str],
			ExamDate: List[datetime.date],
			ExamTime: List[datetime.time],
			Tag: List[int],
			Query: str = None,
			SortOrder: str = None,
			Page: int = 1):
		if not self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser,
				UserPermission(Group = 'Candidate', Permission = 'View')):
			raise Forbidden()

		self.Model.CandidateList = Candidate.Search(
			user = self.CurrentUser,
			candidateFilter = CandidateFilter(
				Campus = Campus or None,
				Course = Course or None,
				Segment = Segment or None,
				Status = Status or None,
				AppointmentStatus = AppointmentStatus or None,
				ExamDate = ExamDate or None,
				ExamTime = ExamTime or None,
				Tag = Tag or None,
				Query = Query,
				SortOrder = SortOrder,
				Page = Page))

		return self.RenderTemplate('Candidate/View/Candidate/_Listing')

class FilterCounts (Handler):
	__method__ = 'GET'
	__path__ = '/Candidate/_FilterCounts/'

	def Handle(self, *,
			Course: List[int],
			Segment: List[int],
			Campus: List[int],
			Status: List[str],
			AppointmentStatus: List[str],
			ExamDate: List[datetime.date],
			ExamTime: List[datetime.time],
			Tag: List[int]):
		if not self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser,
				UserPermission(Group = 'Candidate', Permission = 'View')):
			raise Forbidden()

		import json
		self.ContentType = 'application/json'
		filterCounts = Candidate.FilterCounts(
			user = self.CurrentUser,
			candidateFilter = CandidateFilter(
				Campus = Campus or None,
				Course = Course or None,
				Segment = Segment or None,
				Status = Status or None,
				AppointmentStatus = AppointmentStatus or None,
				ExamDate = ExamDate or None,
				ExamTime = ExamTime or None,
				Tag = Tag or None))
		filterCounts['ExamDate'] = {
			(k.isoformat() if k is not None else k): v
			for k, v in list(filterCounts['ExamDate'].items())}
		filterCounts['ExamTime'] = {
			((datetime.datetime.min + k).time().isoformat() if k is not None else k): v
			for k, v in list(filterCounts['ExamTime'].items())}
		return json.dumps(filterCounts)

class Report (Handler):
	__method__ = 'GET'
	__path__ = '/Candidate/_Report/'

	def GetFieldValues(self, fieldName):
		if fieldName == 'Course':
			return [(row.Id, row.Name) for row in self.DatabaseCursor.Execute('''
						SELECT
							id AS Id,
							name AS Name
						FROM
							Course''')]
		elif fieldName == 'Campus':
			return [(row.Id, row.Name) for row in self.DatabaseCursor.Execute('''
						SELECT
							id AS Id,
							name AS Name
						FROM
							Campus
						ORDER BY Name''')
					if row.Id in self.GetManager(CampusManager).GetAuthorized()]
		elif fieldName == 'Status':
			return list(Candidate.STATUSES.items())
		elif fieldName == 'ExamDate':
			from lawrence.util import LongDay, LongMonth
			return [(row.ExamDate, LongDay(row.ExamDate) + " de " + LongMonth(row.ExamDate)) for row in db.Execute('''
						SELECT DISTINCT
							DATE(ExamDateTime) AS ExamDate
						FROM
							Candidate
						ORDER BY
							ExamDate''') if row.ExamDate is not None]
		elif fieldName == 'ExamTime':
			from lawrence.util import ShortTime
			return [(row.ExamTime, ShortTime(row.ExamTime)) for row in db.Execute('''
						SELECT DISTINCT
							TIME(ExamDateTime) AS ExamTime
						FROM
							Candidate
						ORDER BY ExamTime''') if row.ExamTime is not None]
		elif fieldName == 'Tag':
			return [(row.Id, row.Name) for row in db.Execute('''
			SELECT CandidateTag.Id,
				   CandidateTag.Name
			FROM CandidateTag
			ORDER BY Name''')]
		else:
			raise NotImplementedError()

	def Handle(self, *,
			Course: int = None,
			Segment: int = None,
			Campus: int = None,
			Status: str = None,
			AppointmentStatus: str = None,
			ExamDate: datetime.date = None,
			ExamTime: datetime.time = None,
			Tag: int = None,
			RowField: str = None,
			ColumnField: str = None):
		if not self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser,
				UserPermission(Group = 'Candidate', Permission = 'View')):
			raise Forbidden()

		self.Model.RowField = RowField
		self.Model.ColumnField = ColumnField
		# TODO: check fields for validity
		# TODO: forbid RowField == ColumnField

		self.Model.RowList = self.GetFieldValues(self.Model.RowField) if self.Model.RowField is not None else None
		self.Model.ColumnList = self.GetFieldValues(self.Model.ColumnField) if self.Model.ColumnField is not None else None

		candidateFilter = CandidateFilter(
			Page = 1,
			Campus = Campus,
			Course = Course,
			Segment = Segment,
			Status = Status,
			AppointmentStatus = AppointmentStatus,
			ExamDate = ExamDate,
			ExamTime = ExamTime,
			Tag = Tag)

		if self.Model.RowField is not None and getattr(candidateFilter, self.Model.RowField, None) is not None:
			raise BadRequest()

		if self.Model.ColumnField is not None and getattr(candidateFilter, self.Model.ColumnField, None) is not None:
			raise BadRequest()

		if self.Model.RowField and self.Model.ColumnField:
			self.Model.ReportCounts = {}
			import copy
			for rowValue, rowLabel in self.Model.RowList:
				partialFilter = copy.copy(candidateFilter)
				setattr(partialFilter, self.Model.RowField, rowValue)

				self.Model.ReportCounts[rowValue] = Candidate.FilterCounts(self.CurrentUser,
																		   candidateFilter = partialFilter)[self.Model.ColumnField]
		else:
			self.Model.ReportCounts = None

		return self.RenderTemplate('Candidate/View/Candidate/_Report')

class Detail (Handler):
	__method__ = 'GET'
	__path__ = '/[Cc]andidate/(?P<candidateId>[1-9][0-9]*)/'

	def Handle(self, *,
			candidate: Candidate,
			enem_error: str = None):

		if not self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser,
				UserPermission(Group = 'Candidate', Permission = 'View')):
			raise Forbidden()

		if candidate.status == 'enrolled' and not self.HasPermission(candidate.path, 'admin'):
			self.CurrentUser.Grant(path = candidate.path, permission = 'read')

		self.Model.CanEdit = self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser,
				UserPermission(Group = 'Candidate', Permission = 'Edit'))
		self.Model.CanDeleteNegotiation = self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser,
				UserPermission(Group = 'Candidate', Permission = 'DeleteNegotiation'))
		self.Model.CanEnroll = self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser,
				UserPermission(Group = 'Candidate', Permission = 'Enroll'))
		self.Model.CanCreateAppointment = self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser,
				UserPermission(Group = 'Candidate', Permission = 'CreateAppointment'))
		self.Model.CanDeleteAppointment = self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser,
				UserPermission(Group = 'Candidate', Permission = 'DeleteAppointment'))
		self.Model.CanPostComment = self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser,
				UserPermission(Group = 'Candidate', Permission = 'PostComment'))
		self.Model.CanPostRating = self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser,
				UserPermission(Group = 'Candidate', Permission = 'PostRating'))
		self.Model.CanImportGrades = self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser,
				UserPermission(Group = 'Candidate', Permission = 'ImportGrades'))
		self.Model.CanServiceAppointment = self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser,
				UserPermission(Group = 'Candidate', Permission = 'ServiceAppointment'))
		self.Model.Candidate = candidate

		call_attempts = db.Execute('''SELECT
				Total,
				Successful
			FROM
				CandidateCallAttempts
			WHERE
				CandidateId = %s''',
			(candidate.Id ))
		if len(call_attempts) > 0:
			total_call_attempts = call_attempts[0].Total
			successful_call_attempts = call_attempts[0].Successful
		else:
			total_call_attempts = 0
			successful_call_attempts = 0

		alt_grades = self.DatabaseCursor.Execute('''
				SELECT ExamName, Grade
				FROM CandidateAlternativeGrade
				WHERE CandidateId = %s''', 
				(candidate.Id, ))
		if len(alt_grades) > 0:
			alt_grade_name = alt_grades[0].ExamName
			alt_grade = alt_grades[0].Grade
		else:
			alt_grade_name = ""
			alt_grade = "0"

		qse = self.DatabaseCursor.Execute('''
			SELECT *
			FROM CandidateQSE
			WHERE CandidateId = %s
			''',
			(candidate.Id))
		qse = qse[0] if len(qse) > 0 else None

		familyIncome = self.DatabaseCursor.Execute('''
			SELECT * 
			FROM CandidateFamilyIncome
			WHERE CandidateId = %s
			''', (candidate.Id))

		qse_status = validate(qse, familyIncome)
		qse_link = 'http://'+config.domain+'/candidate/'+str(candidate.Id)+'/qse/1/?token=' +getToken(candidate.Id)

		return self.RenderTemplate('Candidate/View/Candidate/Detail',
							 candidate = candidate,
							 ExamDateTimes = Candidate.ExamDateTimes(),
							 total_call_attempts = total_call_attempts,
							 successful_call_attempts = successful_call_attempts,
							 alt_grade_name = alt_grade_name,
							 alt_grade = alt_grade,
							 qse = self.Model.CanServiceAppointment,
							 qse_status = qse_status,
							 qse_link = qse_link)

class DetailEdit (Handler):
	__method__ = 'POST'
	__path__ = '/candidate/(?P<candidateId>[1-9][0-9]*)/'

	def Handle(self, *,
			candidate: Candidate,
			name: str,
			value: List[str]):

		if not self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser,
				UserPermission(Group = 'Candidate', Permission = 'Edit')):
			raise Forbidden()

		self.Model.Candidate = candidate

		self.RequirePermission(candidate.Campus.Path, 'read')
		if candidate.status == 'enrolled':
			self.RequirePermission(candidate.path, 'admin')

		if name not in (
				'school', 'contact_name', 'contact_email', 'contact_phone', 'status', 'course', 'campus', 'exam_campus',
				'ExamDateTime', 'student_name', 'BirthDate', 'CityName'):
			raise BadRequest()
		if name in ('contact_phone', ):
			try:
				value = [parse.phone(contactPhone) for contactPhone in value]
			except ValueError:
				raise BadRequest()
			self.DatabaseCursor.Execute('''
				DELETE FROM CandidatePhone
				WHERE CandidateId = %s''',
				(candidate.Id, ))
			for contactPhone in value:
				self.DatabaseCursor.Execute('''
					INSERT INTO CandidatePhone (CandidateId, phone)
					VALUES (%s, %s)''',
					(candidate.Id,
					 contactPhone))
		else:
			value = value[0]
			if name == 'campus':
				name = 'CampusId'
			if name == 'exam_campus':
				name = 'ExamCampusId'
			if name == 'course':
				name = 'CourseId'
			try:
				if name == 'ExamDateTime':
					value = parse.datetime(value)
				if name == 'BirthDate':
					value = parse.date(value)
			except ValueError:
				raise BadRequest()
			if name == 'status' and value not in self.GetManager(CandidateManager).ValidStatusTransitions(self.Model.Candidate):
				raise BadRequest()
			if name == 'status' and value == 'failed' and candidate.grades:
				self.DatabaseCursor.Execute('''
					DELETE FROM CandidateGrade
					WHERE CandidateId = %s''',
					(candidate.Id, ))
			if name == 'status' and value == 'cancelled':
				self.DatabaseCursor.Execute('''
					DELETE FROM CandidateAppointment
					WHERE
						DATE(AppointmentDateTime) >= DATE(NOW()) AND
						CandidateId = %s''',
					(candidate.Id, ))

			self.DatabaseCursor.Execute('''
				UPDATE Candidate
				SET %s = %%s
				WHERE id = %%s''' % (name, ),
				(value,
				 candidate.Id))

		auditTableName = 'audit_candidate_' + name
		if hasattr(candidate, name) and \
		   value != getattr(candidate, name) and \
		   self.DatabaseCursor.Execute('''
			SELECT table_name
			FROM information_schema.tables
			WHERE table_name = %s''',
			(auditTableName, )):
			self.DatabaseCursor.Execute('''
				INSERT INTO %s (CandidateId, OldValue, NewValue, UserId)
				VALUES (%%s, %%s, %%s, %%s)''' % (auditTableName, ),
				(candidate.Id,
				 getattr(candidate, name),
				 value,
				 self.CurrentUser.id))

		if name == 'ExamDateTime' and candidate.status == 'absent':
			self.DatabaseCursor.Execute('''
				UPDATE Candidate
				SET status = 'unconfirmed'
				WHERE Id = %s''',
				(candidate.Id, ))
		if name == 'status' and value == 'cancelled':
			lawrence.mail.Queue(config.candidate_email,
								candidate.contact_email,
								self.RenderTemplate('Candidate/View/Candidate/Registration/CancellationEmail',
											  candidate = candidate))
		self.DatabaseCursor.Commit()

		raise NoContent()

class ValidCampusView (Handler):
	__method__ = 'GET'
	__path__ = '/candidate/(?P<candidateId>[1-9][0-9]*)/valid-Campus/'

	def Handle(self, *,
			candidate: Candidate):
		if not self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser,
				UserPermission(Group = 'Candidate', Permission = 'View')):
			raise Forbidden()

		self.RequirePermission(candidate.Campus.Path, 'read')

		campusList = [{'value': row.id, 'text': row.name} for row in db.Execute('''
			SELECT
				Campus.id AS id,
				Campus.name AS name
			FROM
				Campus
				INNER JOIN (SELECT DISTINCT CourseId, CampusId FROM SchoolClass) AS T ON T.CampusId = Campus.id
			WHERE
				T.CourseId = %s
			ORDER BY
				Campus.name''',
			(candidate.Course.Id, ))]

		import json
		self.ContentType = 'application/json'
		return json.dumps(campusList)

class ValidStatusesView (Handler):
	__method__ = 'GET'
	__path__ = '/candidate/(?P<candidateId>[1-9][0-9]*)/valid-statuses/'

	def Handle(self, *,
			candidate: Candidate):
		if not self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser,
				UserPermission(Group = 'Candidate', Permission = 'View')):
			raise Forbidden()

		self.RequirePermission(candidate.Campus.Path, 'read')

		import json
		self.ContentType = 'application/json'
		return json.dumps([{'value': k, 'text': v}
						   for k, v in list(self.GetManager(CandidateManager).ValidStatusTransitions(candidate).items())])

class ValidExamDatetimeView (Handler):
	__method__ = 'GET'
	__path__ = '/candidate/(?P<candidateId>[1-9][0-9]*)/valid-exam-datetimes/'

	def Handle(self, *,
			candidate: Candidate):
		if not self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser,
				UserPermission(Group = 'Candidate', Permission = 'View')):
			raise Forbidden()

		self.RequirePermission(candidate.Campus.Path, 'read')

		import json
		self.ContentType = 'application/json'
		return json.dumps([{'value': str(ExamDateTime), 'text': util.LongDateTime(ExamDateTime)}
						   for ExamDateTime in candidate.valid_ExamDateTimes()])

class CommentNew (Handler):
	__method__ = 'POST'
	__path__ = ('/candidate/(?P<candidateId>[1-9][0-9]*)'
				'/comment/=/new/')

	def Handle(self, *,
			candidate: Candidate,
			value: str):
		if not self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser,
				UserPermission(Group = 'Candidate', Permission = 'PostComment')):
			raise Forbidden()

		self.RequirePermission(candidate.Campus.Path, 'read')

		value = str(markupsafe.escape(value)).replace("\n", "<br>")
		db.Execute(
			'''INSERT INTO CandidateComment (CandidateId, UserId, comment)
			   VALUES (%s, %s, %s)''',
			(candidate.Id,
			 self.CurrentUser.Id,
			 value))
		return self.RenderTemplate('Candidate/View/Candidate/_Fields:candidate_activity',
							 candidate,
							 DBRow(type = 'comment',
								   datetime = datetime.datetime.now(),
								   user = self.CurrentUser,
								   comment = value))
class CommentDelete (Handler):
	__method__ = 'POST'
	__path__ = '/candidate/(?P<candidateId>[1-9][0-9]*)/comment/=/delete/'

	def Handle(self, *,
			candidate: Candidate,
			comment_content: str):

		self.DatabaseCursor.begin()
		try:
			comment = self.DatabaseCursor.Single('''
				SELECT UserId FROM CandidateComment
				WHERE CandidateId = %s AND
					  comment = %s''',
				(candidate.Id,
				 comment_content))
		except ValueError:
			raise NotFound()
		if comment.UserId is not None:
			self.RequirePermission(candidate.path, 'admin')

		self.DatabaseCursor.Execute('''
			DELETE FROM CandidateComment
			WHERE CandidateId = %s AND
				  comment = %s''',
			(candidate.Id,
			 comment_content))
		self.DatabaseCursor.Commit()

		raise NoContent()

class CommentDelete (Handler):
	__method__ = 'POST'
	__path__ = '/candidate/(?P<candidateId>[1-9][0-9]*)/calls/'

	def Handle(self, *,
			candidate: Candidate,
			total: int,
			successful: int):
		
		if not total or total < 0:
			total = 0
		if not successful or successful < 0:
			successful = 0

		db.Execute(
			'''INSERT INTO CandidateCallAttempts (CandidateId, Total, Successful)
			   VALUES (%s, %s, %s)
			   ON DUPLICATE KEY UPDATE Total=%s, Successful=%s''',
			(candidate.Id,
			 total,
			 successful, 
			 total,
			 successful))

		raise SeeOther('/candidate/' + str(candidate.Id) + '/')


