from io import BytesIO
import datetime
from zipfile import ZipFile
from Campus.Business import CampusManager

from Candidate.Entity.CandidateFilter import CandidateFilter
from Ozark.Typing import List
from User.Business.UserPermissionManager import UserPermissionManager
from lawrence import util as util
from Candidate.Entity.Candidate import roll, Candidate, certificate, xlsx
from User.Entity.UserPermission import UserPermission
from lawrence.HTTPStatus import Forbidden
from lawrence.base import Handler

class ExportXLSXRoll (Handler):
	__method__ = 'GET'
	__path__ = '/candidate/=/export/xlsx/roll/'

	def Handle(self):
		if not self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser,
				UserPermission(Group = 'Candidate', Permission = 'Export')):
			raise Forbidden()

		zipContent = BytesIO()
		zipName = 'listas-%s.zip' % datetime.date.today().isoformat()
		with ZipFile(zipContent, 'w') as zipFile:
			for campus in self.GetManager(CampusManager).GetAuthorized():
				campus = self.DatabaseCursor.Single('''
					SELECT Id, Name
					FROM Campus
					WHERE Id = %s''',
					(campus, ))
				candidate = self.DatabaseCursor.Execute('''
					SELECT
						CandidateId AS Id,
						TIME(ExamDateTime) AS ExamTime
					FROM
						Candidate
					WHERE
						ExamCampusId = %s AND
						status != 'cancelled' AND
						status != 'pending_reenrollment' AND
						ExamDateTime IS NOT NULL AND
						DATE(ExamDateTime) = DATE(TIMESTAMPADD(HOUR, -6, NOW()))
					ORDER BY
						student_name''',
					(campus.id, ))
				for examTime, group in util.group_by(candidate, 'ExamTime'):
					candidateIdList = [row.Id for row in group]
					rollFile = BytesIO()
					workbook = roll(candidateIdList)
					workbook.save(rollFile)
					zipFile.writestr('%s/%s.xlsx' % (util.Slug(campus.name), util.time(examTime)), rollFile.getvalue())
		self.Headers.append(('Content-Disposition', 'attachment; filename=%s' % zipName))
		self.Headers.append(('Content-Length', str(len(zipContent.getvalue()))))
		self.ContentType = 'application/zip'
		return [zipContent.getvalue()]


class ExportDOCXCertificate (Handler):
	__method__ = 'GET'
	__path__ = '/candidate/=/export/docx/certificate/'

	def Handle(self, *,
			Campus: int = None,
			Status: str = None,
			Course: int = None,
			Segment: int = None,
			ExamDate: datetime.date = None,
			ExamTime: datetime.time = None,
			Tag: int = None,
			Query: str = None,
			SortOrder: str = None):
		if not self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser,
				UserPermission(Group = 'Candidate', Permission = 'View')):
			raise Forbidden()

		result = Candidate.Search(self.CurrentUser,
								  candidateFilter = CandidateFilter(
									  Campus = Campus,
									  Status = Status,
									  Course = Course,
									  Segment = Segment,
									  ExamDate = ExamDate,
									  ExamTime = ExamTime,
									  Tag = Tag,
									  Query = Query,
									  SortOrder = SortOrder))
		if len(result) > 1 and not self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser, UserPermission(Group = 'Candidate', Permission = 'Export')):
			raise Forbidden()
		candidateIdList = [row.Id for row in result]

		certificateData = certificate(candidateIdList)
		self.ContentType = 'application/vnd.openxmlformats-officedocument.wordprocessingml.document'
		self.Headers.append(('Content-Disposition', 'attachment; filename=%s.docx' % util.Slug("certificado")))
		self.Headers.append(('Content-Length', str(len(certificateData))))
		return [certificateData]


class ExportDOCXCertificateFailed (Handler):
	__method__ = 'GET'
	__path__ = '/candidate/=/export/docx/certificate-failed/'

	def Handle(self,
			Campus: int = None,
			Status: str = None,
			Course: int = None,
			Segment: int = None,
			ExamDate: datetime.date = None,
			ExamTime: datetime.time = None,
			Tag: int = None,
			Query: str = None,
			SortOrder: str = None):
		if not self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser,
				UserPermission(Group = 'Candidate', Permission = 'View')):
			raise Forbidden()

		result = Candidate.Search(self.CurrentUser,
								  candidateFilter = CandidateFilter(
									  Campus = Campus,
									  Status = Status,
									  Course = Course,
									  Segment = Segment,
									  ExamDate = ExamDate,
									  ExamTime = ExamTime,
									  Tag = Tag,
									  Query = Query,
									  SortOrder = SortOrder))
		if len(result) > 1 and not self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser, UserPermission(Group = 'Candidate', Permission = 'View')):
			raise Forbidden()

		candidateIdList = [row.Id for row in result]

		certificateData = certificate(candidateIdList, 'certificate-failed.docx')
		self.ContentType = 'application/vnd.openxmlformats-officedocument.wordprocessingml.document'
		self.Headers.append(('Content-Disposition', 'attachment; filename=%s.docx' % util.Slug("certificado")))
		self.Headers.append(('Content-Length', str(len(certificateData))))
		return [certificateData]


class ExportXLSXRaw (Handler):
	__method__ = 'GET'
	__path__ = '/candidate/=/export/xlsx/raw/'

	def Handle(self, *,
			Campus: List[int],
			Status: List[str],
			AppointmentStatus: List[str],
			Course: List[int],
			Segment: List[int],
			ExamDate: List[datetime.date],
			ExamTime: List[datetime.time],
			Tag: List[int],
			Query: str = None,
			SortOrder: str = None):
		if not self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser,
				UserPermission(Group = 'Candidate', Permission = 'Export')):
			raise Forbidden()

		result = Candidate.Search(
			user = self.CurrentUser,
			candidateFilter = CandidateFilter(
				Campus = Campus or None,
				Course = Course or None,
				Segment = Segment or None,
				Status = Status or None,
				AppointmentStatus = AppointmentStatus or None,
				ExamDate = ExamDate or None,
				ExamTime = ExamTime or None,
				Tag = Tag or None,
				Query = Query,
				SortOrder = SortOrder,
				Page = None))

		candidateIdList = [row.Id for row in result]

		workbook = xlsx(candidateIdList)
		outputFile = BytesIO()
		workbook.save(outputFile)
		self.Headers.append(('Content-Disposition', 'attachment; filename=Alunos.xlsx'))
		self.ContentType = 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'
		return outputFile.getvalue()