import datetime
from Candidate.Business.CandidateQueueManager import CandidateQueueManager

from Candidate.Entity.Candidate import Candidate
from Candidate.Entity.CandidateQueue import CandidateQueue
from User.Business.UserPermissionManager import UserPermissionManager
from User.Entity.UserPermission import UserPermission
from lawrence.base import Handler
from lawrence.HTTPStatus import Forbidden, SeeOther, NoContent

class CandidateQueueNew (Handler):
	__method__ = 'POST'
	__path__ = '/Candidate/Queue/New/'

	def Handle(self, *,
			Name: str,
			Course: int = None,
			Segment: int = None,
			Campus: int = None,
			ExamDate: datetime.date = None,
			ExamTime: datetime.time = None,
			Tag: int = None,
			Status: str = None,
			AppointmentStatus: str = None):
		if not self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser,
				UserPermission(Group = 'Candidate', Permission = 'CreateQueue')):
			raise Forbidden()

		self.DatabaseCursor.Execute('''
			INSERT INTO CandidateQueue (Name)
			VALUES (%s)''',
			(Name, ))
		candidateQueueId = self.DatabaseCursor.ScopeIdentity()
		self.DatabaseCursor.Execute('''
			INSERT INTO CandidateQueueFilter (CandidateQueueId, CourseId, SegmentId, CampusId, ExamDate, ExamTime, Tag, Status, AppointmentStatus)
			VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s)''',
			(candidateQueueId,
			 Course,
			 Segment,
			 Campus,
			 ExamDate,
			 ExamTime,
			 Tag,
			 Status,
			 AppointmentStatus))

		self.DatabaseCursor.Commit()
		raise SeeOther('/Candidate/Queue/%d/' % candidateQueueId)

class CandidateQueueDetail (Handler):
	__method__ = 'GET'
	__path__ = '/Candidate/Queue/(?P<candidateQueueId>[0-9]+)/'

	def Handle(self, *,
			candidateQueue: CandidateQueue):
		if not self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser,
				UserPermission(Group = 'Candidate', Permission = 'ViewQueue')):
			raise Forbidden()

		self.Model.CanDeleteQueue = self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser,
				UserPermission(Group = 'Candidate', Permission = 'DeleteQueue'))

		self.Model.CandidateList = Candidate.Search(user = self.CurrentUser, candidateFilter = candidateQueue.Filter)
		self.Model.CandidateQueue = candidateQueue
		return self.RenderTemplate('/Candidate/View/Candidate/Queue/Detail')


class CandidateQueueDelete (Handler):
	__method__ = 'POST'
	__path__ = '/Candidate/Queue/(?P<candidateQueueId>[0-9]+)/Delete/'

	def Handle(self, *,
			candidateQueue: CandidateQueue):
		if not self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser,
				UserPermission(Group = 'Candidate', Permission = 'DeleteQueue')):
			raise Forbidden()

		self.GetManager(CandidateQueueManager).Delete(candidateQueue)

		raise NoContent()


class CandidateQueueFront (Handler):
	__method__ = 'POST'
	__path__ = '/Candidate/Queue/(?P<candidateQueueId>[0-9]+)/Front/'

	def Handle(self, *,
			candidateQueue: CandidateQueue):
		if not self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser,
				UserPermission(Group = 'Candidate', Permission = 'ViewQueue')):
			raise Forbidden()

		try:
			candidate = self.GetManager(CandidateQueueManager).GetFront(candidateQueue)
		except ValueError:
			return self.RenderTemplate('Candidate/Queue/Empty')

		self.Model.CanEdit = self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser,
				UserPermission(Group = 'Candidate', Permission = 'Edit'))
		self.Model.CanCancel = self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser,
				UserPermission(Group = 'Candidate', Permission = 'Cancel'))
		self.Model.CanDeleteAppointment = self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser,
				UserPermission(Group = 'Candidate', Permission = 'DeleteAppointment'))
		self.Model.CanPostComment = self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser,
				UserPermission(Group = 'Candidate', Permission = 'PostComment'))
		self.Model.CanServiceAppointment = self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser,
				UserPermission(Group = 'Candidate', Permission = 'ServiceAppointment'))
		self.Model.CanScheduleAppointment = self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser,
				UserPermission(Group = 'Candidate', Permission = 'CreateAppointment'))
		self.Model.CanDeleteNegotiation = self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser,
				UserPermission(Group = 'Candidate', Permission = 'DeleteNegotiation'))
		self.Model.CanEnroll = False
		return self.RenderTemplate('/Candidate/View/Candidate/Queue/Front',
			candidate = candidate,
			candidateQueue = candidateQueue)

class PostponeCall (Handler):
	__method__ = 'POST'
	__path__ = '/candidate/(?P<candidateId>[1-9][0-9]*)/postpone/'

	def Handle(self, *,
			candidate: Candidate,
			hours: int):
		if not self.GetManager(UserPermissionManager).HasPermission(
				self.CurrentUser,
				UserPermission(Group = 'Candidate', Permission = 'ViewQueue')):
			raise Forbidden()

		self.RequirePermission(candidate.Campus.Path, 'read')

		self.DatabaseCursor.Execute('''
			UPDATE Candidate
			SET next_call_time = TIMESTAMPADD(HOUR, %s, NOW())
			WHERE id = %s''',
			(hours,
			 candidate.Id))
		self.DatabaseCursor.Commit()

		raise NoContent()