from io import BytesIO
from decimal import Decimal
from collections import OrderedDict, defaultdict
import datetime
import re
import os.path
from Campus.Business import CampusManager
from SchoolClass.Entity import SchoolClass

from lawrence.base import Handler
import lawrence.util as util
import lawrence.db as db

DBRow = db.DBRow
from lawrence.Entity import Entity

from openpyxl import Workbook
from openpyxl.reader.excel import load_workbook

class CandidateActivity (object):
	def __init__(self, **kwargs):
		for key, value in list(kwargs.items()):
			setattr(self, key, value)

class CandidateRegistrationActivity (CandidateActivity):
	def __init__(self, **kwargs):
		CandidateActivity.__init__(self, **kwargs)
		self.type = 'registration'

class SqlQuery (object):
	def __init__(self):
		self.Columns = {}
		self.Tables = []
		self.Constraints = []
		self.GroupBy = []
		self.OrderBy = []
		self.Limit = None

	def __str__(self):
		queryString = ''
		queryString += 'SELECT ' + ', '.join('%s AS %s' % (v, k) for k, v in list(self.Columns.items()))
		queryString += ' FROM ' + ' '.join(self.Tables)
		if self.Constraints:
			queryString += ' WHERE ' + ' AND '.join(self.Constraints)
		if self.GroupBy:
			queryString += ' GROUP BY ' + ', '.join(self.GroupBy)
		if self.OrderBy:
			queryString += ' ORDER BY ' + ', '.join(self.OrderBy)
		if self.Limit:
			queryString += ' LIMIT ' + str(self.Limit.start) + ', ' + str(self.Limit.stop - self.Limit.start)
		return queryString


class Candidate (Entity):
	STATUSES = OrderedDict([
		('unconfirmed', "não confirmado"),
		('confirmed', "confirmado"),
		('failed', "reprovado"),
		('absent', "ausente"),
		('present', "presente"),
		('dropped', "desistente"),
		('enrolled', "matriculado"),
		('cancelled', "cancelado"),
		('pending_reenrollment', "rematrícula pendente"),
	])

	APPOINTMENT_STATUSES = OrderedDict([
		('unscheduled', "nenhum agendamento"),
		('scheduled', "agendado para atendimento"),
		('missed', "faltou ao último atendimento"),
		('met_recently', "atendido recentemente"),
		('met', "atendido há mais de uma semana"),
	])

	PAGE_SIZE = 20

	def __init__(self, **kwargs):
		Entity.__init__(self, **kwargs)
		if hasattr(self, 'Id'):
			self.path = '/candidate/%s/' % self.Id
			self.Path = self.path
			self.EncodedId = util.EncodeId(self.Id)

	@property
	def grades(self):
		return {grade.SubjectName: grade.Grade for grade in db.Execute('''
			SELECT
				subject_name AS SubjectName,
				grade AS Grade
			FROM CandidateGrade
			WHERE CandidateId = %s''',
			(self.Id, ))}

	@property
	def activities(self):
		activities = []
		activities.append(CandidateRegistrationActivity(datetime = self.RegistrationDateTime))
		if self.next_call_time > datetime.datetime.now() + datetime.timedelta(minutes = 15):
			activities.append(CandidateActivity(type = 'call', datetime = self.next_call_time))
		if self.ExamDateTime is not None:
			activities.append(CandidateActivity(type = 'exam', datetime = self.ExamDateTime))
		for comment in self.comments:
			activities.append(CandidateActivity(type = 'comment', datetime = comment.time,
									user = comment.user,
									comment = comment.comment))
		for appointment in self.appointments:
			activities.append(CandidateActivity(type = 'appointment', datetime = appointment.datetime,
									student_name = self.student_name,
									user = appointment.user))
		for rating in self.ratings:
			activities.append(CandidateActivity(type = 'rating', datetime = rating.time,
									user = rating.user,
									rating = rating.rating))
		for row in db.Execute('''
			SELECT OldValue,
				   NewValue,
				   User.Id AS UserId,
				   User.FullName AS user_name,
				   timestamp
			FROM audit_candidate_status
				 LEFT OUTER JOIN User ON User.Id = audit_candidate_status.UserId
			WHERE CandidateId = %s''',
			(self.Id, )):
			activities.append(CandidateActivity(type = 'change', datetime = row.timestamp,
									field = "Situação",
									user = DBRow(id = row.UserId,
												 name = row.user_name),
									old_value = self.__class__.STATUSES[row.OldValue],
									new_value = self.__class__.STATUSES[row.NewValue]))
		for row in db.Execute('''
			SELECT OldValue,
				   NewValue,
				   User.Id AS UserId,
				   User.FullName AS user_name,
				   timestamp
			FROM audit_candidate_ExamDateTime
				 INNER JOIN User ON User.Id = audit_candidate_ExamDateTime.UserId
			WHERE CandidateId = %s''',
			(self.Id, )):
			activities.append(CandidateActivity(type = 'change', datetime = row.timestamp,
									field = "Data/hora do exame",
									user = DBRow(id = row.UserId,
												 name = row.user_name),
									old_value = util.LongDateTime(row.OldValue) if row.OldValue else "(nenhum)",
									new_value = util.LongDateTime(row.NewValue) if row.NewValue else "(nenhum)"))
		activities.sort(key = lambda activity: activity.datetime)
		return activities

	@property
	def comments(self):
		comments = db.Execute('''
			SELECT User.Id AS UserId,
				   User.FullName AS user_name,
				   CandidateComment.time AS time,
				   NOW() AS now,
				   CandidateComment.comment AS comment
			FROM CandidateComment
				 LEFT OUTER JOIN User ON CandidateComment.UserId = User.Id
			WHERE CandidateComment.CandidateId = %s''',
			(self.Id, ))
		for comment in comments:
			comment.user = DBRow(id = comment.UserId, name = comment.user_name)
		return comments

	@property
	def ratings(self):
		ratings = db.Execute('''
			SELECT User.Id AS UserId,
				   User.FullName AS user_name,
				   CandidateRating.time AS time,
				   NOW() AS now,
				   CandidateRating.rating AS rating
			FROM CandidateRating
				 INNER JOIN User ON CandidateRating.UserId = User.Id
			WHERE CandidateRating.CandidateId = %s''',
			(self.Id, ))
		for rating in ratings:
			rating.user = DBRow(id = rating.UserId, name = rating.user_name)
		return ratings

	@property
	def appointments(self):
		appointments = db.Execute('''
			SELECT Candidate.student_name AS student_name,
				   CandidateAppointment.Id,
				   CandidateAppointment.AppointmentDateTime AS datetime,
				   User.Id AS `user.id`,
				   User.FullName AS `user.name`
			FROM CandidateAppointment
				 INNER JOIN Candidate
				   ON CandidateAppointment.CandidateId = Candidate.Id
				 LEFT OUTER JOIN User
				   ON CandidateAppointment.UserId = User.Id
			WHERE CandidateAppointment.CandidateId = %s''',
			(self.Id, ))
		return [Candidate.Appointment(**row) for row in appointments]

	@property
	def discounts(self):
		try:
			totalGrade = sum(self.grades.values())
			schoolClassList = [SchoolClass(**schoolClass) for schoolClass in db.Execute('''
				SELECT DISTINCT
					 SchoolClass.Id AS id,
					 SchoolClass.Id AS Id,
					 SchoolClass.Name AS class_name,
					 SchoolClassOutlay.MonthlyPayment,
					 SchoolClassOutlay.MinimumDiscount,
					 SchoolClassOutlay.AverageDiscount,
					 SchoolClassOutlay.MaximumDiscount,
					 SchoolClass.CourseId AS `course.id`,
					 ClassShift.Id AS `shift.id`,
					 ClassShift.name AS `shift.name`
				FROM SchoolClass
					 INNER JOIN SchoolClassOutlay
					   ON SchoolClassOutlay.SchoolClassId = SchoolClass.Id
					 INNER JOIN ClassShift
					   ON ClassShift.Id = SchoolClass.ShiftId
					 LEFT OUTER JOIN CandidateNegotiation
					   ON CandidateNegotiation.CandidateId = %s AND
						  CandidateNegotiation.SchoolClassId = SchoolClass.Id
				WHERE CampusId = %s AND
					  (SchoolClass.CourseId = %s OR
					   CandidateNegotiation.SchoolClassId IS NOT NULL)''',
				(self.Id,
				 self.Campus.Id,
				 self.Course.Id))]

			for schoolClass in schoolClassList:
				schoolClass.name = schoolClass.class_name
				schoolClass.merit_discount = schoolClass.discount(totalGrade)
				schoolClass.total_payment = ((Decimal('1.000') - schoolClass.merit_discount) * schoolClass.MonthlyPayment) \
						.quantize(Decimal('0.01'))
				try:
					latestNegotiation = db.Single('''
						SELECT MAX(NegotiationDate) AS LatestNegotiationDate
						FROM CandidateNegotiation
						WHERE CandidateId = %s AND
							  SchoolClassId = %s''',
						(self.Id, schoolClass.Id))
					negotiation = db.Single('''
						SELECT LowerBound, UpperBound, FullNumber
						FROM CandidateNegotiation
						WHERE CandidateId = %s AND
							  SchoolClassId = %s AND
							  NegotiationDate = %s''',
						(self.Id, schoolClass.Id, latestNegotiation.LatestNegotiationDate))
					schoolClass.LowerBound = negotiation.LowerBound
					schoolClass.UpperBound = negotiation.UpperBound
					schoolClass.FullNumber = negotiation.FullNumber
				except ValueError:
					schoolClass.LowerBound = schoolClass.merit_discount
					schoolClass.UpperBound = None
					schoolClass.FullNumber = None
			schoolClassList.sort(key = lambda schoolClass: schoolClass.merit_discount)
			return schoolClassList
		except AttributeError as ex:
			raise KeyError() from ex

	@classmethod
	def ExamDateTimes(cls):
		return [row.ExamDateTime for row in db.Execute('''
			SELECT DISTINCT ExamDateTime FROM Candidate ORDER BY ExamDateTime''')]

	FILTER_COLUMNS = {
		'Campus': 'Candidate.CampusId',
		'Course': 'Candidate.CourseId',
		'Segment': 'Course.SegmentId',
		'ExamDate': 'DATE(Candidate.ExamDateTime)',
		'ExamTime': 'TIME(Candidate.ExamDateTime)',
		'Tag': 'CandidateTagMap.TagId',
		'Status': 'Candidate.status',
		'AppointmentStatus': '''
			CASE WHEN CandidateAppointment.AppointmentDateTime IS NULL
				   THEN 'unscheduled'
				 WHEN DATE(CandidateAppointment.AppointmentDateTime) >= date(NOW()) AND CandidateAppointment.UserId IS NULL
				   THEN 'scheduled'
				 WHEN CandidateAppointment.UserId IS NULL
				   THEN 'missed'
				 WHEN DATE(CandidateAppointment.AppointmentDateTime) >= DATE_SUB(DATE(NOW()), INTERVAL 7 DAY)
				   THEN 'met_recently'
				 ELSE 'met' END''',
	}

	@classmethod
	def TimeSeries(cls, user, candidateFilter):
		timeSeries = defaultdict(lambda: 0)
		for candidate in cls.Search(user, candidateFilter):
			timeSeries[candidate.RegistrationDateTime.date()] += 1
		# Remove partial data point
		timeSeries[datetime.date.today()] = 0
		del timeSeries[datetime.date.today()]

		return timeSeries

	@classmethod
	def FilterCounts(cls, user, candidateFilter):
		import copy
		filterCounts = {}
		for filterName, filterColumn in list(cls.FILTER_COLUMNS.items()):
			if filterName in ('AppointmentStatus', ) and candidateFilter.Status != ['present']:
				continue
			if filterName in ('Segment', ):
				continue
			partialFilter = copy.deepcopy(candidateFilter)
			setattr(partialFilter, filterName, None)

			countQuery = cls.ApplyFilter(user, partialFilter)
			countQuery.Columns.update({
				'value': filterColumn,
				'count': 'COUNT(*)',
			})
			countQuery.GroupBy.append(filterColumn)
			if filterName == 'Tag':
				countQuery.Tables.append('LEFT OUTER JOIN CandidateTagMap ON CandidateTagMap.CandidateId = Candidate.Id')
			filterCounts[filterName] = {row.value: row.count for row in db.Execute(str(countQuery))}
			if filterName == 'Status':
				setattr(partialFilter, filterName, 'cancelled')
				countQuery = cls.ApplyFilter(user, partialFilter)
				countQuery.Columns.update({
					'value': filterColumn,
					'count': 'COUNT(*)',
				})
				filterCounts[filterName].update({row.value: row.count for row in db.Execute(str(countQuery))})

				setattr(partialFilter, filterName, 'pending_reenrollment')
				countQuery = cls.ApplyFilter(user, partialFilter)
				countQuery.Columns.update({
					'value': filterColumn,
					'count': 'COUNT(*)',
				})
				filterCounts[filterName].update({row.value: row.count for row in db.Execute(str(countQuery))})

		countQuery = cls.ApplyFilter(user, candidateFilter)
		countQuery.Columns.update({ 'count': 'COUNT(*)' })
		filterCounts['CandidateCount'] = db.Single(str(countQuery)).count
		filterCounts['PageCount'] = (filterCounts['CandidateCount'] + cls.PAGE_SIZE - 1) // cls.PAGE_SIZE
		return filterCounts

	@classmethod
	def ApplyFilter(cls, user, candidateFilter):
		campusManager = CampusManager(currentUser = user, databaseCursor = db.Cursor())

		filterQuery = SqlQuery()

		filterQuery.Tables.extend([
			'Candidate',
		])

		if candidateFilter.Status == ['present']:
			filterQuery.Tables.extend([
				'''LEFT OUTER JOIN CandidateAppointment
					  ON CandidateAppointment.CandidateId = Candidate.id AND
						 COALESCE(CandidateAppointment.AppointmentDateTime > Candidate.ExamDateTime, 1) AND
						 CandidateAppointment.AppointmentDateTime =
						   (SELECT MAX(aux.AppointmentDateTime)
							FROM CandidateAppointment AS aux
							WHERE aux.CandidateId = CandidateAppointment.CandidateId)''',
				'''LEFT OUTER JOIN CandidateAppointment AS last_interviewers
					  ON last_interviewers.CandidateId = Candidate.id AND
						 last_interviewers.UserId IS NOT NULL AND
						 COALESCE(last_interviewers.AppointmentDateTime > Candidate.ExamDateTime, 1) AND
						 last_interviewers.AppointmentDateTime =
						   (SELECT MAX(aux.AppointmentDateTime)
							FROM CandidateAppointment AS aux
							WHERE aux.CandidateId = last_interviewers.CandidateId)''',
			])

		if candidateFilter.Segment is not None:
			filterQuery.Tables.extend([
				'''INNER JOIN Course ON Candidate.CourseId = Course.Id''',
				'''INNER JOIN Course ON Course.CourseId = Course.Id''',
			])

		if candidateFilter.Tag is not None:
			filterQuery.Tables.append('INNER JOIN CandidateTagMap ON CandidateTagMap.CandidateId = Candidate.Id')

		queryFilters = {
			'Campus': campusManager.GetAuthorized(),
			'Course': None,
			'Segment': None,
			'ExamDate': None,
			'ExamTime': None,
			'Tag': None,
			'Status': [status for status in list(cls.STATUSES.keys()) if status not in ('cancelled', 'pending_reenrollment')],
			'AppointmentStatus': None,
		}

		if len(queryFilters['Campus']) == len(campusManager.GetAll()):
			queryFilters['Campus'] = None

		if candidateFilter.Campus is not None:
			queryFilters['Campus'] = candidateFilter.Campus

		if candidateFilter.Course is not None:
			queryFilters['Course'] = candidateFilter.Course

		if candidateFilter.Segment is not None:
			queryFilters['Segment'] = candidateFilter.Segment

		if candidateFilter.ExamDate is not None:
			queryFilters['ExamDate'] = candidateFilter.ExamDate

		if candidateFilter.ExamTime is not None:
			queryFilters['ExamTime'] = candidateFilter.ExamTime

		if candidateFilter.Tag is not None:
			queryFilters['Tag'] = candidateFilter.Tag

		if candidateFilter.Status is not None:
			queryFilters['Status'] = list(map(str, candidateFilter.Status))

		# FIXME: validate these against Status
		if candidateFilter.Status == ['present'] and candidateFilter.AppointmentStatus is not None:
			queryFilters['AppointmentStatus'] = list(map(str, candidateFilter.AppointmentStatus))

		for filterName, filterColumn in list(cls.FILTER_COLUMNS.items()):
			if filterName in ['Query', 'SortOrder', 'Page']:
				continue
			if queryFilters[filterName] is not None:
				filterQuery.Constraints.append(db.query_in(filterColumn, queryFilters[filterName]))

		return filterQuery

	@classmethod
	def Search(cls, user, candidateFilter, nextCallNow = False):
		filterQuery = cls.ApplyFilter(user, candidateFilter)

		if candidateFilter.SortOrder == 'last_comment':
			filterQuery.Tables.append('''
				LEFT OUTER JOIN (SELECT CandidateId, MAX(time) AS comment_time
								 FROM CandidateComment
								 GROUP BY CandidateId) AS last_comments
				ON last_comments.CandidateId = Candidate.id''')
			filterQuery.Columns['last_comment'] = 'last_comments.comment_time'
			filterQuery.OrderBy.append(candidateFilter.SortOrder)
		elif candidateFilter.SortOrder == 'rating':
			filterQuery.Tables.append('''
				LEFT OUTER JOIN (SELECT CandidateRating.*
								 FROM CandidateRating
									  INNER JOIN (SELECT CandidateId, MAX(time) AS max_time
												  FROM CandidateRating
												  GROUP BY CandidateId) AS latest_ratings
									  ON latest_ratings.CandidateId = CandidateRating.CandidateId AND
										 CandidateRating.time = latest_ratings.max_time) AS ratings
				ON ratings.CandidateId = Candidate.id''')
			filterQuery.Columns['rating'] = 'ratings.rating'
			filterQuery.OrderBy.append(candidateFilter.SortOrder + ' DESC')
		elif candidateFilter.SortOrder == 'total_grade':
			filterQuery.Tables.append('''
				LEFT OUTER JOIN (SELECT CandidateId, SUM(grade) AS total_grade
								 FROM CandidateGrade
								 GROUP BY CandidateId) AS total_grades
				ON total_grades.CandidateId = Candidate.id''')
			filterQuery.Columns['total_grade'] = 'total_grades.total_grade'
			filterQuery.OrderBy.append(candidateFilter.SortOrder + ' DESC')
		elif candidateFilter.SortOrder == 'enem':
			filterQuery.Tables.append('''
				LEFT OUTER JOIN CandidateENEM ON CandidateENEM.CandidateId = Candidate.id''')
			filterQuery.Columns['enem'] = '(ch_score + cn_score + lc_score + mt_score + rd_score) / 5'
			filterQuery.OrderBy.append(candidateFilter.SortOrder + ' DESC')
		elif candidateFilter.SortOrder == 'negotiation_gap':
			filterQuery.Tables.append('''
				LEFT OUTER JOIN (SELECT CandidateId,
										MIN(COALESCE(UpperBound, 1) - LowerBound) AS gap
								 FROM CandidateNegotiation
								 GROUP BY CandidateId) AS negotiation_gaps
				ON negotiation_gaps.CandidateId = Candidate.id''')
			filterQuery.Columns['negotiation_gap'] = 'negotiation_gaps.gap'
			filterQuery.OrderBy.append(candidateFilter.SortOrder + ' IS NULL')
			filterQuery.OrderBy.append(candidateFilter.SortOrder)

		if candidateFilter.Query:
			if re.match(r'[1-9][0-9]{7}', candidateFilter.Query):
				decodedId = util.DecodeId(int(candidateFilter.Query))
				filterQuery.Constraints.append('Candidate.id = %d' % decodedId)
			else:
				escapedQuery = candidateFilter.Query.replace("%", "\\%").replace("_", "\\_")
				filterQuery.Constraints.append('Candidate.student_name LIKE ' + db.UnicodeEscape("%" + escapedQuery + "%"))
				filterQuery.OrderBy.insert(0, 'Candidate.student_name LIKE ' + db.UnicodeEscape(escapedQuery + "%") + ' DESC')

		if nextCallNow:
			filterQuery.Constraints.append('next_call_time < NOW()')

		filterQuery.Columns.update({
			'Id': 'Candidate.id',
			'student_name': 'Candidate.student_name',
			'RegistrationDateTime': 'Candidate.RegistrationDateTime',
			'status': 'Candidate.status',
		})

		filterQuery.OrderBy.append('Candidate.student_name')

		if candidateFilter.Page is not None:
			filterQuery.Limit = slice(Candidate.PAGE_SIZE * (candidateFilter.Page - 1),
									  Candidate.PAGE_SIZE * candidateFilter.Page)

		resultSet = db.Execute(str(filterQuery))

		candidateList = [cls(**candidate) for candidate in resultSet]

		return candidateList

	def valid_ExamDateTimes(self):
		#raise NotImplementedError()
		return [row.ExamDateTime
		 		for row in db.Execute('''
		 			SELECT ExamDateTime
		 			FROM CampusExamDateTime
		 			WHERE CampusId = %s
		 			ORDER BY ExamDateTime''',
		 			(self.Campus.Id, ))]

	def ValidAppointmentTimes(self, date):
		# FIXME: consider candidate campus
		return {row.time: row for row in db.Execute('''
			SELECT TIME(AppointmentDateTime) AS time,
				   COUNT(*) AS count
			FROM CandidateAppointment
			WHERE DATE(AppointmentDateTime) = %s
			GROUP BY TIME(AppointmentDateTime)''',
			(date, ))}

	class NetworkError (BaseException):
		pass

	class InvalidLogin (BaseException):
		pass

	def FetchEnemResult(self, cpf, password):
		enemHost = 'http://sistemasenem2.inep.gov.br'

		from urllib.parse import urlencode
		import urllib.request, urllib.error, urllib.parse
		from bs4 import BeautifulSoup

		try:
			request = urllib.request.Request(enemHost + '/resultadosenem/home.seam')
			requestFile = urllib.request.urlopen(request)
			cookie = '; '.join(header.split(': ')[1].split('; ')[0] for header in requestFile.info().getallmatchingheaders('Set-Cookie'))
			requestFile.close()
		except urllib.error.URLError:
			raise Candidate.NetworkError()

		try:
			request = urllib.request.Request(enemHost + '/resultadosenem/home.seam', urlencode([
				('AJAXREQUEST', '_viewRoot'),
				('login2Form', 'login2Form'),
				('anoDecorate:anoSelect', '1401'),
				('cpfDecorate:cpfInput', ''),
				('senha2Decorate:senha2Input', ''),
				('verifyCaptcha', ''),
				('javax.faces.ViewState', 'j_id1'),
				('j_id84', 'j_id84'),
				('ajaxSingle', 'j_id84'),
			]), {'Cookie': cookie})
			requestFile = urllib.request.urlopen(request)
			soup = BeautifulSoup(requestFile.read())
			requestFile.close()
		except urllib.error.URLError:
			raise Candidate.NetworkError()

		try:
			request = urllib.request.Request(enemHost + dict(soup.find('img').attrs)['src'], '', {'Cookie': cookie})
			requestFile = urllib.request.urlopen(request)
			requestFile.read()
			requestFile.close()
		except urllib.error.URLError:
			raise Candidate.NetworkError()

		x, _, y, _ = dict(soup.find('img').attrs)['alt'].split()
		captcha = int(x) + int(y)

		try:
			request = urllib.request.Request(enemHost + '/resultadosenem/home.seam', urlencode([
				('login2Form', 'login2Form'),
				('anoDecorate:anoSelect', '1401'),
				('cpfDecorate:cpfInput', util.cpf(cpf)),
				('senha2Decorate:senha2Input', password),
				('verifyCaptchaSoma', captcha),
				('j_id102.x', '49'),
				('j_id102.y', '20'),
				('javax.faces.ViewState', 'j_id1'),
			]), {'Cookie': cookie})
			requestFile = urllib.request.urlopen(request)
			soup = BeautifulSoup(requestFile.read())
			requestFile.close()
		except urllib.error.URLError:
			raise Candidate.NetworkError()

		if soup.find(id = 'numInscricaoDecorate:numInscricao') is None:
			raise Candidate.InvalidLogin()

		db.Execute('''
			DELETE FROM CandidateENEM WHERE CandidateId = %s''',
			(self.Id, ))
		db.Execute('''
			INSERT INTO CandidateENEM (CandidateId, name, code,
						ch_score, cn_score, lc_score, mt_score,
						rd_score)
			VALUES (%s, %s, %s, %s, %s, %s, %s, %s)''',
			(self.Id,
			 soup.find(id = 'nomeDecorate:nome').text,
			 soup.find(id = 'numInscricaoDecorate:numInscricao').text,
			 float(soup.find(id = 'notasProva:0:j_id104').text),
			 float(soup.find(id = 'notasProva:1:j_id104').text),
			 float(soup.find(id = 'notasProva:2:j_id104').text),
			 float(soup.find(id = 'notasProva:3:j_id104').text),
			 float(soup.find(id = 'notasProva:4:j_id104').text)))
		db.Execute('''
			UPDATE Candidate
			SET status = 'present'
			WHERE Id = %s''',
			(self.Id, ))

	class Appointment (Entity):
		@classmethod
		def get(cls, appointmentId):
			if appointmentId is None:
				return None
			try:
				appointment = db.Single('''
					SELECT Candidate.Id AS `candidate.id`,
						   Candidate.student_name AS student_name,
						   CandidateAppointment.Id,
						   CandidateAppointment.AppointmentDateTime AS datetime,
						   User.Id AS `user.id`,
						   User.FullName AS `user.name`
					FROM CandidateAppointment
						 INNER JOIN Candidate
						   ON CandidateAppointment.CandidateId = Candidate.Id
						 LEFT OUTER JOIN User
						   ON CandidateAppointment.UserId = User.Id
					WHERE CandidateAppointment.Id = %s''',
					(appointmentId, ))
			except ValueError:
				return None
			appointment = cls(**appointment)
			appointment.path = '/candidate/%s/appointment/%s/' % (appointment.candidate.id, appointment.id)

			return appointment

def certificate(candidateIdList, templateFile = 'certificate.docx'):
	from zipfile import ZipFile
	from lawrence.Monte import XmlUtil, WordUtil
	from lawrence.Monte.Util import ZipContents
	from lxml import etree as ElementTree
	import copy

	template = ZipFile(os.path.join(Handler._base_dir, templateFile), 'r')
	for name, file in ZipContents(template):
		if name == 'word/document.xml':
			templateTree = ElementTree.parse(file)
			templateBody = templateTree.find(XmlUtil.ns('w:body'))
			for el in templateBody[:]:
				WordUtil.RemoveRevisions(el)
				if WordUtil.IsEmpty(el):
					templateBody.remove(el)

	candidateList = [Candidate(**candidate) for candidate in db.Execute('''
		SELECT
			Candidate.id AS Id,
			Campus.id AS CampusId,
			Campus.name AS `Campus.Name`,
			Course.id AS CourseId,
			Course.Name AS `Course.Name`,
			Candidate.student_name,
			Course.id AS course_CourseId,
			Course.id AS `course.course.id`
		FROM
			Candidate
			INNER JOIN Campus
			  ON Campus.id = Candidate.CampusId
			INNER JOIN Course
			  ON Course.id = Candidate.CourseId
			INNER JOIN Course
			  ON Course.CourseId = Course.id
		WHERE ''' + db.query_in('Candidate.id', candidateIdList) + '''
		ORDER BY student_name''')]

	def SingleCertificate(candidate, templateBody):
		certificateBody = []
		if not candidate.discounts:
			return
		for el in copy.deepcopy(templateBody)[:]:
			if "[valores]" in WordUtil.ElementText(el):
				for discount in candidate.discounts:
					certificateBody.append(WordUtil.Paragraph(
						"%s (%s) \u2014 %s de bolsa. Mensalidade: %s." % (
							discount.class_name,
							discount.shift.name,
							util.percentage()(discount.merit_discount),
							util.money(discount.total_payment))))
				continue
			XmlUtil.ReplaceText(el, "[número de inscrição]", str(candidate.EncodedId))
			XmlUtil.ReplaceText(el, "[nome do aluno]", candidate.student_name)
			XmlUtil.ReplaceText(el, "[unidade]", candidate.Campus.Name)
			XmlUtil.ReplaceText(el, "[campanha]", candidate.Course.Name)
			XmlUtil.ReplaceText(el, "[total de acertos]", str(sum(candidate.grades.values())))
			for subject_name, grade in list(candidate.grades.items()):
				XmlUtil.ReplaceText(el, "[acertos em %s]" % subject_name, str(grade))
			certificateBody.append(el)
		return certificateBody

	outputFile = BytesIO()
	with ZipFile(outputFile, 'w') as outputDocx:
		for name, file in ZipContents(template):
			if name == 'word/document.xml':
				certificateTree = copy.deepcopy(templateTree)
				certificateBody = certificateTree.find(XmlUtil.ns('w:body'))
				for el in certificateBody[:]:
					certificateBody.remove(el)
				certificateCount = 0
				for candidate in candidateList:
					certificateElements = SingleCertificate(candidate, templateBody)
					if certificateElements is None:
						continue
					if certificateCount > 0:
						certificateBody.append(WordUtil.PageBreak())
					certificateCount += 1
					for el in certificateElements:
						certificateBody.append(el)
				data = XmlUtil.DumpTree(certificateTree)
			else:
				data = file.read()
			outputDocx.writestr(name, data)

	return outputFile.getvalue()

def xlsx(candidateIdList):
	candidateList = db.Execute('''
		SELECT
			Candidate.id AS Id,
			Course.Name AS `Course.Name`,
			Candidate.RegistrationDateTime AS RegistrationDateTime,
			DATE(Candidate.ExamDateTime) AS ExamDate,
			TIME(Candidate.ExamDateTime) AS ExamTime,
			Candidate.status AS status,
			Candidate.contact_name AS contact_name,
			Candidate.contact_email AS contact_email,
			Candidate.student_name AS student_name,
			Candidate.school AS school,
			Candidate.BirthDate AS BirthDate,
			SUBSTRING_INDEX(Candidate.CityName, '||', 1) AS DistrictName,
			SUBSTRING_INDEX(Candidate.CityName, '||', -1) AS CityName,
			Campus.name AS `Campus.Name`,
			SUM(CandidateGrade.grade) AS total_grade,
			CASE WHEN Candidate.status != 'present'
				   THEN ''
				 WHEN CandidateAppointment.AppointmentDateTime IS NULL
				   THEN 'unscheduled'
				 WHEN DATE(CandidateAppointment.AppointmentDateTime) >= date(NOW()) AND CandidateAppointment.UserId IS NULL
				   THEN 'scheduled'
				 WHEN CandidateAppointment.UserId IS NULL
				   THEN 'missed'
				 WHEN DATE(CandidateAppointment.AppointmentDateTime) >= DATE_SUB(DATE(NOW()), INTERVAL 7 DAY)
				   THEN 'met_recently'
				 ELSE 'met' END AS AppointmentStatus,
			last_interviewer_user.FullName AS last_interviewer,
			NextAppointment.AppointmentDateTime as NextAppointmentDateTime
		FROM
			Candidate
			LEFT OUTER JOIN Course
			  ON Candidate.CourseId = Course.id
			LEFT OUTER JOIN Campus
			  ON Candidate.CampusId = Campus.id
			LEFT OUTER JOIN CandidateGrade
			  ON CandidateGrade.CandidateId = Candidate.id
			LEFT OUTER JOIN CandidateAppointment
			  ON CandidateAppointment.CandidateId = Candidate.id AND
				 COALESCE(CandidateAppointment.AppointmentDateTime > Candidate.ExamDateTime, 1) AND
				 CandidateAppointment.AppointmentDateTime =
				   (SELECT MAX(aux.AppointmentDateTime)
					FROM CandidateAppointment AS aux
					WHERE aux.CandidateId = CandidateAppointment.CandidateId)
			LEFT OUTER JOIN CandidateAppointment AS last_interviewer
			  ON last_interviewer.CandidateId = Candidate.id AND
				 last_interviewer.AppointmentDateTime =
				   (SELECT MAX(aux.AppointmentDateTime)
					FROM CandidateAppointment AS aux
					WHERE aux.CandidateId = CandidateAppointment.CandidateId AND
						  aux.UserId IS NOT NULL)
			LEFT OUTER JOIN CandidateAppointment AS NextAppointment
			  ON NextAppointment.CandidateId = Candidate.id AND
				 NextAppointment.AppointmentDateTime =
				   (SELECT MIN(aux.AppointmentDateTime)
					FROM CandidateAppointment AS aux
					WHERE aux.CandidateId = CandidateAppointment.CandidateId AND
						  aux.UserId IS NULL AND
						  DATE(aux.AppointmentDateTime) >= DATE(NOW()))
			LEFT OUTER JOIN User AS last_interviewer_user
			  ON last_interviewer.UserId = last_interviewer_user.id
		WHERE ''' + db.query_in('Candidate.id', candidateIdList) + '''
		GROUP BY Candidate.id''')

	candidateMap = {}
	for candidateRow in candidateList:
		candidate = Candidate(**candidateRow)
		candidate.ContactPhones = []
		candidate.TagList = []
		candidate.MorningDiscount = None
		candidate.AfternoonDiscount = None
		candidateMap[candidateRow.Id] = candidate
	for row in db.Execute('''
			SELECT CandidateId, phone AS Phone
			FROM CandidatePhone
			WHERE ''' + db.query_in('CandidateId', candidateIdList)):
		candidateMap[row.CandidateId].ContactPhones.append(row.Phone)
	for row in db.Execute('''
			SELECT CandidateTagMap.CandidateId, CandidateTag.Name
			FROM
				CandidateTagMap
				INNER JOIN CandidateTag
				  ON CandidateTagMap.TagId = CandidateTag.Id
			WHERE ''' + db.query_in('CandidateId', candidateIdList)):
		candidateMap[row.CandidateId].TagList.append(row.Name)
	for row in db.Execute('''
			SELECT
				CandidateId,
				ShiftId,
				MAX(LowerBound) AS HighestLowerBound
			FROM
				CandidateNegotiation
				INNER JOIN SchoolClass
				  ON CandidateNegotiation.SchoolClassId = SchoolClass.Id
			WHERE ''' + db.query_in('CandidateId', candidateIdList) + '''
			GROUP BY
				CandidateId,
				ShiftId'''):
		if row.ShiftId == 1:
			candidateMap[row.CandidateId].MorningDiscount = row.HighestLowerBound
		if row.ShiftId == 2:
			candidateMap[row.CandidateId].AfternoonDiscount = row.HighestLowerBound

	for candidate in candidateMap.values():
		candidate.ExamTime = util.ShortTime(candidate.ExamTime) if candidate.ExamTime is not None else '=NA()'  # FIXME
		candidate.status = Candidate.STATUSES[candidate.status]
		if candidate.AppointmentStatus:
			candidate.AppointmentStatus = Candidate.APPOINTMENT_STATUSES[candidate.AppointmentStatus]
		candidate.EncodedId = util.EncodeId(candidate.Id)
		if candidate.BirthDate is not None:
			candidate.StudentAge = util.AgeInYearsAndMonths(candidate.BirthDate)
		else:
			candidate.StudentAge = None
		candidate.contact_phone = " / ".join([util.phone(phone) for phone in candidate.ContactPhones])
		candidate.Tags = "; ".join(candidate.TagList)
		if candidate.total_grade is not None:
			candidate.total_grade = str(candidate.total_grade)
	
	columns = [
		("Código", 12, lambda candidate: candidate.EncodedId),
		("Data", 12, lambda candidate: candidate.RegistrationDateTime),
		("Campanha", 12, lambda candidate: candidate.Course.Name),
		("Unidade", 20, lambda candidate: candidate.Campus.Name),
		("Data do exame", 12, lambda candidate: candidate.ExamDate),
		("Hora do exame", 12, lambda candidate: candidate.ExamTime),
		("Situação", 12, lambda candidate: candidate.status),
		("Situação do atendimento", 12, lambda candidate: candidate.AppointmentStatus),
		("Nota no Bolsão", 12, lambda candidate: candidate.total_grade),
		("Nome do contato", 25, lambda candidate: candidate.contact_name),
		("Email", 25, lambda candidate: candidate.contact_email),
		("Telefone", 15, lambda candidate: candidate.contact_phone),
		("Nome do aluno", 40, lambda candidate: candidate.student_name),
		("Data de nascimento", 25, lambda candidate: candidate.BirthDate),
		("Idade", 25, lambda candidate: candidate.StudentAge),
		("Bairro do aluno", 40, lambda candidate: candidate.DistrictName),
		("Cidade do aluno", 40, lambda candidate: candidate.CityName),
		("Colégio atual", 25, lambda candidate: candidate.school),
		("Último entrevistador", 25, lambda candidate: candidate.last_interviewer),
		("Próximo atendimento", 25, lambda candidate: candidate.NextAppointmentDateTime),
		("Marcadores", 25, lambda candidate: candidate.Tags),
		("Bolsa concedida (manhã)", 25, lambda candidate: candidate.MorningDiscount),
		("Bolsa concedida (tarde)", 25, lambda candidate: candidate.AfternoonDiscount),
	]

	workbook = Workbook(write_only = True)
	worksheet = workbook.create_sheet()
	worksheet.title = "Campanhas"

	headerRow = []
	for j, (header, width, field) in enumerate(columns):
		headerRow.append(header)
		# worksheet.column_dimensions[j+1].width = width
	worksheet.append(headerRow)

	for candidate in candidateMap.values():
		candidateRow = [field(candidate) for (header, width, field) in columns]
		worksheet.append(candidateRow)

	return workbook

def roll(candidateIdList):
	workbook = load_workbook(os.path.join(Handler._base_dir, 'roll.xlsx'))
	worksheet = workbook.get_active_sheet()

	candidateList = [Candidate(**candidate) for candidate in db.Execute('''
		SELECT
			Candidate.id AS Id,
			Course.Name AS `course.name`,
			Candidate.student_name AS student_name
		FROM
			Candidate
			INNER JOIN Course
			  ON Candidate.CourseId = Course.id
		WHERE ''' + db.query_in('Candidate.id', candidateIdList) + '''
		ORDER BY student_name''')]

	for i, candidate in enumerate(candidateList):
		row = i+1
		worksheet.cell('A%s' % row).value = candidate.course.name
		worksheet.cell('B%s' % row).value = candidate.EncodedId
		worksheet.cell('C%s' % row).value = candidate.student_name.upper()
		for col in ('A', 'B', 'C'):
			worksheet.cell('%s%s' % (col, row)).style = worksheet.cell('%s1' % col).style
		worksheet.row_dimensions[row].height = 30

	return workbook

