import base64
from functools import wraps
from uuid import UUID

from Ozark import DB, ORM, HTML, HTTP
from Ozark.Request import Request, RequestFactory, RequestFields
from Ozark.Injector import InjectorFactory
from Ozark.Route import RouteMismatchError, RouteRegistry, PrefixedRouteRegistry, RouteTarget
from Ozark.Widget import Widget, WidgetRegistry

def StringEncoder(encoding: str):
	def decorator(f):
		@wraps(f)
		def wrapper(*args, **kwargs):
			for s in f(*args, **kwargs):
				if isinstance(s, bytes):
					yield s
				elif isinstance(s, str):
					yield s.encode(encoding)
				else:
					raise TypeError("invalid type '{0}'".format(type(s)))
		return wrapper
	return decorator

def ScalarFactory(*, argumentName: str, argumentType: type, requestFields: RequestFields):
	return list(map(argumentType, requestFields.PopValues(argumentName)))

def StringFactory(*, argumentName: str, requestFields: RequestFields):
	return list(map(str, requestFields.PopValues(argumentName)))

def NumberFactory(*, argumentName: str, argumentType: type, requestFields: RequestFields):
	if argumentType is bool:
		return list(map(lambda value: {'true': True, 'false': False}[value], requestFields.PopValues(argumentName)))
	elif argumentType is int:
		return list(map(int, requestFields.PopValues(argumentName)))
	else:
		return list(map(float, requestFields.PopValues(argumentName)))

def BytesFactory(*, argumentName: str, requestFields: RequestFields):
	return list(map(lambda s: base64.b64decode(s.replace('.', '='), '-_'), requestFields.PopValues(argumentName)))

def EntityReaderFactory(*, requestTarget, cursor: DB.Connection.Cursor):
	return [ORM.EntityReader(cursor, requestTarget)]

def EntityWriterFactory(*, requestTarget, cursor: DB.Connection.Cursor):
	return [ORM.EntityWriter(cursor, requestTarget)]

def EntityFactory(*,
		argumentName,
		argumentType,
		entityReader: ORM.EntityReader,
		requestFields):
	# TODO: pull several entities in a single query
	entityList = list(map(lambda key: entityReader.GetById(argumentType, UUID(key)), requestFields.PopValues(argumentName + 'Id')))
	if None in entityList:
		raise ValueError
	return entityList

def WidgetFactory(*, widgetRegistry, argumentType, injector):
	return list(injector(widgetRegistry.GetWidgets, widgetType = argumentType))

def RouteTargetFactory(*, routeRegistry: RouteRegistry, argumentName):
	return [
		route.Target for routeName, route in routeRegistry._Routes.items()
		if routeName == argumentName]

class Application:
	def __init__(self, cursorFactory):
		cursor = cursorFactory()
		schema = DB.Database.GetSchema(cursor)
		self.AnchorModel = ORM.Model.Model.CreateFromSchema(schema)

		routeInjectorFactory = InjectorFactory()
		routeInjectorFactory.AddByTypeArgumentFactory(lambda: self.AnchorModel, ORM.Model.Model)
		routeInjectorFactory.AddByTypeArgumentFactory(lambda *, argumentName: self.AnchorModel[argumentName], ORM.Anchor)
		self.RouteRegistry = RouteRegistry(RouteInjectorFactory = routeInjectorFactory)
		routeInjectorFactory.AddByTypeArgumentFactory(lambda: self.RouteRegistry, RouteRegistry)

		widgetInjectorFactory = InjectorFactory()
		widgetInjectorFactory.AddByTypeArgumentFactory(lambda: self.AnchorModel, ORM.Model.Model)
		widgetInjectorFactory.AddByTypeArgumentFactory(lambda *, argumentName: self.AnchorModel[argumentName], ORM.Anchor)
		self.WidgetRegistry = WidgetRegistry(WidgetInjectorFactory = widgetInjectorFactory)
		routeInjectorFactory.AddByTypeArgumentFactory(lambda: self.WidgetRegistry, WidgetRegistry)

		self.RequestInjectorFactory = InjectorFactory()
		self.RequestInjectorFactory.AddByTypeArgumentFactory(StringFactory, str)
		self.RequestInjectorFactory.AddByTypeArgumentFactory(BytesFactory, bytes)
		self.RequestInjectorFactory.AddByTypeArgumentFactory(NumberFactory, int)
		self.RequestInjectorFactory.AddByTypeArgumentFactory(NumberFactory, float)
		self.RequestInjectorFactory.AddByTypeArgumentFactory(ScalarFactory, UUID)
		self.RequestInjectorFactory.AddByTypeArgumentFactory(lambda: cursorFactory(echo = False), DB.Connection.Cursor)
		self.RequestInjectorFactory.AddByTypeArgumentFactory(EntityWriterFactory, ORM.EntityWriter)
		self.RequestInjectorFactory.AddByTypeArgumentFactory(EntityReaderFactory, ORM.EntityReader)
		self.RequestInjectorFactory.AddByTypeArgumentFactory(EntityFactory, ORM.Anchor)
		self.RequestInjectorFactory.AddByTypeArgumentFactory(WidgetFactory, Widget)
		self.RequestInjectorFactory.AddByTypeArgumentFactory(RouteTargetFactory, RouteTarget)

		self.RequestInjectorFactory.AddByNameArgumentFactory(lambda: self, argumentName = 'application')
		self.RequestInjectorFactory.AddByNameArgumentFactory(lambda: self.WidgetRegistry, argumentName = 'widgetRegistry')
		self.RequestInjectorFactory.AddByNameArgumentFactory(lambda: self.RouteRegistry, argumentName = 'routeRegistry')

		self.RequestFactory = RequestFactory(requestInjectorFactory = self.RequestInjectorFactory)

	def GetResponse(self, request: Request):
		for route in self.RouteRegistry._Routes.values():
			try:
				result = request.Dispatch(route)
			except RouteMismatchError:
				pass
			except Exception:
				raise
			# except Exception as ex:
			# 	return HTTP.Response(str(ex), Status = 400)
			else:
				# FIXME: create a ResponseFactoryRegistry
				if isinstance(result, HTTP.Response):
					return result
				elif isinstance(result, HTML.Document):
					return HTTP.Response(result.ToHTML(application = self, routeRegistry = self.RouteRegistry))
				elif isinstance(result, HTTP.Redirect):
					return result.ToResponse(routeRegistry = self.RouteRegistry)
				else:
					raise TypeError("unknown response type {}".format(type(result).__name__))
		else:
			return HTTP.Response(Status = 404)

	@StringEncoder('utf-8')
	def __call__(self, environ: dict, startResponse: callable):
		request = self.RequestFactory(environ)
		response = self.GetResponse(request)

		response.SendHeaders(startResponse)
		return response

	def AddModule(self, moduleFactory: callable, *, routePrefix: str = ''):
		injectorFactory = InjectorFactory()
		injectorFactory.AddByTypeArgumentFactory(lambda: self, Application)
		injectorFactory.AddByTypeArgumentFactory(lambda: self.AnchorModel, ORM.Model.Model)
		injectorFactory.AddByTypeArgumentFactory(lambda: self.WidgetRegistry, WidgetRegistry)
		# injectorFactory.AddByTypeArgumentFactory(lambda: self.RequestFactory, RequestFactory)
		injectorFactory.AddByNameArgumentFactory(lambda: self.RequestInjectorFactory, argumentName = 'requestInjectorFactory')

		# TODO: get rid of this so we can store injectorFactory in a property
		injectorFactory.AddByNameArgumentFactory(lambda: PrefixedRouteRegistry(self.RouteRegistry, routePrefix), argumentName = 'routeRegistry')
		injector = injectorFactory()
		injector(moduleFactory)