from _collections_abc import Iterable
from abc import ABCMeta, abstractmethod

import psycopg2
from psycopg2._psycopg import ProgrammingError
from psycopg2.extras import RealDictCursor, register_uuid

class CursorFactory:
	def __init__(self, *, connection, compilerFactory):
		self._Connection = connection
		self._CompilerFactory = compilerFactory

	def __call__(self, **kwargs):
		return Cursor(self._Connection.cursor(), compiler = self._CompilerFactory(), **kwargs)

class PostgreSQLCursorFactory (CursorFactory):
	def __init__(self, host, port, database, user, password):
		connection = psycopg2.connect(
			host = host,
			port = port,
			database = database,
			user = user,
			password = password,
			cursor_factory = RealDictCursor)
		from Ozark.DB.SQLCompiler.PostgreSQLCompiler import PostgreSQLCompiler
		compilerFactory = PostgreSQLCompiler

		CursorFactory.__init__(self, connection = connection, compilerFactory = compilerFactory)
		register_uuid(conn_or_curs = self._Connection)

class Cursor:
	def __init__(self, cursor, compiler, *, echo: bool = False):
		self._Compiler = compiler
		self._Cursor = cursor
		self._Echo = echo

	def __del__(self):
		if not self._Cursor.closed:
			self.Rollback()
			self.Close()

	def Execute(self, query):
		compiledQuery = self._Compiler.CompileQuery(query)
		if self._Echo:
			print(compiledQuery.decode('utf-8') + '\n')
		self._Cursor.execute(compiledQuery)
		try:
			return EvaluatedResultSet(self._Cursor.fetchall())
		except ProgrammingError:
			return None

	def Begin(self) -> None:
		self._Cursor.execute('BEGIN')

	def Commit(self) -> None:
		self._Cursor.execute('COMMIT')

	def Rollback(self) -> None:
		self._Cursor.execute('ROLLBACK')

	def Close(self):
		if not self._Cursor.closed:
			self._Cursor.close()

class ResultSet (Iterable, metaclass = ABCMeta):
	class NoResultTuples (Exception):
		pass

	@abstractmethod
	def __iter__(self):
		raise NotImplementedError()

	def Scalar(self) -> object:
		# TODO: remove this method
		resultIterator = iter(self)
		try:
			resultTuple = resultIterator.__next__()
		except StopIteration as ex:
			raise ResultSet.NoResultTuples() from ex
		return resultTuple.popitem()[1]

class EvaluatedResultSet (ResultSet):
	def __init__(self, resultSet):
		self._ResultSet = resultSet

	def __iter__(self):
		return iter(self._ResultSet)