from psycopg2.extensions import adapt

from Ozark import QueryAST

class PostgreSQLCompiler:
	#  11:  (function call)
	#  10:  .	left	table/column name separator
	#   9:  ::	left	PostgreSQL-style typecast
	#   8:  [ ]	left	array element selection
	#   7:  -	right	unary minus
	#   6:  ^	left	exponentiation
	#   5:  * / %	left	multiplication, division, modulo
	#   4:  + -	left	addition, subtraction
	#   3:  IS	 	IS TRUE, IS FALSE, IS UNKNOWN, IS NULL
	#   2:  ISNULL	 	test for null
	#   1:  NOTNULL	 	test for not null
	#   0:  (any other)	left	all other native and user-defined operators
	#  -1:  IN	 	set membership
	#  -2:  BETWEEN	 	range containment
	#  -3:  OVERLAPS	 	time interval overlap
	#  -4:  LIKE ILIKE SIMILAR	 	string pattern matching
	#  -5:  < >	 	less than, greater than
	#  -6:  =	right	equality, assignment
	#  -7:  NOT	right	logical negation
	#  -8:  AND	left	logical conjunction
	#  -9:  OR	left	logical disjunction
	# -10:  (parenthesised expressions)
	# -10:  (function call arguments)

	_OperatorPrecedence = {
		QueryAST.Expression.Minimum: 11,
		QueryAST.Expression.Maximum: 11,

		QueryAST.Expression.Boole: 9,

		QueryAST.Expression.Multiplication: 5,
		QueryAST.Expression.Division: 5,

		QueryAST.Expression.Addition: 4,
		QueryAST.Expression.Subtraction: 4,

		QueryAST.Expression.ElementOf: -1,
		QueryAST.Expression.NotElementOf: -1,

		QueryAST.Expression.LessThanOrEqualTo: -5,
		QueryAST.Expression.LessThan: -5,
		QueryAST.Expression.GreaterThanOrEqualTo: -5,
		QueryAST.Expression.GreaterThan: -5,

		QueryAST.Expression.NotEqualTo: -6,
		QueryAST.Expression.EqualTo: -6,

		QueryAST.Expression.Conjunction: -8,

		QueryAST.Expression.Disjunction: -9,

		type(None): -10}

	_OperatorName = {
		QueryAST.Expression.Disjunction: b'OR',
		QueryAST.Expression.Conjunction: b'AND',

		QueryAST.Expression.LessThanOrEqualTo: b'<=',
		QueryAST.Expression.LessThan: b'<',
		QueryAST.Expression.GreaterThanOrEqualTo: b'>=',
		QueryAST.Expression.GreaterThan: b'>',

		QueryAST.Expression.NotEqualTo: b'<>',
		QueryAST.Expression.EqualTo: b'=',

		QueryAST.Expression.ElementOf: b'IN',
		QueryAST.Expression.NotElementOf: b'NOT IN',

		QueryAST.Expression.Multiplication: b'*',
		QueryAST.Expression.Division: b'/',

		QueryAST.Expression.Addition: b'+',
		QueryAST.Expression.Subtraction: b'-'}

	def CompileOrderBy(self, orderBy):
		if isinstance(orderBy, QueryAST.Ordering.Descending):
			return self.CompileExpression(orderBy.Expression) + b' DESC'
		if isinstance(orderBy, QueryAST.Ordering.Ascending):
			return self.CompileExpression(orderBy.Expression) + b' ASC'
		raise NotImplementedError(orderBy)

	def CompileScalar(self, scalar):
		if isinstance(scalar, frozenset):
			return b'(' + b', '.join(self.CompileScalar(element) for element in scalar) + b')'
		if isinstance(scalar, str):
			return self.QuoteString(scalar)
		# noinspection PyArgumentList
		return adapt(scalar).getquoted()

	def CompileExpression(self, expression: QueryAST.Expression.Expression, *, parentExpression = None) -> bytes:
		if not isinstance(expression, QueryAST.Expression.Expression):
			return self.CompileScalar(expression)
		if isinstance(expression, QueryAST.Expression.FunctionCall):
			return self.CompileFunctionCall(expression)
		if isinstance(expression, QueryAST.Expression.Column):
			return self.QuoteIdentifier(expression.Name)

		wrapExpression = lambda compiledExpression: compiledExpression
		if PostgreSQLCompiler._OperatorPrecedence.get(expression.__class__, 0) <= PostgreSQLCompiler._OperatorPrecedence.get(parentExpression.__class__, 0):
			wrapExpression = lambda compiledExpression: b'(' + compiledExpression + b')'
		if expression.__class__ in self.__class__._OperatorName:
			operatorName = b' ' + self.__class__._OperatorName[expression.__class__] + b' '
			return wrapExpression(operatorName.join(
				self.CompileExpression(constraint, parentExpression = expression)
				for constraint in expression.Terms))
		if isinstance(expression, QueryAST.Expression.Minimum):
			return b'LEAST(' + b', '.join(self.CompileExpression(expression) for expression in expression.Terms) + b')'
		if isinstance(expression, QueryAST.Expression.Maximum):
			return b'GREATEST(' + b', '.join(self.CompileExpression(expression) for expression in expression.Terms) + b')'
		if isinstance(expression, QueryAST.Expression.Boole):
			return b'CASE WHEN ' + self.CompileExpression(expression.Expression) + b' THEN 1.0 ELSE 0.0 END'
		if isinstance(expression, QueryAST.Expression.StartsWith):
			return self.CompileExpression(expression.LeftTerm, parentExpression = expression) + b' ~~ (REGEXP_REPLACE(' + self.CompileExpression(expression.RightTerm, parentExpression = expression) + b', \'([_%])\', \'\\\\\\1\', \'g\') || \'%\')'
		raise NotImplementedError(expression)

	def CompileTable(self, table: QueryAST.Table.Table) -> bytes:
		if isinstance(table, QueryAST.Table.SchemaTable):
			if table.Schema == 'public':
				return self.QuoteIdentifier(table.Name)
			else:
				return self.QuoteIdentifier(table.Schema) + b'.' + self.QuoteIdentifier(table.Name)
		if isinstance(table, QueryAST.Table.QueryTable):
			return b'(' + self.CompileQuery(table.Query) + b') AS ' + self.QuoteIdentifier(table.Name)
		raise NotImplementedError(table)

	def CompileColumn(self, column, columnName):
		if isinstance(column, QueryAST.Expression.Column) and column.Name == columnName:
			return self.QuoteIdentifier(columnName)
		return self.CompileExpression(column) + b' AS ' + self.QuoteIdentifier(columnName)

	def CompileQuery(self, query: QueryAST.Query.Query) -> bytes:
		if isinstance(query, QueryAST.Query.Raw):  # TODO: kill this object
			# noinspection PyArgumentList
			return (query._Statement % {key: adapt(value).getquoted().decode('utf-8') for key, value in query._Parameters.items()}).encode('utf-8')
		if isinstance(query, QueryAST.Query.Select):
			statement = b'SELECT DISTINCT\n'  # TODO: make this switchable
			statement += b',\n'.join(b'\t' + self.CompileColumn(column, columnName) for columnName, column in sorted(query.Column.items(), key = lambda t: t[0]))
			if query.Table:
				firstTable, _ = query.Table[0]
				statement += b'\nFROM\n'
				statement += b'\t' + self.CompileTable(firstTable)
				for table, join in query.Table[1:]:
					statement += b'\n\tNATURAL ' + join + b' JOIN '
					statement += self.CompileTable(table)
			if query.Constraint != True:
				statement += b'\nWHERE ' + self.CompileExpression(query.Constraint)
			if query.GroupBy:
				statement += b'\nGROUP BY\n'
				statement += b', \n'.join(b'\t' + self.CompileExpression(expression) for expression in query.GroupBy)
			if query.OrderBy:
				statement += b'\nORDER BY\n'
				statement += b', \n'.join(b'\t' + self.CompileOrderBy(expression) for expression in query.OrderBy)
			return statement
		if isinstance(query, QueryAST.Query.InsertSelect):
			statement = b'INSERT INTO ' + self.CompileTable(query.Table)
			statement += b' (\n' + b',\n'.join(b'\t' + self.QuoteIdentifier(columnName) for columnName, column in sorted(query.Query.Column.items(), key = lambda t: t[0])) + b')\n'
			statement += self.CompileQuery(query.Query)
			return statement
		if isinstance(query, QueryAST.Query.InsertTuple):
			statement = b'INSERT INTO ' + self.CompileTable(query.Table)
			statement += b' (\n' + b',\n'.join(b'\t' + self.QuoteIdentifier(columnName) for columnName, column in sorted(query.Table.Column.items(), key = lambda t: t[0])) + b')\n'
			statement += b'  VALUES\n'
			statement += b',\n'.join(b'\t(' + b', '.join(self.CompileExpression(column) for columnName, column in sorted(t.items(), key = lambda t: t[0])) + b')' for t in query.Tuple)
			return statement
		if isinstance(query, QueryAST.Query.Delete):
			statement = b'DELETE FROM ' + self.CompileTable(query.Table)
			if query.Constraint != True:
				statement += b'\nWHERE ' + self.CompileExpression(query.Constraint)
			return statement
		if isinstance(query, QueryAST.Query.CompositeQuery):
			return b';\n'.join(self.CompileQuery(query) for query in query.Queries)
		raise NotImplementedError(query)

	def CompileFunctionCall(self, function) -> bytes:
		if isinstance(function, QueryAST.Expression.CurrentTimestamp):
			return b'CURRENT_TIMESTAMP'
		if isinstance(function, QueryAST.Expression.AggregateMaximum):
			return b'MAX(' + self.CompileExpression(function.Expression) + b')'
		if isinstance(function, QueryAST.Expression.Generic):
			return self.QuoteIdentifier(function.Name) + b'(' + b', '.join(self.CompileExpression(obj) for obj in function.Arguments) + b')'
		raise NotImplementedError(function)

	@staticmethod
	def QuoteString(string: str):
		string = string.encode('utf-8')
		return b"'" + string.replace(b"'", b"''") + b"'"

	@staticmethod
	def QuoteIdentifier(identifier: str):
		# TODO: create an Identifier class
		identifier = identifier.encode('utf-8')
		return b'"' + identifier.replace(b'"', b'""') + b'"'
