import re
from ..Typing import List

def Escape(s, escapeQuotes = True):
	"""Replace special characters by HTML-safe sequences.

:param str s: string to be escaped
:param bool escapeQuotes: also escape single and double quotation marks"""

	s = s.replace('&', '&amp;')
	s = s.replace('<', '&lt;')
	s = s.replace('>', '&gt;')
	if escapeQuotes:
		s = s.replace('"', '&quot;')
		s = s.replace('\'', '&#x27;')
	return s

class Node:
	def ToHTML(self, **kwargs):
		"""Serialize this node to HTML."""
		raise NotImplementedError()

class Text (Node):
	"""A piece of text on the page."""
	def __init__(self, Content: str):
		Node.__init__(self)
		self.Content = Content

	def ToHTML(self, **kwargs):
		yield Escape(self.Content, escapeQuotes = False)

class Element (Node):
	_AttributeRegex = re.compile(r'^[^\t\n\f />"\'=]+$')
	_TagName = NotImplemented

	def __init__(self, *, Id: str = None, Class: str = None, Data: dict = None):
		Node.__init__(self)
		self.Id = Id
		self.Class = Class
		self.Data = Data

	def _Attributes(self, **kwargs) -> dict:
		attributes = {}
		if self.Id is not None:
			attributes['id'] = self.Id
		if self.Class is not None:
			attributes['class'] = self.Class
		if self.Data is not None:
			for key in self.Data:
				attributes['data-' + re.compile('[A-Z]').sub(lambda t: '-' + t.group(0).lower(), key)] = str(self.Data[key])
		return attributes

	def _OpeningTag(self, **kwargs):
		yield "<"
		assert isinstance(self._TagName, str)
		yield self._TagName
		for key, value in self._Attributes(**kwargs).items():
			yield " "
			assert Element._AttributeRegex.match(key)
			yield str(key)
			yield "=\""
			yield Escape(value, escapeQuotes = True)
			yield "\""
		yield ">"

class EmptyElement (Element):
	"""An element that cannot have any child nodes."""

	def ToHTML(self, **kwargs):
		yield from self._OpeningTag(**kwargs)

class NonVoidElement (Element):
	def __init__(self, *children, **kwargs):
		Element.__init__(self, **kwargs)
		self._Children = []
		# No direct assignment so that the filtering in .Append can be applied.
		for node in children:
			self.Append(node)

	def Append(self, node):
		if isinstance(node, str):
			node = Text(node)
		elif not isinstance(node, Node):
			raise TypeError("elements can contain Node and str, not {}".format(type(node).__name__))
		self._Children.append(node)

	# noinspection PyUnusedLocal
	def _ClosingTag(self, **kwargs):
		yield "</"
		assert isinstance(self._TagName, str)
		yield self._TagName
		yield ">"

	def ToHTML(self, **kwargs):
		yield from self._OpeningTag(**kwargs)
		for node in self._Children:
			yield from node.ToHTML(**kwargs)
		yield from self._ClosingTag(**kwargs)

class Form (NonVoidElement):
	_TagName = 'form'

	def __init__(self, *children, Target, **kwargs):
		NonVoidElement.__init__(self, *children, **kwargs)
		self.Target = Target

	def _Attributes(self, *, routeRegistry, **kwargs):
		attributes = NonVoidElement._Attributes(self, **kwargs)
		attributes['action'], attributes['method'] = routeRegistry.GetTargetAndMethod(self.Target)
		return attributes

class TextInput (EmptyElement):
	_TagName = 'input'

	def __init__(self, *, Name: str, Required: bool = False, **kwargs):
		EmptyElement.__init__(self, **kwargs)
		self.Name = Name
		self.Required = Required

	def _Attributes(self, **kwargs):
		attributes = EmptyElement._Attributes(self, **kwargs)
		attributes['type'] = 'text'
		attributes['name'] = self.Name
		if self.Required:
			attributes['required'] = 'required'
		return attributes

class PasswordInput (EmptyElement):
	_TagName = 'input'

	def __init__(self, *, Name: str, Required: bool = False, **kwargs):
		EmptyElement.__init__(self, **kwargs)
		self.Name = Name
		self.Required = Required

	def _Attributes(self, **kwargs):
		attributes = EmptyElement._Attributes(self, **kwargs)
		attributes['type'] = 'password'
		attributes['name'] = self.Name
		if self.Required:
			attributes['required'] = 'required'
		return attributes

class FileInput (EmptyElement):
	_TagName = 'input'

	def __init__(self, *, Name: str, Accept: List[str] = None, **kwargs):
		EmptyElement.__init__(self, **kwargs)
		self.Name = Name
		self.Accept = Accept

	def _Attributes(self, **kwargs):
		attributes = EmptyElement._Attributes(self, **kwargs)
		attributes['type'] = 'file'
		attributes['autocomplete'] = 'off'
		attributes['multiple'] = 'multiple'
		attributes['name'] = self.Name
		if self.Accept is not None:
			attributes['accept'] = ','.join(self.Accept)
		return attributes

# class PasswordInput (EmptyElement):
# 	_TagName = 'input'
#
# 	def __init__(self, *, name: str):
# 		EmptyElement.__init__(self)
# 		self.Name = name
#
# 	def _Attributes(self, **kwargs):
# 		return {
# 			'type': 'password',
# 			'name': self.Name,
# 		}
#
# class SearchInput (EmptyElement):
# 	_TagName = 'input'
#
# 	def __init__(self, *, name: str):
# 		EmptyElement.__init__(self)
# 		self.Name = name
#
# 	def _Attributes(self, **kwargs):
# 		return {
# 			'type': 'search',
# 			'name': self.Name,
# 		}

class SubmitButton (NonVoidElement):
	_TagName = 'button'

	def __init__(self, *children, Enabled: bool = True, **kwargs):
		NonVoidElement.__init__(self, *children, **kwargs)
		self.Enabled = Enabled

	def _Attributes(self, **kwargs):
		attributes = NonVoidElement._Attributes(self, **kwargs)
		attributes['type'] = 'submit'
		if not self.Enabled:
			attributes['disabled'] = 'disabled'
		return attributes

class Heading (NonVoidElement):
	def __init__(self, *children, level: int, **kwargs):
		NonVoidElement.__init__(self, *children, **kwargs)
		self._TagName = 'h{}'.format(level)

class InlineScript (NonVoidElement):
	_TagName = 'script'

class ExternalScript (NonVoidElement):
	_TagName = 'script'

	def __init__(self, *children, Target, **kwargs):
		NonVoidElement.__init__(self, *children, **kwargs)
		self.Target = Target

	def _Attributes(self, routeRegistry, **kwargs):
		attributes = NonVoidElement._Attributes(self, **kwargs)
		attributes['src'] = routeRegistry.GetTarget(self.Target)
		return attributes

class Paragraph (NonVoidElement):
	_TagName = 'p'

class Time (NonVoidElement):
	_TagName = 'time'

	def __init__(self, *children, DateTime, **kwargs):
		NonVoidElement.__init__(self, *children, **kwargs)
		self.DateTime = DateTime

	def _Attributes(self, **kwargs):
		attributes = NonVoidElement._Attributes(self, **kwargs)
		attributes['datetime'] = self.DateTime.isoformat()
		return attributes

class AnchorLink (NonVoidElement):
	_TagName = 'a'

	def __init__(self, *children, Target, **kwargs):
		NonVoidElement.__init__(self, *children, **kwargs)
		self.Target = Target

	def _Attributes(self, *, routeRegistry, **kwargs):
		attributes = NonVoidElement._Attributes(self, **kwargs)
		attributes['href'] = routeRegistry.GetTarget(self.Target)
		return attributes

class UnorderedList (NonVoidElement):
	_TagName = 'ul'

class ListItem (NonVoidElement):
	_TagName = 'li'

class Section (NonVoidElement):
	_TagName = 'section'

class Article (NonVoidElement):
	_TagName = 'article'

class Bold (NonVoidElement):
	_TagName = 'b'

class TextArea (NonVoidElement):
	_TagName = 'textarea'

	def __init__(self, *, Name: str, **kwargs):
		NonVoidElement.__init__(self, **kwargs)
		self.Name = Name

	def _Attributes(self, **kwargs):
		attributes = NonVoidElement._Attributes(self, **kwargs)
		attributes['name'] = self.Name
		return attributes

class ProgressBar (NonVoidElement):
	_TagName = 'progress'

	def __init__(self, *, Value: float = None, Max: float = None, **kwargs):
		NonVoidElement.__init__(self, **kwargs)
		self.Value = Value
		self.Max = Max

	def _Attributes(self, **kwargs):
		attributes = NonVoidElement._Attributes(self, **kwargs)
		if self.Max is not None:
			attributes['max'] = str(self.Max)
		if self.Value is not None:
			attributes['value'] = str(self.Value)
		return attributes

class HiddenInput (EmptyElement):
	_TagName = 'input'

	def __init__(self, *, Name: str, Value: str, **kwargs):
		EmptyElement.__init__(self, **kwargs)
		self.Name = Name
		self.Value = Value

	def _Attributes(self, **kwargs):
		attributes = EmptyElement._Attributes(self, **kwargs)
		attributes['name'] = self.Name
		attributes['value'] = self.Value
		attributes['type'] = 'hidden'
		return attributes

class InlineStyle (NonVoidElement):
	_TagName = 'style'
	def __init__(self, Content: str, **kwargs):
		NonVoidElement.__init__(self, **kwargs)
		self.Content = Content

	def ToHTML(self, **kwargs):
		yield from self._OpeningTag(**kwargs)
		yield self.Content  # TODO: escaping
		yield from self._ClosingTag(**kwargs)

class ExternalStyle (NonVoidElement):
	_TagName = 'link'

	def __init__(self, *children, Target, **kwargs):
		NonVoidElement.__init__(self, *children, **kwargs)
		self.Target = Target

	def _Attributes(self, routeRegistry, **kwargs):
		attributes = NonVoidElement._Attributes(self, **kwargs)
		attributes['rel'] = 'stylesheet'
		attributes['href'] = routeRegistry.GetTarget(self.Target)
		return attributes

class FormLabel (NonVoidElement):
	_TagName = 'label'

class Footer (NonVoidElement):
	_TagName = 'footer'

class NavigationBar (NonVoidElement):
	_TagName = 'nav'

class LineBreak (EmptyElement):
	_TagName = 'br'

class Small (NonVoidElement):
	_TagName = 'small'

###############################################################################

class Head (NonVoidElement):
	_TagName = 'head'

class Body (NonVoidElement):
	_TagName = 'body'

class Document:
	def __init__(self):
		self.Head = Head()
		self.Body = Body()

	def Append(self, node):
		self.Body.Append(node)

	def ToHTML(self, *, application, **kwargs):
		# TODO: create a templating engine
		# TODO: collect needed JS programatically
		self.Head.Append(ExternalStyle(Target = application.RouteRegistry._Routes['GetCoreCSS'].Target))
		self.Body.Append(ExternalScript(Target = application.RouteRegistry._Routes['GetCoreJS'].Target))
		self.Body.Append(ExternalScript(Target = application.RouteRegistry._Routes['GetGridJS'].Target))

		yield "<!DOCTYPE html>"
		yield "<html>"
		yield from self.Head.ToHTML(**kwargs)
		yield from self.Body.ToHTML(**kwargs)
		yield "</html>"

###############################################################################

class Field (NonVoidElement):
	_TagName = 'div'

	class FieldControls (NonVoidElement):
		_TagName = 'div'

		def _Attributes(self, **kwargs):
			attributes = NonVoidElement._Attributes(self, **kwargs)
			attributes['class'] = 'Controls'
			return attributes

	class FieldDescription (NonVoidElement):
		_TagName = 'p'

		def _Attributes(self, **kwargs):
			attributes = NonVoidElement._Attributes(self, **kwargs)
			attributes['class'] = 'Description'
			return attributes

	def __init__(self, *children, Label: str, Description: str = None, **kwargs):
		NonVoidElement.__init__(self, *children, **kwargs)
		self.Label = Label
		self.Description = Description

		self._Children = [FormLabel(Text(self.Label), Paragraph(Class = 'Error')), Field.FieldControls(*self._Children)]
		if self.Description is not None:
			self._Children[0].Append(Field.FieldDescription(self.Description))

	def _Attributes(self, **kwargs):
		attributes = NonVoidElement._Attributes(self, **kwargs)
		attributes['class'] = 'Field'  # TODO: what about composing classes?
		return attributes

class Output (NonVoidElement):
	_TagName = 'output'

class Span(NonVoidElement):
	_TagName = 'span'

class Image (EmptyElement):
	_TagName = 'img'

	def __init__(self, *, Target, Width: float = None, Height: float = None, **kwargs):
		EmptyElement.__init__(self, **kwargs)
		self.Target = Target
		self.Width = Width
		self.Height = Height

	def _Attributes(self, *, routeRegistry, **kwargs):
		attributes = EmptyElement._Attributes(self, **kwargs)
		attributes['src'] = routeRegistry.GetTarget(self.Target)
		if self.Width is not None:
			attributes['width'] = str(round(self.Width))
		if self.Height is not None:
			attributes['height'] = str(round(self.Height))
		return attributes

class Div(NonVoidElement):
	_TagName = 'div'