from collections import OrderedDict
import inspect
from Ozark.Typing import TypedList

class InjectorFactory:
	def __init__(self):
		self._ByNameFactories = OrderedDict()
		self._ByTypeFactories = OrderedDict()

	def __call__(self):
		return Injector(
			byNameFactories = self._ByNameFactories,
			byTypeFactories = self._ByTypeFactories)

	def AddByNameArgumentFactory(self, argumentFactory, argumentName: str):
		self._ByNameFactories[argumentName] = argumentFactory

	def AddByTypeArgumentFactory(self, argumentFactory, argumentType: type):
		self._ByTypeFactories[argumentType] = argumentFactory

class Injector:
	def __init__(self, *, byNameFactories: dict, byTypeFactories: dict):
		self._ByNameFactories = byNameFactories
		self._ByTypeFactories = byTypeFactories
		self.Cache = {'injector': self}

	def __setitem__(self, key: str, value):
		assert key not in self.Cache, "{} argument already set"
		self.Cache[key] = value

	def __call__(self, callTarget, **kwargs):
		callArguments = {
			argument.name: self.GetArgument(argument, **kwargs)
			for argument in inspect.signature(callTarget).parameters.values()}

		return callTarget(**callArguments)

	def GetArgument(self, argument, **kwargs):
		argumentName, argumentType, argumentDefault = argument.name, argument.annotation, argument.default
		#if argument.kind != inspect.Parameter.KEYWORD_ONLY:
		#	raise TypeError("can't inject positional argument {}".format(argumentName))
		try:
			return kwargs[argumentName]
		except KeyError:
			pass

		try:
			return self.Cache[argumentName]
		except KeyError:
			pass

		if argumentName in self._ByNameFactories:
			self.Cache[argumentName] = self(
				self._ByNameFactories[argumentName])
			return self.Cache[argumentName]

		if argumentType is inspect.Parameter.empty:
			raise TypeError("unknown argument name '{}'".format(argumentName))

		try:
			isVariadicArgument = issubclass(argumentType, TypedList)
		except TypeError:
			isVariadicArgument = False
		if isVariadicArgument:
			argumentType = argumentType.ElementType

		for factoryType, argumentFactory in self._ByTypeFactories.items():
			try:
				isTypeMatch = issubclass(argumentType, factoryType)
			except TypeError:
				isTypeMatch = isinstance(argumentType, factoryType)
			if isTypeMatch:
				fieldValues = self(
					argumentFactory,
					argumentName = argumentName,
					argumentType = argumentType)
				break
		else:
			raise TypeError("unknown type '{}'".format(argumentType.__name__))

		if not isVariadicArgument:
			if not isinstance(fieldValues, list):
				self.Cache[argumentName] = fieldValues
			elif len(fieldValues) == 1:
				self.Cache[argumentName] = fieldValues.pop()
			elif len(fieldValues) == 0 and argumentDefault is not inspect.Parameter.empty:
				self.Cache[argumentName] = argumentDefault
			else:
				self.Cache[argumentName] = fieldValues
				#raise ValueError("argument '{}' takes exactly one value".format(argumentName))
		else:
			self.Cache[argumentName] = fieldValues
		return self.Cache[argumentName]
