from uuid import uuid4

class Entity:
	"""Entities are the nouns of the model — they are not necessarily
physical entities; a book and a book order are both (related) entities.

``_Thunk``, if supplied, must be a dictionary mapping attribute names to
functions taking no arguments; these functions must return the value to
be assigned to the corresponding attribute. They are evaluated whenever the
corresponding attribute is read, written or deleted.

Any extra keyword arguments are assigned to instance attributes *after*
``_Thunk`` is processed (and therefore any side effects caused by thunk
evaluation are triggered). No attempt is made to restrict attribute names
to those present in the database schema.

This is an abstract base class, and should *not* be instantiated directly;
call the parent anchors instead."""
	def __init__(self, *, _Thunk = None, **kwargs):
		self._Thunk = _Thunk or {}
		self._Value = {}
		for key, value in kwargs.items():
			setattr(self, key, value)
		if not hasattr(self, 'Id'):
			object.__setattr__(self, 'Id', uuid4())

	def __hash__(self):
		return hash(self.Id)

	def __eq__(self, other):
		if self.__class__ != other.__class__:
			return NotImplemented
		return self.Id == other.Id

	def __getattr__(self, key):
		try:
			thunk = self._Thunk.pop(key)
		except KeyError:
			raise AttributeError("{} instance has no attribute {}".format(self.__class__.__name__, key))
		self._Value[key] = value = thunk()
		object.__setattr__(self, key, value)
		return value

	def __setattr__(self, key, value):
		if key.startswith('_'):
			object.__setattr__(self, key, value)
			return
		# Force thunk evaluation
		try:
			getattr(self, key)
		except AttributeError:
			pass
		if key == 'Id':
			raise AttributeError("entity id is immutable")
		if key in self._Anchor._Child:
			value = frozenset(value)
			for child in value:
				parentName = list(
					parentName
					for parentName, parentAnchor in child._Anchor._Parent.items()
					if parentAnchor is self._Anchor)[0]
				if getattr(child, parentName, self) is not self:
					raise AttributeError("can\'t reassign child entity to another parent")
				object.__setattr__(child, parentName, self)
		object.__setattr__(self, key, value)

	def __delattr__(self, key):
		if not key.startswith('_'):
			# Force thunk evaluation
			try:
				getattr(self, key)
			except AttributeError:
				pass
		if key == 'Id':
			raise AttributeError("entity id is immutable")
		object.__delattr__(self, key)

	def __repr__(self):
		return "<{} Id={}>".format(self.__class__.__name__, self.Id)