from collections import defaultdict

class AttributeUsageEstimator (defaultdict):
	"""Aggregate attribute usage data for an entity within a given function.

The entity and function for which this estimator collects data are *not*
stored here — see the implementation of :py:class:`EntityReader` instead."""
	def __init__(self):
		defaultdict.__init__(self, lambda: 0)
		self.InstanceCount = 0

	def __repr__(self):
		return "<AttributeUsageEstimator InstanceCount={} {}>".format(self.InstanceCount, dict.__repr__(self))

	def EntityConstructorDecorator(self):
		"""Decorate an entity constructor so that, whenever it is called,
that entity's instance count is incremented."""
		def Decorator(constructor):
			def DecoratedConstructor(_Thunk):
				self.InstanceCount += 1
				return constructor(_Thunk = {
					attributeName: self._AttributeThunkDecorator(attributeName)(thunk)
					for attributeName, thunk in _Thunk.items()})
			return DecoratedConstructor
		return Decorator

	def _AttributeThunkDecorator(self, attributeName):
		"""Decorate an attribute thunk so that, whenever it is evaluated,
that attribute's evaluation count is incremented."""
		def Decorator(thunk):
			def DecoratedThunk():
				self[attributeName] += 1
				return thunk()
			return DecoratedThunk
		return Decorator

	def EstimateUsage(self, attributeName):
		"""Estimate the probability that the attribute named `attributeName`
will be evaluated at some point during the function's execution."""
		try:
			return self[attributeName] / self.InstanceCount
		except ZeroDivisionError:
			return 0.0  # 0.5