from .Anchor import Anchor
from .Attribute import StaticAttribute

class Model:
	"""The ORM model is the aggregate of all anchors, attributes and ties
between entities; it is inferred from the database schema passed to the
constructor.

The anchors within can be accessed as a dictionary, e.g. `model['Document']`
is the Document anchor."""
	def __init__(self):
		self._Anchor = {}

	@staticmethod
	def CreateFromSchema(schema):
		"""
:param schema: dictionary mapping table names to tables
		"""
		model = Model()
		model._Schema = schema

		anchorTable = []
		attributeTable = []
		for table in schema.values():
			if table.Name in table.Column:
				# delay processing of attributes until we get all the anchors
				attributeTable.append(table)
			elif table.Name + '.Id' in table.Column:
				anchorTable.append(table)
			else:
				# TODO: do something with the remaining unmapped tables
				# print(table.Name)
				pass

		for table in anchorTable:
			model._MapAnchorTable(table)
		for table in attributeTable:
			model._MapAttributeTable(table)

		# Create hierarquical structure to anchors, e.g. "A"."B"."C" == "A.B.C"
		for childName in model._Anchor:
			if '.' not in childName:
				continue
			parentName, childFragment = childName.rsplit('.', 1)
			setattr(model[parentName], childFragment, model[childName])

			try:
				_, parentFragment = parentName.rsplit('.', 1)
			except ValueError:
				parentFragment = parentName
			model[childName]._Parent[parentFragment] = model[parentName]
			model[parentName]._Child[childFragment] =model[childName]

		return model

	def __getitem__(self, key):
		return self._Anchor[key]

	def __setitem__(self, key, value):
		self._Anchor[key] = value

	def __iter__(self):
		return iter(self._Anchor.values())

	def _MapAnchorTable(self, table):
		keyColumn = table.Column[table.Name + '.Id']
		self[table.Name] = Anchor(
			name = table.Name,
			table = table,
			key = keyColumn)

	def _MapAttributeTable(self, table):
		anchorName, attributeName = table.Name.rsplit('.', 1)
		keyColumn = table.Column[anchorName + '.Id']
		valueColumn = table.Column[table.Name]
		# TODO: the information schema tells whether a table is updateable
		isWritable = True

		self[anchorName][attributeName] = StaticAttribute(
			anchor = self[anchorName],
			table = table,
			keyColumn = keyColumn,
			valueColumn = valueColumn,
			isWritable = isWritable)

	def ToPostgreSQL(self):
		# TODO: move this to PostgreSQL compiler
		for anchorName, anchor in sorted(self._Anchor.items()):
			anchorCreateTableStatement = list()
			anchorCreateTableStatement.append('CREATE TABLE "{}" ('.format(anchorName))
			anchorCreateTableStatement.append('\t"{}.Id" UUID NOT NULL,'.format(anchorName))
			for _, parentAnchor in anchor._Parent.items():
				parentName = list(anchorName for anchorName, anchor in self._Anchor.items() if anchor is parentAnchor)[0]
				anchorCreateTableStatement.append('\t"{}.Id" UUID NOT NULL,'.format(parentName))
				anchorCreateTableStatement.append('\tFOREIGN KEY ("{}.Id") REFERENCES "{}" ("{}.Id") ON DELETE CASCADE,'.format(parentName, parentName, parentName))
			anchorCreateTableStatement.append('\tUNIQUE ("{}.Id"))'.format(anchorName))
			yield '\n'.join(anchorCreateTableStatement)
			for attributeName, attribute in anchor._Attribute.items():
				if isinstance(attribute, StaticAttribute):
					yield '''CREATE TABLE "{}.{}" (
	"{}.Id" UUID NOT NULL,
	"{}.{}" {} NOT NULL,
	UNIQUE ("{}.Id"),
	FOREIGN KEY ("{}.Id") REFERENCES "{}" ("{}.Id") ON DELETE CASCADE)'''.format(
						anchorName,
						attributeName,
						anchorName,
						anchorName,
						attributeName,
						attribute._Type,
						anchorName,
						anchorName,
						anchorName,
						anchorName)