from Ozark import DB, QueryAST, ORM
from Ozark.Injector import Injector, InjectorFactory

def ResetSchema(*,
		cursor: DB.Connection.Cursor):
	resetQuery = QueryAST.Query.Raw("DROP SCHEMA public CASCADE; CREATE SCHEMA public")
	resetQuery.Execute(cursor)

def GetPieceByName(*,
		cursor: DB.Connection.Cursor,
		ChessPiece: ORM.Anchor,
		pieceName):
	entityReader = ORM.EntityReader(cursor, None)
	pieceQuery = QueryAST.Query.Select(ChessPiece['Name']._Table).Where(QueryAST.Expression.EqualTo(ChessPiece['Name']._Column, pieceName))
	return entityReader.GetSingle(ChessPiece, pieceQuery)

def ModelFactory(*,
		cursor: DB.Connection.Cursor):
	schema = DB.Database.GetSchema(cursor)
	model = ORM.Model.Model.CreateFromSchema(schema)
	return model

def AnchorFactory(*,
		argumentName,
		model: ORM.Model.Model):
	return model[argumentName]

def RunHierarquicalTest(*,
		injector: Injector):

	def CreateHierarquicalSchema(*,
			cursor: DB.Connection.Cursor):
		ResetSchema(cursor = cursor)
		creationQuery = QueryAST.Query.Raw("""
	CREATE TABLE "ChessPiece" (
		"ChessPiece.Id" UUID NOT NULL,
		UNIQUE ("ChessPiece.Id"));

	CREATE TABLE "ChessPiece.Name" (
		"ChessPiece.Id" UUID NOT NULL,
		"ChessPiece.Name" TEXT NOT NULL,
		UNIQUE ("ChessPiece.Id"),
		FOREIGN KEY ("ChessPiece.Id") REFERENCES "ChessPiece" ("ChessPiece.Id") ON DELETE CASCADE);

	CREATE TABLE "ChessPiece.Move" (
		"ChessPiece.Move.Id" UUID NOT NULL,
		"ChessPiece.Id" UUID NOT NULL,
		UNIQUE ("ChessPiece.Move.Id"),
		FOREIGN KEY ("ChessPiece.Id") REFERENCES "ChessPiece" ("ChessPiece.Id") ON DELETE CASCADE);

	CREATE TABLE "ChessPiece.Move.DeltaX" (
		"ChessPiece.Move.Id" UUID NOT NULL,
		"ChessPiece.Move.DeltaX" INT NOT NULL,
		UNIQUE ("ChessPiece.Move.Id"),
		FOREIGN KEY ("ChessPiece.Move.Id") REFERENCES "ChessPiece.Move" ("ChessPiece.Move.Id") ON DELETE CASCADE);

	CREATE TABLE "ChessPiece.Move.DeltaY" (
		"ChessPiece.Move.Id" UUID NOT NULL,
		"ChessPiece.Move.DeltaY" INT NOT NULL,
		UNIQUE ("ChessPiece.Move.Id"),
		FOREIGN KEY ("ChessPiece.Move.Id") REFERENCES "ChessPiece.Move" ("ChessPiece.Move.Id") ON DELETE CASCADE);
	""")
		creationQuery.Execute(cursor)

	def CreateKnight(*,
			cursor: DB.Connection.Cursor,
			ChessPiece: ORM.Anchor):
		entityWriter = ORM.EntityWriter(cursor, None)
		knight = ChessPiece(Name = 'Knight')
		knight.Move = {
			ChessPiece.Move(DeltaX = deltaX, DeltaY = deltaY)
			for deltaX in (-2, -1, 0, +1, +2)
			for deltaY in (-2, -1, 0, +1, +2)}
		entityWriter.Save(knight)
		entityWriter.Commit()
		knight.Move = {
			move
			for move in knight.Move
			if move.DeltaX != 0 and move.DeltaY != 0 and abs(move.DeltaX) != abs(move.DeltaY)}
		entityWriter.Save(knight)
		entityWriter.Commit()

	def CreateKing(*,
			cursor: DB.Connection.Cursor,
			ChessPiece: ORM.Anchor):
		entityWriter = ORM.EntityWriter(cursor, None)
		king = ChessPiece(Name = 'King')
		king.Move = {
			ChessPiece.Move(DeltaX = DeltaX, DeltaY = DeltaY)
			for DeltaX in (-1, 0, +1)
			for DeltaY in (-1, 0, +1)
			if any(delta != 0 for delta in {DeltaX, DeltaY})}
		entityWriter.Save(king)
		entityWriter.Commit()

	def TestChildEntityDeletion(*,
			cursor: DB.Connection.Cursor,
			ChessPiece: ORM.Anchor):
		entityReader = ORM.EntityReader(cursor, None)
		assert len(entityReader.Get(ChessPiece)) == 1
		assert len(entityReader.Get(ChessPiece.Move)) == 8
		print([move.ChessPiece for move in entityReader.Get(ChessPiece.Move)])

	def DeleteKnight(*,
			cursor: DB.Connection.Cursor,
			injector: Injector):
		entityWriter = ORM.EntityWriter(cursor, None)
		knight = injector(GetPieceByName, pieceName = 'Knight')
		entityWriter.Delete(knight)
		entityWriter.Commit()

	def TestCountPieceMovements(*,
			injector: Injector):
		king = injector(GetPieceByName, pieceName = 'King')
		assert len(king.Move) == 8
		assert set({move.DeltaX ** 2 + move.DeltaY ** 2 for move in king.Move}) == {1, 2}
		knight = injector(GetPieceByName, pieceName = 'Knight')
		assert len(knight.Move) == 8
		assert set({move.DeltaX ** 2 + move.DeltaY ** 2 for move in knight.Move}) == {5}

	injector(CreateHierarquicalSchema)
	injector(CreateKing)
	injector(CreateKnight)
	injector(TestCountPieceMovements)
	injector(DeleteKnight)
	injector(TestChildEntityDeletion)

def Main():
	# TODO: load this from somewhere else
	cursorFactory = DB.Connection.PostgreSQLCursorFactory(
		host = 'localhost',
		port = 5432,
		database = 'oklahoma',
		user = 'fmoreira',
		password = None)

	injectorFactory = InjectorFactory()
	injectorFactory.AddByTypeArgumentFactory(lambda: cursorFactory(echo = True), DB.Connection.Cursor)
	injectorFactory.AddByTypeArgumentFactory(ModelFactory, ORM.Model.Model)
	injectorFactory.AddByTypeArgumentFactory(AnchorFactory, ORM.Anchor)

	injectorFactory()(RunHierarquicalTest)

if __name__ == '__main__':
	Main()