from collections import defaultdict
import functools

from toposort import toposort_flatten

from Ozark import QueryAST
from Ozark.DB.Connection import ResultSet
from Ozark.ORM import Anchor, Attribute
from Ozark.ORM.Attribute import Attribute
from Ozark.ORM.Anchor import Anchor
from Ozark.ORM.Instrumentation import AttributeUsageEstimator
from Ozark.ORM.Entity import Entity
from . import Model

_MapperHintCache = defaultdict(lambda: defaultdict(lambda: AttributeUsageEstimator()))
class EntityReader:
	"""Read-only object-relational mapper.

Initially, the entities returned by this mapper have their attributes
lazy-loaded as they are accessed. On subsequent usage, the access patterns
from previous iterations are used to prefetch the attributes most likely to
be used.

`cacheHint` can be any hashable object. Roughly speaking, mapper instances
passed as arguments to subsequent invocations of the same function should
have tha same `cacheHint` for optimum performance; the function itself is
usually a good choice of `cacheHint`.

:param cursor: database cursor
:param object cacheHint: cache key"""

	def __init__(self, cursor, cacheHint):
		self._Cursor = cursor
		self._CacheHint = _MapperHintCache[cacheHint]
		self._Cache = dict()

	def _GetAttributeThunk(self, attribute, entityQuery):
		def _():
			try:
				return entityQuery.Join(attribute.Query).Execute(self._Cursor).Scalar()
			except ResultSet.NoResultTuples:
				raise AttributeError()
		return _

	def _GetChildThunk(self, childAnchor, entityQuery):
		return functools.partial(self.Get, childAnchor, entityQuery.Join(childAnchor._Table))

	def _GetEntityThunks(self, anchor, prefetchRow):
		entityId = prefetchRow[anchor._Key.Name]
		entityQuery = QueryAST.Query.Select().Where(QueryAST.Expression.EqualTo(anchor._Key, entityId))
		entityThunk = {'Id': lambda entityId = entityId: entityId}
		for attributeName, attribute in anchor._Attribute.items():
			try:
				entityThunk[attributeName] = lambda attributeValue = prefetchRow[attribute._Column.Name]: attributeValue
			except KeyError:
				entityThunk[attributeName] = self._GetAttributeThunk(attribute, entityQuery)

		for childName, childAnchor in anchor._Child.items():
			try:
				entityThunk[childName] = lambda childAnchor = childAnchor, children = prefetchRow[childName]: children
			except KeyError:
				entityThunk[childName] = self._GetChildThunk(childAnchor, entityQuery)

		for parentName, parentAnchor in anchor._Parent.items():
			entityThunk[parentName] = functools.partial(self.GetById, parentAnchor, prefetchRow[parentAnchor._Key.Name])

		return entityThunk

	def Get(self, anchor, entityQuery = None):
		"""Return a list of all entities of type ``anchor`` that satisfy the constraints in the given query.

:returns: the list of satisficing entities"""

		if entityQuery is None:
			entityQuery = QueryAST.Query.Select()
		else:
			entityQuery = entityQuery.Copy()
		# Don't fetch columns unrelated to this entity
		# TODO: make this more robust
		entityQuery.Column = {ordering.Expression.Name: ordering.Expression for ordering in entityQuery.OrderBy}
		mapperHint = self._CacheHint[anchor]
		# We need the anchor key to map the prefetch results
		prefetchQuery = entityQuery.Join(anchor._Table)
		for attributeName, attribute in anchor._Attribute.items():
			if mapperHint.EstimateUsage(attributeName) < 0.05:
				continue
			prefetchQuery = prefetchQuery.LeftOuterJoin(attribute.Query)
		prefetchResult = prefetchQuery.Execute(self._Cursor)
		for row in prefetchResult:
			for k, v in list(row.items()):
				if v is None:
					del row[k]

		for childName, childAnchor in anchor._Child.items():
			if mapperHint.EstimateUsage(childName) < 0.05:
				continue

			childQuery = entityQuery.Join(childAnchor._Table)
			childEntityMap = {row[anchor._Key.Name]: set() for row in prefetchResult}
			for row in childQuery.Execute(self._Cursor):
				childEntityMap[row[anchor._Key.Name]].add(row[childAnchor._Key.Name])
			children = {child.Id: child for child in self.Get(childAnchor, entityQuery.Join(childQuery))}

			for row in prefetchResult:
				entityId = row[anchor._Key.Name]
				row[childName] = frozenset(children[childId] for childId in childEntityMap[entityId])

		for parentName, parentAnchor in anchor._Parent.items():
			# The parent IDs are already fetched in the appropriate columns; the only decision is whether to fetch the
			# entities themselves.
			if mapperHint.EstimateUsage(parentName) < 0.05:
				continue
			# TODO: populate self._Cache with the parent entities without triggering recursion

		entityConstructor = mapperHint.EntityConstructorDecorator()(anchor)
		entityList = []
		for row in prefetchResult:
			entityList.append(entityConstructor(_Thunk = self._GetEntityThunks(anchor, row)))
		for entity in entityList:
			self._Cache[entity.Id] = entity
		return entityList

	def GetSingle(self, anchor, entityQuery):
		"""Return the unique entity of type ``anchor`` satifying the given query.

:returns: the requested entity
:raises IndexError: if there is no satisficing entity
:raises IndexError: if there is more than one satisficing entity"""
		entitySet = self.Get(anchor, entityQuery)
		if len(entitySet) != 1:
			raise IndexError()
		return list(entitySet)[0]

	def GetById(self, anchor, entityId):
		"""Return the unique entity of type ``anchor`` with the given id.

:returns: the requested entity
:raises IndexError: if there is no entity with that id"""
		if entityId in self._Cache:
			return self._Cache[entityId]

		entityQuery = QueryAST.Query.Select(anchor._Table)
		entityQuery = entityQuery.Where(QueryAST.Expression.EqualTo(anchor._Key, entityId))

		try:
			return self.GetSingle(anchor, entityQuery)
		except IndexError:
			raise KeyError()

	def Begin(self):
		"""Start a transaction."""
		self._Cursor.Begin()

	def Rollback(self):
		"""Rollback the current transaction and release all held locks."""
		self._Cursor.Rollback()


# noinspection PyProtectedMember
class EntityWriter (EntityReader):
	def __init__(self, cursor, cacheHint):
		EntityReader.__init__(self, cursor = cursor, cacheHint = cacheHint)
		self.PendingSave = set()
		self.PendingDelete = set()

	def _SaveAnchor(self, anchor, entityList):
		insertQuery = QueryAST.Query.InsertTuple(anchor._Table)
		for entity in entityList:
			if 'Id' not in entity._Value and 'Id' not in entity._Thunk:
				entity._Value['Id'] = entity.Id
				entity._New = True
				insertTuple = {anchor._Key.Name: entity.Id}
				for parentName, parentAnchor in anchor._Parent.items():
					insertTuple[parentAnchor._Key.Name] = getattr(entity, parentName).Id
				insertQuery.Tuple.append(insertTuple)
			else:
				entity._New = False
		yield insertQuery

	def _SaveAttribute(self, anchor, entityList, attributeName, attribute):
		if not attribute._IsWritable:
			return

		deleteIdList = []
		insertQueryList = []
		for entity in entityList:
			# Ignore unevaluated attributes
			if attributeName in entity._Thunk:
				continue

			# Ignore unchanged attributes
			if getattr(entity, attributeName, None) == entity._Value.get(attributeName, None):
				continue

			# Update local object
			try:
				entity._Value[attributeName] = getattr(entity, attributeName)
			except AttributeError:
				del entity._Value[attributeName]

			# Update database
			if not entity._New:
				deleteIdList.append(entity.Id)
			if hasattr(entity, attributeName):
				insertQueryList.append(attribute.InsertQuery(getattr(entity, attributeName)).Where(QueryAST.Expression.EqualTo(anchor._Key, entity.Id)))
		yield attribute.DeleteQuery.Where(QueryAST.Expression.ElementOf(anchor._Key, frozenset(deleteIdList)))
		# TODO: de-dupe this code
		insertQueryTableSet = set(query.Table for query in insertQueryList)
		if all(isinstance(query, QueryAST.Query.InsertTuple) for query in insertQueryList) and \
				len(insertQueryTableSet) == 1:
			insertQueryTable = list(insertQueryTableSet)[0]
			insertTupleList = []
			for query in insertQueryList:
				insertTupleList.extend(query.Tuple)
			yield QueryAST.Query.InsertTuple(insertQueryTable, *insertTupleList)
		else:
			yield from insertQueryList

	def _DeleteAnchor(self, anchor, entityList):
		# TODO: make this compatible with versioned stuff
		yield QueryAST.Query.Delete(anchor._Table).Where(QueryAST.Expression.ElementOf(anchor._Key, frozenset(entity.Id for entity in entityList)))

	def Save(self, entity):
		"""Schedule the entity for later insertion or update.

A subsequent call to :py:meth:`Delete` cancels the insertion or update.
Conversely, calling :py:meth:`Save` on an entity scheduled for deletion
cancels that operation.

This will also recursively insert or update any tied entities.

This method is idempotent — if the entity has already been scheduled for
insertion or update, :py:meth:`Save` returns immediately."""

		if entity in self.PendingSave:
			return
		if entity in self.PendingDelete:
			self.PendingDelete.remove(entity)
		self.PendingSave.add(entity)

		anchor = entity._Anchor

		# Recursively save all children
		for childName, childAnchor in anchor._Child.items():
			if childName in entity._Thunk:  # Skip untouched entities:
				return
			setattr(entity, childName, frozenset(getattr(entity, childName, ())))
			entity._Value[childName] = frozenset(entity._Value.get(childName, ()))
			for childEntity in entity._Value[childName] - getattr(entity, childName):
				self.Delete(childEntity)
			for childEntity in getattr(entity, childName):
				self.Save(childEntity)
			entity._Value[childName] = getattr(entity, childName)

	def Delete(self, entity):
		"""Schedule the entity for later deletion.

A subsequent call to :py:meth:`Save` cancels the deletion. Conversely,
calling :py:meth:`Delete` on an entity scheduled for insertion or update
cancels that operation.

This method is idempotent — if the entity has already been schedule for
deletion, :py:meth:`Save` returns immediately."""

		if entity in self.PendingDelete:
			return
		if entity in self.PendingSave:
			self.PendingSave.remove(entity)
		self.PendingDelete.add(entity)

		# No need to delete children — the database will take care of the cascading.

	def Commit(self):
		"""Commit all pending changes to the database."""

		pendingQueries = []

		pendingSave = defaultdict(list)
		for entity in self.PendingSave:
			pendingSave[entity._Anchor].append(entity)
		topologicalOrder = toposort_flatten({anchor: set(anchor._Parent.values()) for anchor in pendingSave.keys()}, sort = False)

		# Insert all anchors before attributes and ties
		for anchor, entityList in sorted(pendingSave.items(), key = lambda t: topologicalOrder.index(t[0])):
			pendingQueries.extend(self._SaveAnchor(anchor, entityList))
		for anchor, entityList in pendingSave.items():
			for attributeName, attribute in anchor._Attribute.items():
				pendingQueries.extend(self._SaveAttribute(anchor, entityList, attributeName, attribute))

		pendingDelete = defaultdict(list)
		for entity in self.PendingDelete:
			pendingDelete[entity._Anchor].append(entity)
		# Delete only anchors; ties and attributes will be cascaded.
		for anchor, entityList in pendingDelete.items():
			pendingQueries.extend(self._DeleteAnchor(anchor, entityList))

		compositeQuery = QueryAST.Query.CompositeQuery(*pendingQueries)._Simplify()
		if len(compositeQuery.Queries) > 0:
			compositeQuery.Execute(self._Cursor)
			# TODO: invalidate thunks left lying around
		self._Cursor.Commit()

		self.PendingSave = set()
		self.PendingDelete = set()

	def Rollback(self):
		"""Rollback the current transaction, release all held locks and
cancel all insertions, updates and deletions."""

		EntityReader.Rollback(self)
		self.PendingSave = set()
		self.PendingDelete = set()