class Expression:
	def _Simplify(self):
		return self

	@staticmethod
	def Simplify(obj):
		if isinstance(obj, Expression):
			return obj._Simplify()
		return obj

	# Boolean operators
	def __and__(self, other):
		return Conjunction(self, other)

	def __rand__(self, other):
		return Conjunction(other, self)

	def __or__(self, other):
		return Disjunction(self, other)

	def __ror__(self, other):
		return Disjunction(other, self)

	def __xor__(self, other):
		return NotEqualTo(self, other)

	def __rxor__(self, other):
		return NotEqualTo(other, self)

	# TODO: implement __invert__

	# Arithmetic operators
	def __add__(self, other):
		return Addition(self, other)

	def __radd__(self, other):
		return Addition(other, self)

	def __sub__(self, other):
		return Subtraction(self, other)

	def __rsub__(self, other):
		return Subtraction(other, self)

	def __mul__(self, other):
		return Multiplication(self, other)

	def __rmul__(self, other):
		return Multiplication(other, self)

	def __truediv__(self, other):
		return Division(self, other)

	def __rtruediv__(self, other):
		return Division(other, self)

class Column (Expression):
	def __init__(self, Name):
		self.Name = Name

	def __eq__(self, other):
		if self.__class__ != other.__class__:
			return NotImplemented
		return self.Name == other.Name

	def __hash__(self):
		return hash(('Column', self.Name))

class BinaryOperation (Expression):
	def __init__(self, LeftTerm, RightTerm):
		self.LeftTerm = LeftTerm
		self.RightTerm = RightTerm

	@property
	def Terms(self):
		return [self.LeftTerm, self.RightTerm]

	def _Simplify(self):
		self.LeftTerm = Expression.Simplify(self.LeftTerm)
		self.RightTerm = Expression.Simplify(self.RightTerm)
		return self

# TODO: what if less than two terms are received?
class SemigroupOperation (Expression):
	"""An operation that is associative."""
	def __init__(self, *Terms):
		self.Terms = Terms

	def _Simplify(self):
		self.Terms = [Expression.Simplify(expression) for expression in self.Terms]
		for term in self.Terms[:]:
			if isinstance(term, self.__class__):
				# Flatten tree
				index = self.Terms.index(term)
				self.Terms = self.Terms[:index] + term.Terms[:] + self.Terms[index+1:]
		return self

class MonoidOperation (SemigroupOperation):
	Identity = NotImplemented

	def _Simplify(self):
		simplifiedExpression = SemigroupOperation._Simplify(self)
		if not isinstance(simplifiedExpression, self.__class__):
			return simplifiedExpression

		self.Terms = [expression for expression in self.Terms if expression != self.__class__.Identity]
		if len(self.Terms) == 0:
			return self.__class__.Identity
		if len(self.Terms) == 1:
			return self.Terms[0]
		return self

class LoopRightDivisionOperation (BinaryOperation):
	"""Represents an operation with a right inverse loop (e.g. subtraction, division)."""
	RightInverse = NotImplemented

	def _Simplify(self):
		simplifiedExpression = BinaryOperation._Simplify(self)
		if not isinstance(simplifiedExpression, self.__class__):
			return simplifiedExpression

		if self.LeftTerm == self.RightTerm:
			return self.__class__.RightInverse.Identity
		if self.RightTerm == self.__class__.RightInverse.Identity:
			return self.LeftTerm
		if isinstance(self.RightTerm, self.__class__):
			return self.__class__(self.__class__.RightInverse(self.LeftTerm, self.RightTerm.RightTerm), self.RightTerm.LeftTerm)
		return self

class Addition (MonoidOperation, Expression):
	Identity = 0

class Multiplication (MonoidOperation, Expression):
	Identity = 1

class Subtraction (LoopRightDivisionOperation, Expression):
	RightInverse = Addition

class Division (LoopRightDivisionOperation, Expression):
	RightInverse = Multiplication

class Maximum (MonoidOperation, Expression):
	Identity = float('-inf')

class Minimum (MonoidOperation, Expression):
	Identity = float('+inf')

class Boole (Expression):
	def __init__(self, Expression):
		self.Expression = Expression

	def _Simplify(self):
		self.Expression = Expression.Simplify(self.Expression)
		return self

class Disjunction (MonoidOperation, Expression):
	Identity = False

class Conjunction (MonoidOperation, Expression):
	Identity = True

# TODO: convert these names to nouns
class ElementOf (BinaryOperation, Expression):
	def _Simplify(self):
		if len(self.RightTerm) == 0:
			return False
		return self

class NotElementOf (BinaryOperation, Expression):
	def _Simplify(self):
		if len(self.RightTerm) == 0:
			return True
		return self

class EqualTo (BinaryOperation, Expression):
	pass

class NotEqualTo (BinaryOperation, Expression):
	pass

class LessThan (BinaryOperation, Expression):
	pass

class LessThanOrEqualTo (BinaryOperation, Expression):
	pass

class GreaterThan (BinaryOperation, Expression):
	pass

class GreaterThanOrEqualTo (BinaryOperation, Expression):
	pass

class StartsWith (BinaryOperation, Expression):
	pass

class FunctionCall (Expression):
	pass

class Generic (FunctionCall):
	def __init__(self, Name, *Arguments):
		self.Name = Name
		self.Arguments = Arguments

	# TODO: remove this when this gets brought to the Expression hierarchy
	def Simplify(self):
		return self

class CurrentTimestamp (FunctionCall):
	pass

class AggregateFunctionCall (FunctionCall):
	pass

class AggregateMaximum (FunctionCall):
	def __init__(self, Expression):
		self.Expression = Expression

	def _Simplify(self):
		self.Expression = Expression.Simplify(self.Expression)
		return self