from .Expression import Expression
import Ozark.QueryAST

class Query:
	def _Simplify(self):
		return self

	def Execute(self, cursor):
		return cursor.Execute(self._Simplify())

class Raw (Query):
	def __init__(self, statement, **parameters):
		self._Statement = statement
		self._Parameters = parameters

class Select (Query):
	class JoinTypes:
		Inner = b'INNER'
		LeftOuter = b'LEFT OUTER'

	def __init__(self, *Table, Constraint = True):
		self.Table = [(t, Select.JoinTypes.Inner) for t in Table]
		self.Column = {}
		for t in Table:
			self.Column.update({columnName: Ozark.QueryAST.Expression.Column(columnName) for columnName, column in t.Column.items()})
		self.Constraint = Constraint
		# TODO: restrict usage of this when joining
		self.GroupBy = []
		self.OrderBy = []

	def __eq__(self, other):
		if not isinstance(other, self.__class__):
			return NotImplemented
		return self.Column == other.Column and self.Table == other.Table and self.Constraint == other.Constraint

	def Copy(self):
		query = Select()
		query.Column = self.Column.copy()
		query.Table = self.Table
		query.Constraint = self.Constraint

		query.GroupBy = self.GroupBy[:]
		query.OrderBy = self.OrderBy[:]
		return query

	def Join(self, obj):
		"""Compute the natural inner join of two queries."""
		query = self.Copy()
		if isinstance(obj, Ozark.QueryAST.Table.Table):
			query._AddTable(obj, join = Select.JoinTypes.Inner)
			# noinspection PyUnresolvedReferences
			query.Column.update({columnName: Ozark.QueryAST.Expression.Column(columnName) for columnName, column in obj.Column.items()})

		elif isinstance(obj, Select):
			# TODO: audit this function to make sure it is sound.
			query.Column.update(obj.Column)
			for otherTable, join in obj.Table:
				query._AddTable(otherTable, join = join)
			query = query.Where(obj.Constraint)
		return query

	def LeftOuterJoin(self, obj):
		"""Compute the natural left outer join of two queries."""
		# TODO: audit this function to make sure it is sound.
		joinedQuery = self.Copy()
		joinedQuery.Column.update(obj.Column)
		for otherTable, join in obj.Table:
			# TODO: is this always correct?
			joinedQuery._AddTable(otherTable, join = Select.JoinTypes.LeftOuter)
		joinedQuery.Constraint &= obj.Constraint
		return joinedQuery

	def _AddTable(self, table, *, join = JoinTypes.Inner):
		# TODO: make INNER JOIN override LEFT OUTER JOIN
		if table not in (t for t, j in self.Table):
			self.Table = self.Table + [(table, join)]

	def Where(self, constraint):
		query = self.Copy()
		query.Constraint = query.Constraint & constraint
		return query

	def _Simplify(self):
		self.Constraint = Ozark.QueryAST.Expression.Expression.Simplify(self.Constraint)
		for columnName, column in self.Column.items():
			self.Column[columnName] = Ozark.QueryAST.Expression.Expression.Simplify(column)
		for table, join in self.Table:
			# if isinstance(table, QueryTable):
			if hasattr(table, 'Query'):
				table.Query = table.Query._Simplify()
		return self

class Delete (Query):
	def __init__(self, Table, *, Constraint = True):
		self.Table = Table
		self.Constraint = Constraint

	def _Simplify(self):
		self.Constraint = Expression.Simplify(self.Constraint)
		return self

	def Where(self, constraint):
		return Delete(self.Table, Constraint = self.Constraint & constraint)

class Insert (Query):
	pass

class InsertTuple (Insert):
	def __init__(self, Table, *Tuple):
		self.Table = Table
		self.Tuple = list(Tuple)

	def Where(self, constraint):
		# TODO: is this semantically correct?
		if isinstance(constraint, Ozark.QueryAST.Expression.EqualTo):
			if isinstance(constraint.LeftTerm, Ozark.QueryAST.Expression.Column):
				attributeName = constraint.LeftTerm.Name
				attributeValue = constraint.RightTerm
			elif isinstance(constraint.RightTerm, Ozark.QueryAST.Expression.Column):
				attributeName = constraint.RightTerm.Name
				attributeValue = constraint.LeftTerm
			else:
				# TODO: fallback to an InsertSelect
				raise NotImplementedError()

			# TODO: check that attributeValue can be computed without accessing other columns
			newTupleList = [t.copy() for t in self.Tuple]
			for newTuple in newTupleList:
				newTuple.update({attributeName: attributeValue})
			return InsertTuple(self.Table, *newTupleList)
		# TODO: fall back to an InsertSelect
		raise NotImplementedError()

class InsertSelect (Insert):
	def __init__(self, Table, Query):
		self.Table = Table
		self.Query = Query

	def _Simplify(self):
		self.Query = self.Query._Simplify()
		return self

	def Where(self, constraint):
		return InsertSelect(self.Table, self.Query.Where(constraint))

class CompositeQuery (Query):
	def __init__(self, *Queries):
		self.Queries = Queries

	def _Simplify(self):
		self.Queries = [query._Simplify() for query in self.Queries]
		self.Queries = [query for query in self.Queries if not isinstance(query, InsertTuple) or len(query.Tuple) > 0]
		self.Queries = [query for query in self.Queries if not isinstance(query, Delete) or query.Constraint != False]
		return self

	def Where(self, constraint):
		return CompositeQuery(*[query.Where(constraint) for query in self.Queries])