import re
import functools
import urllib.parse

class RouteMismatchError (Exception):
	pass

class RouteTarget:
	"""Abstract class; means a route target when specified to an injector"""
	def __init__(self):
		raise NotImplementedError()

class Route:
	PathArgumentRegex = re.compile(r'{([a-zA-Z]+)}')

	def __init__(self, target, pattern, *, method = 'GET'):
		"""
		:parameter str pattern:
		:parameter ? target:
		:parameter None | str method:
		"""
		self.Pattern = pattern
		self.Target = target
		self.Method = method

		self._PathRegex = re.compile(r'^' + Route.PathArgumentRegex.sub(r'(?P<\1>[^/]+)', self.Pattern) + r'$')

	def GetPathFields(self, request):
		"""
		:parameter Request request:
		:return list[(str, str)]:
		:raises RouteMismatchError:
		"""
		if request.Method != self.Method:
			raise RouteMismatchError("methods '{0}' and '{1}' are incompatible".format(request.Method, self.Method))

		match = self._PathRegex.match(request.Path)
		if match is None:
			raise RouteMismatchError("path '{0}' does not match pattern '{1}'".format(request.Path, self.Pattern))
		return list(match.groupdict().items())

class RouteRegistry:
	def __init__(self, *, RouteInjectorFactory):
		self._Routes = {}
		self._RouteInjectorFactory = RouteInjectorFactory

	def AddRoute(self, routeFactory, pattern, *, method = 'GET'):
		routeInjector = self._RouteInjectorFactory()
		routeTarget = routeInjector(routeFactory)
		# TODO: remove duplicate view names to allow this assertion to be enabled
		# assert routeFactory.__name__ not in self._Routes.keys()
		self._Routes[routeFactory.__name__] = Route(routeTarget, pattern, method = method)

	def GetRoute(self, target) -> Route:
		validRoutes = [route for route in self._Routes.values() if route.Target is target]
		if len(validRoutes) == 0:
			raise KeyError("no route registered for target {}".format(target.__name__))
		elif len(validRoutes) > 1:
			# FIXME: shouldn't happen, block this at route registration
			assert False
		return validRoutes[0]

	def GetTargetAndMethod(self, target):
		pathArguments = {}
		if isinstance(target, functools.partial):
			target, pathArguments = target.func, target.keywords
		route = self.GetRoute(target)

		target, method = route.Pattern, route.Method
		for match in Route.PathArgumentRegex.finditer(route.Pattern):
			placeholder = match.group(0)
			argumentName = match.group(1)
			try:
				target = target.replace(placeholder, str(pathArguments.pop(argumentName)))
			except KeyError:
				raise TypeError("{} missing 1 required path argument: {}".format(route.Target.__name__, argumentName))
		if pathArguments:
			if method == 'GET':
				target += '?' + urllib.parse.urlencode(pathArguments)
			else:
				# TODO: return remaining parameters so they can be inserted as hidden fields
				raise NotImplementedError()
		return target, method

	def GetTarget(self, target, *, method = 'GET'):
		"""Get the target URL for the given target objects.

		Throws if the method supplied does not correspond to the target."""
		actualTarget, actualMethod = self.GetTargetAndMethod(target)
		if method != actualMethod:
			raise ValueError("methods '{0}' and '{1}' are incompatible".format(method, actualMethod))
		return actualTarget

class PrefixedRouteRegistry:
	def __init__(self, parent, prefix):
		self.Parent = parent
		self.Prefix = prefix

	def AddRoute(self, routeFactory, pattern, *, method = 'GET'):
		self.Parent.AddRoute(routeFactory, self.Prefix + pattern, method = method)

	def GetRoute(self, target) -> Route:
		return self.Parent.GetRoute(target)