import bcrypt

from lawrence.base import EntityManager
from User.Entity.User import User
from User.Entity.UserRole import UserRole
from lawrence.db import DBRow


class UserManager (EntityManager):
	__entity__ = User

	def GetById(self, userId):
		if userId is None:
			return None

		user = User(**self.DatabaseCursor.Single('''
			SELECT User.Id AS Id,
				   User.FullName,
				   User.email,
				   User.username,
					User.FullName AS name,
				   (User.password_hash IS NOT NULL) AS has_password
			FROM User
			WHERE User.Id = %s''',
			(userId, )))

		user.id = user.Id
		user.path = '/user/%s/' % user.Id

		user.Permissions = self.DatabaseCursor.Execute(
			'''SELECT path, permission
			   FROM UserPermission
			   WHERE UserId = %s''',
			(user.Id, ))
		user.Permissions.append(DBRow(path = '/user/%s/' % user.Id, permission = 'write'))
		for row in self.DatabaseCursor.Execute('''
			SELECT CampusId
			FROM UserCampus
			WHERE UserId = %s''', (user.Id, )):
			if row.CampusId is None:
				user.Permissions.append(DBRow(path = '/campus/', permission = 'read'))
			else:
				user.Permissions.append(DBRow(path = '/campus/%s/' % row.CampusId, permission = 'read'))
		return user

	def GetAll(self):
		return self.DatabaseCursor.Execute('''
			SELECT
				User.id AS id,
				User.id AS Id,
				User.FullName
			FROM
				User
			ORDER BY FullName''')

	def IsSwitchingAllowed(self, targetUser):
		if self.CurrentUser is None:
			return False
		return self.CurrentUser.HasPermission('/', 'admin')

	def GetByUsernameAndPassword(self, username, password):
		try:
			user = self.DatabaseCursor.Single('''
				SELECT
					password_hash AS PasswordHash,
					id AS Id
				FROM
					User
				WHERE
					username = %s''',
				(username, ))
		except ValueError:
			raise ValueError('invalid username')

		if user.PasswordHash is None:
			raise ValueError('invalid password')
		if bcrypt.hashpw(password.encode('utf-8'), user.PasswordHash.encode('utf-8')) != user.PasswordHash.encode('utf-8'):
			raise ValueError('invalid password')

		return user

	def GetPicture(self, userId):
		user = self.DatabaseCursor.Single('''
			SELECT
				User.Picture AS Picture
			FROM
				User
			WHERE
				User.id = %s''',
			(userId, ))
		if user.Picture is None:
			raise ValueError('user has no picture')
		return user.Picture

	def SendInvite(self, user):
		import lawrence.mail
		import lawrence.config
		import random

		user.InviteCode = "".join([random.choice("ABCDEFGHIJKLMNOPQRSTUVWXYZ") for _ in range(12)])
		self.DatabaseCursor.Begin()
		self.DatabaseCursor.Execute('''
			UPDATE
				User
			SET
				invite_code = %s
			WHERE
				id = %s''',
			(user.InviteCode, user.Id))
		lawrence.mail.Queue('nao-responda@colegioapogeu.com.br',
							user.email,
							("""From: nao-responda@colegioapogeu.com.br
To: %s
Subject: Senha de acesso

Olá %s,

Crie agora a sua senha de acesso no sistema Admissão Apogeu através
do link abaixo:

http://%s/invite/%s/

	Atenciosamente,

		%s.""" % (user.email,
				  user.FullName,
				  lawrence.config.domain,
				  user.InviteCode,
				  lawrence.config.FullName)).encode('utf-8'))
		self.DatabaseCursor.Commit()

	def Create(self, newUser):
		self.DatabaseCursor.Begin()
		self.DatabaseCursor.Execute('''
			INSERT INTO User (FullName, email)
			VALUES (%s, %s)''',
			(newUser.FullName, newUser.email))
		newUser.Id = self.DatabaseCursor.ScopeIdentity()
		self.SendInvite(newUser)
		self.DatabaseCursor.Commit()
		return newUser

	def Delete(self, user):
		self.DatabaseCursor.Begin()
		self.DatabaseCursor.Execute('''
			DELETE FROM User
			WHERE Id = %s''',
			(user.Id, ))
		self.DatabaseCursor.Commit()

	def SetRole(self, user, role, grant):
		assert isinstance(user, User)
		assert isinstance(role, UserRole)
		assert isinstance(grant, bool)

		self.DatabaseCursor.Begin()
		self.DatabaseCursor.Execute('''
			DELETE FROM UserRoleMap
			WHERE
				UserRoleMap.UserId = %s AND
				UserRoleMap.RoleId = %s''',
			(user.Id, role.Id))
		if grant:
			self.DatabaseCursor.Execute('''
				INSERT INTO UserRoleMap (UserId, RoleId)
				VALUES (%s, %s)''',
				(user.Id, role.Id))
		self.DatabaseCursor.Commit()

	def AddRole(self, role, permission):
		self.SetRole(role, permission, True)

	def RemoveRole(self, role, permission):
		self.SetRole(role, permission, False)
