class Entity (object):
    """Generic model container.

    >>> Entity(id = 7, name = u"Abacaxi")
    <Entity {'id': 7, 'name': u'Abacaxi'}>
    >>> Entity(id = 7, class_ = u"Abacaxi")
    <Entity {'class': u'Abacaxi', 'id': 7}>
    >>> Entity(**{'id': 7, 'class': u"Abacaxi"})
    <Entity {'class': u'Abacaxi', 'id': 7}>
    >>> Entity(**{'id': 7, 'class.id': 12, 'class.name': u"Abacaxi"})
    <Entity {'class': <Entity {'id': 12, 'name': u'Abacaxi'}>, 'id': 7}>
    >>> Entity(id = 7, class__id = 12, class__name = u"Abacaxi")
    <Entity {'class': <Entity {'id': 12, 'name': u'Abacaxi'}>, 'id': 7}>
    >>> Entity(**{'id': 7, 'class.id': 12, 'class.name': u"Abacaxi", 'class.campus.id': 19, 'class.campus.name': u"Tomate"})
    <Entity {'class': <Entity {'campus': <Entity {'id': 19, 'name': u'Tomate'}>, 'id': 12, 'name': u'Abacaxi'}>, 'id': 7}>
    >>> m = Entity(id = 7, name = u"Abacaxi")
    >>> m.id = 14
    >>> m
    <Entity {'id': 14, 'name': u'Abacaxi'}>
    >>> m.class_ = Entity(id = 28, name = u"Maçã")
    >>> m
    <Entity {'class': <Entity {'id': 28, 'name': u'Ma\\xc3\\xa7\\xc3\\xa3'}>, 'id': 14, 'name': u'Abacaxi'}>
    """

    def __init__(self, **kwargs):
        object.__setattr__(self, '_data', {})
        kwargs = {Entity.NormalizeKey(key): value for key, value in list(kwargs.items())}
        while kwargs:
            key = list(kwargs.keys())[0]
            if '.' not in key:
                self._data[key] = kwargs[key]
                del kwargs[key]
                continue
            head, _ = key.split('.', 1)
            tails = [key.split('.', 1)[1] for key in list(kwargs.keys())
                     if '.' in key and key.split('.', 1)[0] == head]
            self._data[head] = Entity(**{tail: kwargs[head + '.' + tail] for tail in tails})
            for key in (head + '.' + tail for tail in tails):
                del kwargs[key]

    def __getattr__(self, key):
        assert isinstance(key, str)
        key = Entity.NormalizeKey(key)
        try:
            return object.__getattribute__(self, '_data')[key]
        except KeyError:
            raise AttributeError(key)

    def __setattr__(self, key, value):
        assert isinstance(key, str)
        if key.startswith('_'):
            object.__setattr__(self, key, value)
            return
        key = Entity.NormalizeKey(key)
        self._data[key] = value

    def __repr__(self):
        return '<%s {%s}>' % (self.__class__.__name__,
                              ', '.join(['%s: %s' % (repr(k), repr(v))
                                         for k, v
                                         in sorted(self._data.items())]))

    def __contains__(self, key):
        assert isinstance(key, str)
        return key in self._data

    @classmethod
    def NormalizeKey(cls, key):
        """Transform a key to a canonical form.

        This allows using e.g. class_ instead of class as a keyword."""
        assert isinstance(key, str)
        if key.endswith('_'):
            key = key[:-1]
        key = key.replace('__', '.')
        if isinstance(key, bytes):
            key = key.decode('utf-8')
        return key
