#!/usr/bin/env python2.7
# encoding: utf-8

import copy
import random

from lawrence.Monte.XmlUtil import *
from lawrence.Monte.WordUtil import *

class Chunk (object):
    def __init__(self, monte):
        self.monte = monte
        self.Content = []
        self._AlreadyParsed = False

    def Parse(self):
        """Try to parse this chunk. This method is idempotent whenever it
        fails; after it succeeds, all subsequent calls are also idempotent."""
        return False

    def ExamContent(self):
        return []

    def AnswerContent(self):
        return []

class SubjectHeader (Chunk):
    def __init__(self, monte, subject):
        Chunk.__init__(self, monte)

        self.SubjectName = subject

    def Header(self):
        return [Paragraph(self.SubjectName or '???', style = self.monte.SubjectStyle)]

    def ExamContent(self):
        return self.Header()

    def AnswerContent(self):
        return self.Header()

class NoiseText (Chunk):
    pass

class SharedText (Chunk):
    def ExamContent(self):
        return self.Content

class AnswerKey (Chunk):
    def Parse(self):
        RemoveAnswerKeyHeader(self.Content[0])
        if IsEmpty(self.Content[0]):
            self.Content.pop(0)

class Question (Chunk):
    def __init__(self, monte, number):
        Chunk.__init__(self, monte)

        self.number = number
        self.numId = self.monte.AllocateChoiceNumId()

        self.Statement = []
        self.Choices = []
        self.AnswerStrength = []
        self._answer = None
        self.tail = []

    def Shuffle(self):
        if not self.Choices:
            return
        permutation = list(range(len(self.Choices)))
        random.shuffle(permutation)
        self.Choices = [self.Choices[_] for _ in permutation]
        self._answer = [permutation.index(_) for _ in self.answer]

    def MightBeChoice(self, el):
        if HasChoiceHeader(el):
            return True
        if el.find(ns('./w:pPr/w:numPr/w:numId')) is not None:
            return self.monte.NumberingStyle(el) in ('lowerLetter', 'upperLetter')
        return False

    def ConvertToChoice(self, el):
        self.AnswerStrength.append(CleanTreeAggressively(el))
        if HasChoiceHeader(el):
            RemoveChoiceHeader(el)
            el = StripTag(el, 'w:br')

        SetProp(el, 'w:pPr/w:numPr/w:numId[w:val]', self.numId)
        SetProp(el, 'w:pPr/w:numPr/w:ilvl[w:val]', '0')
        SetProp(el, 'w:pPr/w:pStyle[w:val]', GetProp(self.monte.ChoiceStyle, 'w:styleId'))
        self.Choices.append(el)

    def TryParseChoices(self):
        if len(self.Content) < 4:
            return False

        numChoices = 0
        for i in range(len(self.Content) - 1, -1, -1):
            if not self.MightBeChoice(self.Content[i]):
                break
            numChoices += 1
        if numChoices not in (4, 5):
            return False
        # It's rare to find 5 choices when there are only 4, but the converse
        # is a lot more common. We usually try 5 choices first, but we try
        # some extra heuristics here to avoid accidents.
        if numChoices == 4:
            # Sometimes there's a line break between two choices that we
            # accidentally leave in. Try removing *one* of them.
            newContent = ReplaceLastLineBreak(self.Content)
            if newContent:
                oldContent = self.Content
                self.Content = newContent
                if self.TryParseChoices():
                    return True
                self.Content = oldContent
            # Sometimes a paragraph breaks the next-to-last choice apart. Join
            # a single paragraph if it would make one more choice available.
            if len(self.tail) >= 2 and self.MightBeChoice(self.tail[1]):
                pass  # FIXME
        # Maybe alternatives are incorrectly ordered; this might happen if
        # they were originally placed one per line.
        self.Statement = self.Content[:-numChoices]
        choices = self.Content[-numChoices:]
        if all(HasChoiceHeader(el) for el in choices):
            choiceMap = [(HasChoiceHeader(el), el) for el in choices]
            choiceMap.sort()
            choices = [choice for _, choice in choiceMap]
        for choice in choices:
            self.ConvertToChoice(choice)
        self.Content = []
        return True

    def TryParseTail(self):
        """Search upwards for consecutive paragraphs that look like choices."""
        while self.Content:
            if self.TryParseChoices():
                return True
            self.tail.insert(0, self.Content.pop())
        self.Content, self.tail = self.tail, []
        return False

    def ParseChoices(self):
        oldContent = copy.deepcopy(self.Content)

        if self.TryParseTail():
            self._AlreadyParsed = True
            return True

        # Replace all line breaks by paragraphs and try again.
        while True:
            newContent = ReplaceLastLineBreak(self.Content)
            if not newContent:
                break
            self.Content = newContent
            if self.TryParseTail():
                self._AlreadyParsed = True
                return True

        # Replace all tabs by paragraphs and try again.
        newContent = []
        for paragraph in self.Content:
            remainder = paragraph
            while True:
                tabs = remainder.findall(ns('.//w:tab'))
                if not tabs:
                    if not IsEmpty(remainder):
                        newContent.append(remainder)
                    break
                tab = tabs[0]
                piece, remainder = SplitElement(remainder, tab)
                CleanTree(piece)
                CleanTree(remainder)
                if not IsEmpty(piece):
                    newContent.append(piece)
        self.Content = newContent
        if self.TryParseTail():
            self._AlreadyParsed = True
            return True

        # Now try the reverse approach: try to join paragraphs to create full
        # choices.
        candidateGroups = []
        for paragraph in self.Content:
            if self.MightBeChoice(paragraph):
                candidateGroups.append([])
            if candidateGroups:
                candidateGroups[-1].append(paragraph)
        if len(candidateGroups) >= 4:
            candidateGroups = candidateGroups[-5:]
            self.Statement = self.Content[:-sum(len(group) for group in candidateGroups)]
            for group in candidateGroups:
                if group is not candidateGroups[-1]:
                    pieces = group
                else:
                    # The last group is an exception: it might contain part of
                    # a shared text.
                    for i, piece in enumerate(group):
                        if i > 0 and HasSharedTextHeader(piece):
                            pieces = group[:i]
                            self.tail = group[i:]
                            break
                    else:
                        pieces = group
                paragraph = Element('w:p')
                # Arbitrarily use the pPr of the first paragraph for the union
                # (if it exists).
                paragraphProperties = pieces[0].find(ns('.//w:pPr'))
                if paragraphProperties is None:
                    paragraphProperties = Element('w:pPr')
                paragraph.append(paragraphProperties)
                for piece in pieces:
                    for bit in piece:
                        if bit.tag != ns('w:pPr'):
                            paragraph.append(bit)
                    if piece is not pieces[-1]:
                        run = Element('w:r')
                        br = Element('w:br')
                        run.append(br)
                        paragraph.append(run)
                self.ConvertToChoice(paragraph)
                # Line breaks usually interact badly with justification;
                # let's fix it just in case.
                SetProp(paragraph, 'w:pPr/w:jc[w:val]', 'left')
            self._AlreadyParsed = True
            return True

        # FIXME: try more ideas
        self.Content = oldContent
        return False

    def ParseAnswer(self):
        ANSWER_FOOTER = re.compile(r"\s*[[({]\s*(\d*)\s*linhas?\s*[])}]\s*$", re.IGNORECASE)

        modified = False
        pos = 0
        while pos < len(self.Content):
            paragraph = self.Content[pos]
            pos += 1
            answerFooter = ANSWER_FOOTER.search(ElementText(paragraph))
            if not answerFooter:
                continue

            modified = True
            self._AlreadyParsed = True
            SliceText(paragraph, end = -len(answerFooter.group(0)))
            if IsEmpty(paragraph):
                self.Content.remove(paragraph)
                pos -= 1

            lineCount = int(answerFooter.group(1))
            for _ in range(lineCount):
                blankLine = Paragraph("\t", style = self.monte.AnswerStyle)
                self.Content.insert(pos, blankLine)
            pos += lineCount

        if modified:
            self.Statement = self.Content

        return modified

    def Parse(self):
        if self._AlreadyParsed:
            return True
        if not self.Content:
            return True
        RemoveQuestionHeader(self.Content[0])
        paragraphProperties = self.Content[0].find(ns('.//w:pPr'))
        if paragraphProperties is not None:
            numPr = paragraphProperties.find(ns('w:numPr'))
            if numPr is not None:
                paragraphProperties.remove(numPr)
        CleanTree(self.Content[0])
        if IsEmpty(self.Content[0]):
            self.Content.pop(0)

        if not self.ParseAnswer():
            if not self.ParseChoices():
                return False

        if self.monte.CitationStyle:
            for el in self.Statement:
                if CITATION_RE.match(ElementText(el)):
                    SetProp(el, 'w:pPr/w:pStyle[w:val]', GetProp(self.monte.CitationStyle, 'w:styleId'))
        return True

    def Header(self):
        return [Paragraph("\u00A0Questão %d\u00A0" % (self.number, ), style = self.monte.QuestionStyle)]

    def ExamContent(self):
        if not self._AlreadyParsed:
            self.Parse()
        if not self._AlreadyParsed:
            return self.Header() + self.Content
        else:
            return self.Header() + self.Statement + self.Choices + self.tail

    def _getAnswer(self):
        if self._answer is None:
            self._answer = [_ for _ in range(len(self.Choices)) if self.AnswerStrength[_] == max(self.AnswerStrength)]
            self._answer = [_ for _ in self._answer if _ < len(self.Choices)]
            if len(self._answer) == len(self.Choices):
                self._answer = []
        return self._answer
    def _setAnswer(self, answer):
        self._answer = answer
        if self.Choices:
            self._answer = [_ for _ in self._answer if _ < len(self.Choices)]
    answer = property(_getAnswer, _setAnswer)

    def AnswerContent(self):
        if len(self.answer) == 0:
            answerText = "Sem gabarito."
        elif len(self.answer) == 1:
            answerText = "Letra %s." % ("ABCDE"[self.answer[0]])
        else:
            answerText = "Letras %s." % ("/".join(["ABCDE"[_] for _ in sorted(self.answer)]), )
        return self.Header() + [Paragraph(answerText)]
