#!/usr/bin/env python2.7
# encoding: utf-8

import urllib.request, urllib.parse, urllib.error
from zipfile import ZipFile
from lawrence.Monte.Util import ZipContents

import lawrence.util
from lawrence.Monte.Chunk import *

class ImportedDocument (object):
    def __init__(self, monte, source, filename = None):
        self.Monte = monte
        self.ZipFile = ZipFile(source, 'r')
        self.Filename = filename

        self.DocumentTree = None
        self.NumberingTree = None
        self.RelationshipTree = None
        for name, xmlFile in ZipContents(self.ZipFile):
            if name == 'word/document.xml':
                self.DocumentTree = ElementTree.parse(xmlFile)
            if name == 'word/numbering.xml':
                self.NumberingTree = ElementTree.parse(xmlFile)
            if name == 'word/_rels/document.xml.rels':
                self.RelationshipTree = ElementTree.parse(xmlFile)
        self.ImportRelationships()
        self.ImportNumbering()
        self.ImportDocument()

    def ImportRelationships(self):
        idMap = {}
        for el in self.RelationshipTree.findall(ns('.//rel:Relationship')):
            if GetProp(el, 'Type') not in (PREFIX_MAP['r'] + '/image',
                                            PREFIX_MAP['r'] + '/oleObject',
                                            PREFIX_MAP['r'] + '/hyperlink'):
                continue  # FIXME
            target = GetProp(el, 'Target')
            if GetProp(el, 'Type') in (PREFIX_MAP['r'] + '/image', PREFIX_MAP['r'] + '/oleObject'):
                # Embed remote resources
                if target.startswith('http://') or target.startswith('https://'):
                    extension = target.split('.')[-1]
                    target = self.Monte.ImportMedia(lawrence.util.Slug(target) + '.' + extension, urllib.request.urlopen(target).read())
                else:
                    try:
                        targetData = self.ZipFile.read('word/' + target)
                    except KeyError:
                        continue
                    target = self.Monte.ImportMedia(target, targetData)
            newId = self.Monte.AddRelationship(relationshipType = GetProp(el, 'Type'), target = target)
            idMap[GetProp(el, 'Id')] = newId

        for node in self.DocumentTree.findall('.//*'):
            for attr in node.attrib:
                try:
                    node.attrib[attr] = idMap[node.attrib[attr]]
                except KeyError:
                    pass

    def ImportNumbering(self):
        if self.NumberingTree is None:
            return

        idMap = {}
        for el in self.NumberingTree.findall(ns('.//w:abstractNum')):
            oldId = GetProp(el, '[w:abstractNumId]')
            newId = self.Monte.AddAbstractNum(el)
            idMap[oldId] = newId
        for ref in self.NumberingTree.findall(ns('.//w:abstractNumId')):
            try:
                newId = idMap[GetProp(ref, 'w:val')]
            except KeyError:
                continue
            SetProp(ref, 'w:val', newId)

        idMap = {}
        for el in self.NumberingTree.findall(ns('.//w:num')):
            oldId = GetProp(el, '[w:numId]')
            abstractId = GetProp(el, 'w:abstractNumId[w:val]')
            newId = self.Monte.AddNum(abstractId)
            idMap[oldId] = newId
        for ref in self.DocumentTree.findall(ns('.//w:numPr/w:numId')):
            try:
                newId = idMap[GetProp(ref, 'w:val')]
            except KeyError:
                continue
            SetProp(ref, 'w:val', newId)

    def ImportDocument(self):
        chunks = [NoiseText(self.Monte)]

        content = []
        for el in self.DocumentTree.find(ns('w:body')):
            PreserveWhitespace(el)
            InlineDrawings(el)
            StripText(el)
            CleanTree(el)
            JoinRuns(el)
            CleanTree(el)
            StripTag(el, 'w:br')

            if IsEmpty(el):
                continue
            if el.tag == ns('w:sectPr'):
                continue
            content.append(el)
            # FIXME: do something clever about drawings here.

        if len([_ for _ in content if HasTightQuestionHeader(_)]) >= 3:
            hasQuestionHeader = HasTightQuestionHeader
        else:
            hasQuestionHeader = HasLooseQuestionHeader

        documentQuestions = 0
        answerKeyStyle = 'unknown'
        for el in content:
            if not chunks[-1].Content:
                lastElement = None
            else:
                lastElement = chunks[-1].Content[-1]

            if hasQuestionHeader(el) and answerKeyStyle == 'appendix':
                answerKey = AnswerKey(self.Monte)
                chunks.append(answerKey)
            # If this paragraph has a question header, assume it is a question
            # *unless* it is preceded by a paragraph that is not a question by
            # itself.
            elif hasQuestionHeader(el) and \
                hasQuestionHeader(el)[1] == documentQuestions + 1:
                documentQuestions += 1
                self.Monte.Questions += 1
                question = Question(self.Monte, number = self.Monte.Questions)
                chunks.append(question)
            elif HasAnswerKeyHeader(el):
                if answerKeyStyle == 'unknown':
                    questionsSeen = len([chunk for chunk in chunks
                                          if isinstance(chunk, Question)])
                    if questionsSeen == 1:
                        answerKeyStyle = 'interleaved'
                    else:
                        answerKeyStyle = 'appendix'
                answerKey = AnswerKey(self.Monte)
                chunks.append(answerKey)
            elif HasSharedTextHeader(el) and not isinstance(chunks[-1], Question):
                sharedText = SharedText(self.Monte)
                chunks.append(sharedText)
            elif LooksLikeNoiseText(el):
                noiseText = NoiseText(self.Monte)
                chunks.append(noiseText)
            elif self.Monte.NumberingStyle(el) in ('decimal', 'decimalZero') and \
                 ((isinstance(chunks[-1], (NoiseText, SharedText)) and
                  self.Monte.NumberingStyle(lastElement) is None) or
                  chunks[-1].Parse()):
                documentQuestions += 1
                self.Monte.Questions += 1
                question = Question(self.Monte, number = self.Monte.Questions)
                chunks.append(question)
            chunks[-1].Content.append(el)

        i = 0
        while i < len(chunks):
            chunk = chunks[i]
            if isinstance(chunk, Question):
                if chunk.Parse() and chunk.tail:
                    sharedText = SharedText(self.Monte)
                    sharedText.Content = chunk.tail
                    chunk.tail = []
                    chunks.insert(i+1, sharedText)
            else:
                chunk.Parse()
            i += 1

        # It's very unusual for the last chunk of an input file to be a shared text; it's more likely to be an answer key missing a
        # header.
        if len(chunks) > 0 and isinstance(chunks[-1], SharedText) and \
           all((LOOSE_ANSWER_KEY if self.Monte.NumberingStyle(el) in ('decimal', 'decimalZero') else TIGHT_ANSWER_KEY).match(ElementText(el)) for el in chunks[-1].Content):
            answerKey = AnswerKey(self.Monte)
            answerKey.Content = chunks[-1].Content
            chunks[-1] = answerKey
        questionChunks = [_ for _ in chunks if isinstance(_, Question)]
        answerKeyChunks = [_ for _ in chunks if isinstance(_, AnswerKey)]
        answers = []
        for chunk in answerKeyChunks:
            for el in chunk.Content:
                text = ElementText(el)
                if self.Monte.NumberingStyle(el) in ('decimal', 'decimalZero'):
                    for match in LOOSE_ANSWER_KEY.finditer(text):
                        answers.append(match.group('answer'))
                else:
                    for match in TIGHT_ANSWER_KEY.finditer(text):
                        answers.append(match.group('answer'))
        if len(questionChunks) == len(answers):
            for chunk, answer in zip(questionChunks, answers):
                chunk.answer = [ord(answer.lower()) - ord('a')]

        subjectHints = []
        if self.Filename:
            subjectHints.append(self.Filename)
        for chunk in chunks:
            if isinstance(chunk, NoiseText):
                for el in chunk.Content:
                    subjectHints.append(ElementText(el))
        guessedSubject = GuessSubject(*subjectHints)
        chunks.insert(0, SubjectHeader(self.Monte, guessedSubject))
        self.Monte.Chunks.extend(chunks)
