#!/usr/bin/env python2.7
# encoding: utf-8

import copy

from lawrence.Monte.XmlUtil import *

_TIGHT_QUESTION_PATTERN = r"""(?:
    (\d+)ª\s+quest[ãÃ]o:? |
    quest[ãÃ]o\s+(\d+) |
    question\s+(\d+)
)"""
_LOOSE_QUESTION_PATTERN = r"""(?:
    (\d+)ª\s+quest[ãÃ]o:? |
    quest[ãÃ]o\s+(\d+) |
    question\s+(\d+) |
    (\d+)\s*[°ºª]?\s*[).–—-]
)"""
_TIGHT_QUESTION_HEADER = re.compile(
    r"^\s*" + _TIGHT_QUESTION_PATTERN + r"\s*",
    re.IGNORECASE | re.VERBOSE)
_LOOSE_QUESTION_HEADER = re.compile(
    r"^\s*" + _LOOSE_QUESTION_PATTERN + r"\s*",
    re.IGNORECASE | re.VERBOSE)
_ANSWER_KEY_HEADER = re.compile(
    r"^\s*(?:gabaritos?|gab\.?):?\s*",
    re.IGNORECASE)
_CHOICE_HEADER = re.compile(r"^\s*\(?\s*([a-e])\s*[.)]\s*", re.IGNORECASE)
_SHARED_TEXT_HEADER = re.compile(
    r"""(?:
              ^\s*texto\s*: |
              ^\s*texto\s+(?:\d+|[ivxlcdm]+) |
              (?:texto|figura|mapa)\s+(?:a\s+seguir|abaixo) |
              (?:text|picture|cartoon|comic\s+strip)\s+(?:below|bellow)
         )\s*""",
    re.IGNORECASE | re.VERBOSE)
_NOISE_TEXT_FILTER = re.compile(
    r"""(?:
             ^\s*teste |
             ^\s*quest[õÕ]es
         )\s*""",
    re.IGNORECASE | re.VERBOSE)
LOOSE_ANSWER_KEY = re.compile(
    r"\s*" + _LOOSE_QUESTION_PATTERN +
    r"?\s*(?:letra|op[çÇ][ãÃ]o)?\s*[([]?(?P<answer>[a-e])[])]?\s*",
    re.IGNORECASE | re.VERBOSE)
TIGHT_ANSWER_KEY = re.compile(
    r"\s*" + _LOOSE_QUESTION_PATTERN +
    r"\s*(?:letra|op[çÇ][ãÃ]o)?\s*[([]?(?P<answer>[a-e])[])]?\s*",
    re.IGNORECASE | re.VERBOSE)
CITATION_RE = re.compile(
r"""
    \(?\s*(?: fonte | extra[íÍ]do\s+de | adaptado\s+de )\s*:?
""", re.IGNORECASE | re.VERBOSE)

def HasTightQuestionHeader(el):
    if not el:
        return False
    match = _TIGHT_QUESTION_HEADER.match(ElementText(el))
    if match:
        return True, int([_ for _ in match.groups() if _][0])
    else:
        return False

def HasLooseQuestionHeader(el):
    if not el:
        return False
    match = _LOOSE_QUESTION_HEADER.match(ElementText(el))
    if match:
        return True, int([_ for _ in match.groups() if _][0])
    else:
        return False

def HasAnswerKeyHeader(el):
    return bool(_ANSWER_KEY_HEADER.match(ElementText(el)))

def HasChoiceHeader(el):
    match = _CHOICE_HEADER.match(ElementText(el))
    if match:
        return match.group(1).lower()
    else:
        return False

def HasSharedTextHeader(el):
    return bool(_SHARED_TEXT_HEADER.search(ElementText(el)))

def LooksLikeNoiseText(el):
    return bool(_NOISE_TEXT_FILTER.match(ElementText(el)))

def RemoveHeader(el, header):
    SliceText(el, start = len(header))

def RemoveQuestionHeader(el):
    header = _LOOSE_QUESTION_HEADER.match(ElementText(el))
    if header:
        RemoveHeader(el, header.group(0))

def RemoveAnswerKeyHeader(el):
    header = _ANSWER_KEY_HEADER.match(ElementText(el))
    if header:
        RemoveHeader(el, header.group(0))

def RemoveChoiceHeader(el):
    header = _CHOICE_HEADER.match(ElementText(el))
    if header:
        RemoveHeader(el, header.group(0))

def IsEmpty(el):
    if el.tag != ns('w:p'):
        return False
    if el.findall(ns('.//w:drawing')) or el.findall(ns('.//w:object')):
        return False
    if ElementText(el).strip():
        return False
    return True

def CleanAttributes(el, xpath):
    def _ParseXPath(xpath):
        tag, attr = xpath.split('[')
        tag = ns(tag)
        attr = attr[:-1]
        val = None
        if '=' in attr:
            attr, val = attr.split('=')
        attr = ns(attr)
        return tag, attr, val
    tag, attr, val = _ParseXPath(xpath)
    for child in el:
        CleanAttributes(child, xpath)
    if tag not in ('', '*') and el.tag != tag:
        return
    if val and el.attrib.get(attr, None) != val:
        return
    try:
        del el.attrib[attr]
    except KeyError:
        pass

def CleanEmptyElements(el, tagName, blacklist = None):
    if not blacklist:
        blacklist = []
    blacklist = [ns(_) for _ in blacklist]
    retval = False
    for child in el.getchildren()[:]:
        if child.tag == ns(tagName) and \
           not child.getchildren() and \
           not child.attrib and \
           not child.text:
            el.remove(child)
            retval = True
        elif child.tag not in blacklist:
            retval |= CleanEmptyElements(child, tagName, blacklist)
    return retval

def RemoveRevisions(el):
    def _RemoveRevisions(el):
        for attr in ('w:rsidDel',
                     'w:rsidP',
                     'w:rsidR',
                     'w:rsidRDefault',
                     'w:rsidRPr',
                     'w:rsidSect',
                     'w:rsidTr'):
            try:
                del el.attrib[ns(attr)]
            except KeyError:
                pass
    TreeRecurse(el, _RemoveRevisions)

def CleanTree(el):
    RemoveRevisions(el)

    def _CleanFakeSubscripts(el):
        if el.tag != ns('w:r'):
            return
        fontSize = GetProp(el, 'w:rPr/w:sz[w:val]')
        if fontSize is None:
            return
        fontSize = int(fontSize)
        if fontSize <= 14:
            SetProp(el, 'w:rPr/w:vertAlign[w:val]', 'subscript')
    TreeRecurse(el, _CleanFakeSubscripts)

    BLACKLIST = {
        ns('w:p'): ['w:proofErr', 'w:bookmarkStart', 'w:bookmarkEnd'],
        ns('w:pPr'): ['w:pStyle', 'w:spacing', 'w:jc', 'w:ind', 'w:rPr',
                      'w:widowControl', 'w:adjustRightInd', 'w:adjustLeftInd',
                      'w:autoSpaceDE', 'w:autoSpaceDN', 'w:shd', 'w:tabs',
                      'w:keepLines', 'w:keepNext', 'w:contextualSpacing'],
        ns('w:rPr'): ['w:sz', 'w:iCs', 'w:bdr', 'w:smallCaps',
                      'w:snapToGrid', 'w:bCs', 'w:noProof', 'w:shd',
                      'w:rStyle', 'w:lang', 'w:szCs'],
        ns('w:tblPr'): ['w:tblInd'],
        ns('w:smartTag'): ['w:smartTagPr'],
    }
    for tag in BLACKLIST:
        BLACKLIST[tag] = set([ns(_) for _ in BLACKLIST[tag]])
    def _CleanTree(el):
        if el.tag in BLACKLIST:
            for child in el[:]:
                if child.tag in BLACKLIST[el.tag]:
                    el.remove(child)
    TreeRecurse(el, _CleanTree)

    def _CleanText(s):
        s = s.replace("\u2113", "l")
        s = s.replace("\uFB00", "ff")
        s = s.replace("\uFB01", "fi")
        s = s.replace("\uFB02", "fl")
        s = s.replace("\uFB03", "ffi")
        s = s.replace("\uFB04", "ffl")
        s = s.replace("\uFB06", "st")
        return s

    def _CleanCharacters(el):
        if el.text is not None:
            el.text = _CleanText(el.text)
        if el.tail is not None:
            el.tail = _CleanText(el.tail)
    TreeRecurse(el, _CleanCharacters)

    def _UnwrapTree(el):
        for child in el[:]:
            if child.tag == ns('w:smartTag'):
                for r in child.findall(ns('.//w:r')):
                    el.insert(el[:].index(child), r)
                el.remove(child)
    TreeRecurse(el, _UnwrapTree)

    CleanEmptyElements(el, 'w:t')

    CleanAttributes(el, 'w:color[w:val=000000]')
    CleanEmptyElements(el, 'w:color')
    CleanAttributes(el, 'w:highlight[w:val=FFFFFF]')
    CleanEmptyElements(el, 'w:highlight')
    for node in el.findall(ns('.//w:rFonts')):
        for attr in list(node.attrib.keys()):
            if node.attrib[attr] not in ('Symbol', 'MT Extra'):
                del node.attrib[attr]
    CleanEmptyElements(el, 'w:rFonts')
    CleanEmptyElements(el, 'w:rPr')
    CleanEmptyElements(el, 'w:r')

    CleanAttributes(el, 'w:tab[w:pos]')
    CleanAttributes(el, 'w:tab[w:val]')
    CleanEmptyElements(el, 'w:tab', ('w:r', ))
    CleanEmptyElements(el, 'w:tabs')
    CleanEmptyElements(el, 'w:pPr')
    CleanEmptyElements(el, 'w:p', ('w:tc', ))

def CleanTreeAggressively(el):
    """Clean colors and font styles that CleanTree leaves behind.

    Returns a rough count of how many modifications were made."""
    CleanTree(el)
    retval = 0
    CleanAttributes(el, 'w:color[w:val]')
    CleanAttributes(el, 'w:highlight[w:val]')
    CleanEmptyElements(el, 'w:tab')
    retval += CleanEmptyElements(el, 'w:color')
    retval += CleanEmptyElements(el, 'w:highlight')
    CleanTree(el)
    return retval

def PreserveWhitespace(el):
    def _PreserveWhitespace(el):
        if el.tag == ns('w:t'):
            el.attrib[ns('xml:space')] = 'preserve'
    TreeRecurse(el, _PreserveWhitespace)

def JoinRuns(el):
    def _IsPlainRun(el):
        return (el.tag == ns('w:r') and
                len(el) == 1 and
                el[0].tag == ns('w:t'))
    def _JoinRuns(el):
        if el.tag != ns('w:p'):
            return
        i = 1
        while i < len(el):
            leftChild, rightChild = el[i-1], el[i]
            if (not _IsPlainRun(leftChild) or
                not _IsPlainRun(rightChild)):
                i += 1
            else:
                leftChild[0].text += rightChild[0].text
                el.remove(rightChild)
        return False
    TreeRecurse(el, _JoinRuns)

def InlineDrawings(el):
    def _InlineDrawings(el):
        if el.tag != ns('w:drawing'):
            return
        if el.find(ns('wp:inline')) is None:
            anchor = el.find(ns('wp:anchor'))
            assert anchor is not None
            for tag in ('wp:positionH',
                        'wp:positionV',
                        'wp:simplePos',
                        'wp:wrapNone',
                        'wp:wrapSquare',
                        'wp:wrapThrough',
                        'wp:wrapTight',
                        'wp:wrapTopAndBottom'):
                child = anchor.find(ns(tag))
                if child is not None:
                    anchor.remove(child)
            inline = Element('wp:inline')
            inline[:] = anchor[:]
            el[:] = [inline]
        return False
    TreeRecurse(el, _InlineDrawings)

def SplitElement(el, barrier):
    before = Element(el.tag)
    after = Element(el.tag)
    before.attrib = el.attrib.copy()
    after.attrib = el.attrib.copy()
    if el.tag == ns('w:p'):
        pPr = el.find(ns('w:pPr'))
        if pPr:
            before.append(copy.deepcopy(pPr))
            after.append(copy.deepcopy(pPr))
    if el.tag == ns('w:r'):
        rPr = el.find(ns('w:rPr'))
        if rPr:
            before.append(copy.deepcopy(rPr))
            after.append(copy.deepcopy(rPr))
    current = before
    current.text = el.text
    for e in el:
        if e.tag in (ns('w:pPr'), ns('w:rPr')):
            continue
        if e is barrier:
            current = after
            continue
        pieces = SplitElement(e, barrier)
        if len(pieces) > 1:
            current.append(pieces[0])
            assert current is before
            current = after
            current.append(pieces[1])
        else:
            current.append(pieces[0])
    current.tail = el.tail
    if current is before:
        return (before, )
    else:
        return (before, after)

def StripTag(el, tag):
    remainder = el
    while True:
        tags = remainder.findall('.//' + ns(tag))
        if not tags:
            break
        head, tail = SplitElement(remainder, tags[0])
        if IsEmpty(head):
            remainder = tail
            CleanTree(remainder)
        else:
            break
    while True:
        tags = remainder.findall('.//' + ns(tag))
        if not tags:
            break
        head, tail = SplitElement(remainder, tags[-1])
        if IsEmpty(tail):
            remainder = head
            CleanTree(remainder)
        else:
            break
    return remainder

def Paragraph(text = "", style = None):
    paragraph = Element('w:p')
    if style:
        SetProp(paragraph, 'w:pPr/w:pStyle[w:val]', style.attrib[ns('w:styleId')])
    for i, piece in enumerate(text.split("\t")):
        if i > 0:
            run = Element('w:r')
            run.append(Element('w:tab'))
            paragraph.append(run)
        if piece:
            run = Element('w:r')
            text = Element('w:t')
            text.text = piece
            run.append(text)
            paragraph.append(run)
    return paragraph

def PageBreak():
    paragraph = Element('w:p')
    run = Element('w:r')
    br = Element('w:br')
    br.attrib[ns('w:type')] = "page"
    run.append(br)
    paragraph.append(run)
    return paragraph

def ReplaceLastLineBreak(content):
    newContent = []
    replaced = False
    for paragraph in content[::-1]:
        if replaced:
            newContent.insert(0, paragraph)
            continue
        remainder = paragraph
        replacement = []
        while True:
            lineBreaks = remainder.findall(ns('.//w:br'))
            if not lineBreaks:
                if not IsEmpty(remainder):
                    replacement.append(remainder)
                break
            replaced = True
            lineBreak = lineBreaks[0]
            piece, remainder = SplitElement(remainder, lineBreak)
            CleanTree(piece)
            CleanTree(remainder)
            if not IsEmpty(piece):
                replacement.append(piece)
        newContent = replacement + newContent
    if not replaced:
        return
    return newContent
