#!/usr/bin/env python2.7
# encoding: utf-8

import re
from io import StringIO
from elementtree import ElementTree

PREFIX_MAP = {
    'a': 'http://schemas.openxmlformats.org/drawingml/2006/main',
    'ct': 'http://schemas.openxmlformats.org/package/2006/content-types',
    'm': 'http://schemas.openxmlformats.org/officeDocument/2006/math',
    'mc': 'http://schemas.openxmlformats.org/markup-compatibility/2006',
    'o': 'urn:schemas-microsoft-com:office:office',
    'pic': 'http://schemas.openxmlformats.org/drawingml/2006/picture',
    'r': 'http://schemas.openxmlformats.org/officeDocument/2006/relationships',
    'rel': 'http://schemas.openxmlformats.org/package/2006/relationships',
    'v': 'urn:schemas-microsoft-com:vml',
    'w': 'http://schemas.openxmlformats.org/wordprocessingml/2006/main',
    'w10': 'urn:schemas-microsoft-com:office:word',
    'w14': 'http://schemas.microsoft.com/office/word/2010/wordml',
    'wne': 'http://schemas.microsoft.com/office/word/2006/wordml',
    'wp': 'http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing',
    'wp14': 'http://schemas.microsoft.com/office/word/2010/wordprocessingDrawing',
    'wpc': 'http://schemas.microsoft.com/office/word/2010/wordprocessingCanvas',
    'wpg': 'http://schemas.microsoft.com/office/word/2010/wordprocessingGroup',
    'wpi': 'http://schemas.microsoft.com/office/word/2010/wordprocessingInk',
    'wps': 'http://schemas.microsoft.com/office/word/2010/wordprocessingShape',
    'xml': 'http://www.w3.org/XML/1998/namespace',
}

for prefix, uri in list(PREFIX_MAP.items()):
    ElementTree._namespace_map[uri] = prefix

_SORTED_PREFIX_ITEMS = sorted(list(PREFIX_MAP.items()), key = lambda k_v: len(k_v[0]))
_ns_cache = {}
def ns(ref):
    if ref in _ns_cache:
        return _ns_cache[ref]
    key = ref
    for prefix, uri in _SORTED_PREFIX_ITEMS:
        ref = re.sub(r'\b' + prefix + ':', '{' + uri + '}', ref)
    _ns_cache[key] = ref
    return ref

def Element(tag):
    return ElementTree.Element(ns(tag))

def GetProp(el, ref):
    try:
        tags, attr = ref.split('[')
        attr = attr[:-1]
        tag = el.find('./' + ns(tags))
        if tag is None:
            return None
    except ValueError:
        tag = el
        attr = ref
    return tag.attrib[ns(attr)]

def SetProp(el, ref, val):
    try:
        tags, attr = ref.split('[')
        attr = attr[:-1]
        tagParts = [ns(part) for part in tags.split('/')]
    except ValueError:
        tags, attr = '', ref
        tagParts = []
    parent = el
    for part in tagParts:
        child = parent.find('./' + part)
        if child is None:
            child = Element(part)
            parent.insert(0, child)
        parent = child
    parent.attrib[ns(attr)] = val

def SearchProp(el, ref, val):
    tags, attr = ref.split('[')
    attr = attr[:-1]
    for e in el.findall('./' + ns(tags)):
        if e.attrib[ns(attr)] == val:
            return e

def TextNodes(el):
    nodes = []
    def _TextNodes(el):
        if el.tag == ns('w:t'):
            nodes.append(el)
            return False
    TreeRecurse(el, _TextNodes)
    return nodes

def ElementText(el):
    return "".join([node.text for node in TextNodes(el)])

def SliceText(el, start = None, end = None):
    nodes = TextNodes(el)
    length = sum(len(node.text) for node in nodes)

    if start is None:
        start = 0
    elif start < 0:
        start += length
    start = max(0, min(length, start))

    if end is None:
        end = length
    elif end < 0:
        end += length
    end = max(0, min(length, end))

    if start > end:
        start = end

    for node in nodes:
        consumedCharacters = min(start, len(node.text))
        end -= consumedCharacters
        length -= consumedCharacters
        start -= consumedCharacters
        node.text = node.text[consumedCharacters:]
        if start == 0:
            break
    assert start == 0

    for node in nodes[::-1]:
        consumedCharacters = min(length - end, len(node.text))
        length -= consumedCharacters
        node.text = node.text[:len(node.text) - consumedCharacters]
        if end == length:
            break
    assert end == length

def ReplaceText(el, oldText, newText):
    def ReplaceTextOnce(el, oldText, newText):
        start = ElementText(el).find(oldText)
        if start < 0:
            return False

        nodes = TextNodes(el)
        remainingChars = len(oldText)

        state = 'start'
        for node in nodes:
            if state == 'start':
                consumedCharacters = min(start, len(node.text))
                start -= consumedCharacters
                if start == 0:
                    if len(node.text) - consumedCharacters >= remainingChars:
                        node.text = node.text[:consumedCharacters] + newText + node.text[consumedCharacters + remainingChars:]
                        return True
                    else:
                        remainingChars -= len(node.text) - consumedCharacters
                        node.text = node.text[:consumedCharacters] + newText
                        state = 'end'
                        continue
            elif state == 'end':
                consumedCharacters = min(remainingChars, len(node.text))
                remainingChars -= consumedCharacters
                node.text = node.text[consumedCharacters:]
                if remainingChars == 0:
                    return True
        assert False

    while ReplaceTextOnce(el, oldText, newText):
        pass

def StripText(el):
    text = ElementText(el)
    leadingSpaces = len(text) - len(text.lstrip())
    trailingSpaces = len(text) - len(text.rstrip())
    SliceText(el, start = leadingSpaces, end = len(text) - trailingSpaces)

def TreeRecurse(el, before = None, after = None):
    if before and before(el) is False:
        return
    for child in el:
        TreeRecurse(child, before, after)
    if after:
        after(el)

def SetPrefixes(elem):
    def FixElementPrefixes(elem, uriMap, memo):
        def fixup(name):
            try:
                return memo[name]
            except KeyError:
                if name[0] != '{':
                    return
                uri, tag = name[1:].split('}')
                if uri in uriMap:
                    newName = uriMap[uri] + ':' + tag
                    memo[name] = newName
                    return newName
        # fix element name
        name = fixup(elem.tag)
        if name:
            elem.tag = name
        # fix attribute names
        for key, value in list(elem.items()):
            name = fixup(key)
            if name:
                elem.set(name, value)
                del elem.attrib[key]
    # check if this is a tree wrapper
    if not ElementTree.iselement(elem):
        elem = elem.getroot()

    # build uri map and add to root element
    uriMap = {}
    for prefix, uri in list(PREFIX_MAP.items()):
        if prefix == 'xml':
            continue  # FIXME
        uriMap[uri] = prefix
        elem.set('xmlns:' + prefix, uri)

    # fixup all elements in the tree
    memo = {}
    for elem in elem.getiterator():
        FixElementPrefixes(elem, uriMap, memo)

def DumpTree(tree):
    SetPrefixes(tree)
    buf = StringIO()
    tree.write(buf)
    return buf.getvalue()
