import datetime as _datetime
import decimal
Decimal = decimal.Decimal
import re

def time(value: str):
	if value is None:
		raise ValueError
	match = re.match(r"^(\d{1,2})h(?:(\d{2})m?)?$", value)
	if match:
		hour, minute = int(match.group(1)), int(match.group(2) or '00')
		if not 0 <= hour < 24 or not 0 <= minute < 60:
			raise ValueError(value)
		return _datetime.time(hour, minute)
	match = re.match(r"^(\d{1,2}):(\d{2})(?::\d{2})?$", value)
	if match:
		hour, minute = int(match.group(1)), int(match.group(2))
		if not 0 <= hour < 24 or not 0 <= minute < 60:
			raise ValueError(value)
		return _datetime.time(hour, minute)
	raise ValueError(value)

def date(value: str):
	if value is None:
		raise ValueError
	match = re.match(r"(\d{1,2})[/-](\d{1,2})[/-](\d{1,4})", value)
	if match:
		day, month, year = int(match.group(1)), \
						   int(match.group(2)), \
						   int(match.group(3))
		try:
			if year < 100:
				if year < 30:
					year += 2000
				else:
					year += 1900
			return _datetime.date(year, month, day)
		except:
			raise ValueError(value)
	MONTHS = {
		'jan': 1,
		'fev': 2,
		'feb': 2,
		'mar': 3,
		'abr': 4,
		'apr': 4,
		'mai': 5,
		'may': 5,
		'jun': 6,
		'jul': 7,
		'ago': 8,
		'aug': 8,
		'set': 9,
		'sep': 9,
		'out': 10,
		'oct': 10,
		'nov': 11,
		'dez': 12,
		'dec': 12,
	}
	match = re.match(r"(\d{1,2})[/-](" + "|".join(list(MONTHS.keys())) +
					 r")[/-](\d{1,4})", value)
	if match:
		day, month, year = int(match.group(1)), \
						   MONTHS[match.group(2)], \
						   int(match.group(3))
		try:
			return _datetime.date(year, month, day)
		except:
			raise ValueError(value)
	match = re.match(r"(\d{1,4})[/-](\d{1,2})[/-](\d{1,2})", value)
	if match:
		day, month, year = int(match.group(3)), \
						   int(match.group(2)), \
						   int(match.group(1))
		try:
			return _datetime.date(year, month, day)
		except:
			raise ValueError
	raise ValueError

def datetime(value: str):
	valueDate, valueTime = value.split()
	return _datetime.datetime.combine(date(valueDate), time(valueTime))

def fraction(value: str):
	match = re.match(r"\+?(-?\d+(?:[.,]\d+)?)%", value)
	if match:
		try:
			return Decimal(match.group(1).replace(",", ".")) / Decimal(100)
		except decimal.InvalidOperation:
			raise ValueError
	if "/" in value:
		try:
			numerator, denominator = value.split("/")
			return Decimal(numerator) / Decimal(denominator)
		except decimal.InvalidOperation:
			raise ValueError
	try:
		return Decimal(value.replace(",", "."))
	except decimal.InvalidOperation:
		raise ValueError

def percentage(value: str):
	value = value.rstrip('%')
	return fraction(value) / Decimal(100)

def money(value: str):
	match = re.match(r"(?:R?\$)?[  ]*([0-9.]+)(?:,(\d\d))?\s*", value)
	if not match:
		raise ValueError

	dollars, cents = match.group(1), match.group(2) or "00"
	dollars = dollars.replace(".", "")
	return Decimal('%s.%s' % (dollars, cents))

def weekday(value: str):
	value = int(value)
	if not 1 <= value <= 7:
		raise ValueError
	return value

def gender(value: str):
	if value not in ("M", "F"):
		raise ValueError
	return value

def cpf(value: str):
	if not value:
		return None
	cleanValue = ""
	for ch in value:
		if ch in "0123456789":
			cleanValue += ch
	if len(cleanValue) > 11:
		raise ValueError
	cleanValue = "0" * (11 - len(cleanValue)) + cleanValue
	firstVerifier = sum([x * y for x, y in zip(list(range(10, 1, -1)), [int(ch) for ch in cleanValue[:9]])]) % 11
	if firstVerifier < 2:
		firstDigit = "0"
	else:
		firstDigit = str(11 - firstVerifier)
	if cleanValue[9] != firstDigit:
		raise ValueError
	secondVerifier = sum([x * y for x, y in zip(list(range(11, 1, -1)), [int(ch) for ch in cleanValue[:10]])]) % 11
	if secondVerifier < 2:
		secondDigit = "0"
	else:
		secondDigit = str(11 - secondVerifier)
	if cleanValue[10] != secondDigit:
		raise ValueError
	return cleanValue

def bool_(value):
	if value.lower() == 'true':
		return True
	if value.lower() == 'false':
		return False
	return bool(value.strip())

def int_(value):
	if not value:
		return 0
	return int(value.strip())

def unicode_(value):
	return value.strip()

def file_(value):
	if not hasattr(value, 'file'):
		raise ValueError(value)
	return {'name': value.filename.decode('utf-8'),
			'data': value.file.read()}

email = unicode_  # FIXME

def color(value):
	value = unicode_(value).lstrip("#")
	if len(value) == 3:
		value = '%s%s%s%s%s%s' % (value[0], value[0],
								  value[1], value[1],
								  value[2], value[2])
	match = re.match(r"[0-9a-fA-F]{6}", value)
	if not match:
		raise ValueError(value)
	return value.upper()

def cep(value):
	if not value:
		return None
	value = value.decode('utf-8')
	cleanValue = ""
	for ch in value:
		if ch in "0123456789":
			cleanValue += ch
	if len(cleanValue) > 8:
		raise ValueError(value)
	cleanValue = "0" * (8 - len(cleanValue)) + cleanValue
	return cleanValue

def phone(value):
	"""Parses a Brazilian telephone number (the DDD is required).

	>>> phone('(21) 2568-6834')
	u'02125686834'
	>>> phone('(0xx32) 3239-3300')
	u'03232393300'
	>>> phone('11 99763 7727')
	u'011997637727'
	>>> phone('2529 5077')
	Traceback (most recent call last):
		...
	ValueError: 2529 5077
	"""
	if not value:
		return None
	value = value.decode('utf-8')
	cleanValue = ""
	for ch in value:
		if ch in "0123456789":
			cleanValue += ch
	if not cleanValue.startswith("0"):
		cleanValue = "0" + cleanValue
	if len(cleanValue) < 11 or 12 < len(cleanValue):
		raise ValueError(value)
	return cleanValue

