import datetime
import copy
import urllib.request, urllib.parse, urllib.error
import mako.filters
import unicodedata
import warnings
from decimal import Decimal
from collections import OrderedDict

from lawrence.lang import pt_br

def phone(phoneNumber):
    """Format a phone number (only digits) as a Brazilian phone number.

    >>> phone('02125686834')
    u'(21) 2568-6834'
    >>> phone('011997637727')
    u'(11) 99763-7727'
    >>> phone('25295077')
    u'2529-5077'
    """
    if phoneNumber.startswith('00'):
        return "%s" % phoneNumber  # FIXME
    elif phoneNumber.startswith('0'):
        return "(%2s) %s-%4s" % (phoneNumber[1:3], phoneNumber[3:-4], phoneNumber[-4:])
    else:
        return "%s-%4s" % (phoneNumber[:-4], phoneNumber[-4:])

def js(value):
    return str(value).replace('\\', '\\\\').replace('"', '\\"').replace('\'', '\\\'')

def time(value):
    if isinstance(value, datetime.timedelta):
        value = datetime.datetime(1, 1, 1, 0, 0, 0) + value
    if isinstance(value, str):
        value = datetime.time(*[int(part) for part in value.split(':')])
    if value.minute:
        return "%dh%02d" % (value.hour, value.minute)
    else:
        return "%dh" % (value.hour, )

def TimeDelta(value, reference = None):
    """
    >>> from datetime import timedelta, datetime
    >>> now = datetime(2014, 1, 20, 18, 0, 0)
    >>> TimeDelta(timedelta(seconds = 0), now)
    u'agora'
    >>> TimeDelta(timedelta(seconds = -20), now)
    u'instantes atr\\xe1s'
    >>> TimeDelta(timedelta(seconds = 30), now)
    u'daqui a instantes'
    >>> TimeDelta(timedelta(hours = 5), now)
    u'daqui a 5 horas'
    >>> TimeDelta(timedelta(hours = 6), now)
    u'amanh\\xe3'
    >>> TimeDelta(timedelta(hours = 24), now)
    u'amanh\\xe3'
    >>> TimeDelta(timedelta(hours = 29), now)
    u'amanh\\xe3'
    >>> TimeDelta(timedelta(hours = 30), now)
    u'daqui a 2 dias'
    >>> TimeDelta(timedelta(hours = -17), now)
    u'17 horas atr\\xe1s'
    >>> TimeDelta(timedelta(hours = -18), now)
    u'18 horas atr\\xe1s'
    >>> TimeDelta(timedelta(hours = -19), now)
    u'ontem'
    >>> TimeDelta(timedelta(hours = -42), now)
    u'ontem'
    >>> TimeDelta(timedelta(hours = -43), now)
    u'2 dias atr\\xe1s'
    """

    if reference is None:
        reference = datetime.datetime.now()

    # We start counting days: otherwise any interval of [24, 48) hours will be
    # printed as "tomorrow".

    daysDelta = int(((reference + value).date() - reference.date()).total_seconds()) // 86400

    if daysDelta == 0:
        secondsDelta = int(value.total_seconds())

        if secondsDelta == 0:
            return "agora"

        if secondsDelta < 0:
            secondsDeltaAbsolute = -secondsDelta
            template = "%s atrás"
        else:
            secondsDeltaAbsolute = secondsDelta
            template = "daqui a %s"

        if secondsDeltaAbsolute < 60:
            return template % "instantes"
        minutesDeltaAbsolute = secondsDeltaAbsolute // 60

        if minutesDeltaAbsolute < 60:
            return template % unit("minuto")(minutesDeltaAbsolute)
        hoursDeltaAbsolute = minutesDeltaAbsolute // 60

        return template % unit("hora")(hoursDeltaAbsolute)
    else:
        if daysDelta < 0:
            daysDeltaAbsolute = -daysDelta
            template = "%s atrás"
        else:
            daysDeltaAbsolute = daysDelta
            template = "daqui a %s"

        if daysDeltaAbsolute < 30:
            if daysDeltaAbsolute == 1:
                return "ontem" if daysDelta < 0 else "amanhã"
            return template % unit("dia")(daysDeltaAbsolute)
        monthsDeltaAbsolute = daysDeltaAbsolute // 30

        if monthsDeltaAbsolute < 12:
            return template % unit("mês")(monthsDeltaAbsolute)
        yearsDeltaAbsolute = monthsDeltaAbsolute // 12

        return template % unit("ano")(yearsDeltaAbsolute)

def ShortTime(value):
    s = time(value)
    if s.endswith("00"):
        return s[:-2]
    return s

def gender(gender):
    if gender != 'F':
        return (lambda x: x)
    else:
        return pt_br.Feminize

def br(value):
    htmlEscapedValue = str(mako.filters.html_escape(value)).strip()
    return htmlEscapedValue.replace("\n", "<br>")

def QueryString(parameters):
    if not parameters:
        return ''
    return '?' + urllib.parse.urlencode(parameters, doseq = True)

def ShortDay(value):
    if isinstance(value, str):
        value = datetime.date(*[int(part) for part in value.split('-')])
    return str(value.day)

def LongDay(value):
    if isinstance(value, str):
        value = datetime.date(*[int(part) for part in value.split('-')])
    if value.day == 1:
        return "1º"
    else:
        return str(value.day)

_MONTH_SHORT = {
     1: "jan",
     2: "fev",
     3: "mar",
     4: "abr",
     5: "mai",
     6: "jun",
     7: "jul",
     8: "ago",
     9: "set",
    10: "out",
    11: "nov",
    12: "dez",
}
def ShortMonth(value):
    if isinstance(value, str):
        value = datetime.date(*[int(part) for part in value.split('-')])
    return _MONTH_SHORT[value.month]

_MONTH_LONG = {
     1: "janeiro",
     2: "fevereiro",
     3: "março",
     4: "abril",
     5: "maio",
     6: "junho",
     7: "julho",
     8: "agosto",
     9: "setembro",
    10: "outubro",
    11: "novembro",
    12: "dezembro",
}
def LongMonth(value):
    if isinstance(value, str):
        value = datetime.date(*[int(part) for part in value.split('-')])
    return _MONTH_LONG[value.month]

def year(value):
    if isinstance(value, str):
        value = datetime.date(*[int(part) for part in value.split('-')])
    return str(value.year)

def date_s(value):
    return "%s/%s/%s" % (ShortDay(value), ShortMonth(value), year(value))

def ShortDate(value):
    if isinstance(value, str):
        import lawrence.parse
        value = lawrence.parse.date(value)
    return "%02d/%02d/%04d" % (value.day, value.month, value.year)

def LongDateTime(value):
    return "%s de %s, %s" % (LongDay(value), LongMonth(value), time(value))

def decimal(precision = None):
    def DecimalFormatter(value):
        if isinstance(value, Decimal):
            if precision != None:
                value = value.quantize(Decimal('0.1') ** precision)
            value = value.normalize()
            if value.as_tuple().exponent > 0:
                value = value.quantize(Decimal('1'))
        value = str(value).replace("-", "\u2212")
        value = str(value).replace(".", ",")
        return value
    return DecimalFormatter

def dot(value):
    value = str(value)
    output = ""
    for i, char in enumerate(value):
        output += char
        if (len(value) - i - 1) % 3 == 0:
            output += "."
    return output.strip(".")

def money(value):
    if value is None:
        return "\u2014"
    if value < 0:
        return "(%s)" % money(-value)
    cents = int(value * 100)
    integer, cents = cents // 100, cents % 100
    return "R$\u00A0%s,%02d" % (dot(integer), cents)

def percentage(places = None):
    def PercentageFormatter(value):
        if value is None:
            return ""
        return "%s%%" % (decimal(places)(Decimal(value) * Decimal(100)), )
    return PercentageFormatter

def int_():
    def IntegerFormatter(value):
        if value is None:
            return ""
        return str(int(value))
    return IntegerFormatter

def int2_():
    def Integer2Formatter(value):
        if value is None:
            return "0"
        return str(int(value))
    return Integer2Formatter


def cpf(value):
    value = str(value)
    if len(value) != 11:
        return ''
    return "%11s-%2s" % (dot(value[0:9]), value[9:11])

def cep(value):
    value = str(value)
    if len(value) != 8:
        return ''
    return "%5s-%3s" % (dot(value[0:5]), value[5:8])

def unit(singular, plural = None, zero = None):
    def _unitize(value, singular = singular, plural = plural, zero = zero):
        if plural == None:
            plural = pt_br.Pluralize(singular)
        if abs(value) > 1:
            return "%s %s" % (decimal()(value), plural)
        elif value == 0 and zero != None:
            return zero
        else:
            return "%s %s" % (decimal()(value), singular)
    return _unitize

class Grouping (object):
    __slots__ = [
        'Key',
        '_Elements'
    ]
    def __init__(self, key, elements):
        self.Key = key
        self._Elements = tuple(elements)

    def __iter__(self):
        return self._Elements.__iter__()

def GroupBy(source, keySelector):
    groups = OrderedDict()
    for obj in source:
        key = keySelector(obj)
        if key not in groups:
            groups[key] = []
        groups[key].append(obj)
    outputList = []
    for k, v in list(groups.items()):
        outputList.append(Grouping(k, v))
    return outputList

def group_by(things, key):
    def _group_key(thing, key):
        if isinstance(key, str):
            return thing[key]
        else:
            return tuple([thing.get(subKey) for subKey in key])
    groups = {}
    for thing in things:
        attr = _group_key(thing, key)
        groups.setdefault(attr, []).append(thing)
    return sorted(list(groups.items()), key = lambda k_v: k_v[0])

def n_partition(numParts, weights, things):
    assert len(weights) == len(things)
    numThings = len(things)

    # Putting all items in a single column is a solution.
    class Solution (object):
        pass
    best = Solution()
    best.cost = sum(weights)
    best.distribution = [list(range(numThings))] + [[] for _ in range(1, numParts)]

    # Sort things by decreasing weight.
    thingOrder = list(range(numThings))
    thingOrder.sort(key = lambda j: weights[j])
    thingOrder = thingOrder[::-1]
    weights = [weights[j] for j in thingOrder]
    things = [things[j] for j in thingOrder]

    def PartitionDfs(distribution, cost, p, lastColumn):
        if p >= numThings:
            best.cost = max(cost)
            best.distribution = copy.deepcopy(distribution[:])
            return
        if p > 0 and weights[p-1] == weights[p]:
            validColumns = list(range(lastColumn, numParts))
        else:
            validColumns = list(range(numParts))
        for i in validColumns:
            if cost[i] + weights[p] >= best.cost:
                continue
            cost[i] += weights[p]
            distribution[i].append(p)
            PartitionDfs(distribution, cost, p+1, i)
            distribution[i].pop()
            cost[i] -= weights[p]

    PartitionDfs([[] for _ in range(numParts)], [0] * numParts, 0, 0)

    # Return columns in lexicographical order, but push empty lists to the
    # end.
    sol = sorted([I for I in best.distribution if I])
    while len(sol) < numParts:
        sol.append([])

    return [[things[_] for _ in I] for I in sol]

def columns(numColumns, things):
    if numColumns == 0:
        return []
    if len(things) >= numColumns:
        columnSize = len(things) // numColumns
        C = []
        k = 0
        for i in range(numColumns):
            l = k + columnSize
            if i < len(things) % numColumns:
                l += 1
            C.append(things[k:l])
            k = l
        return C
    else:
        return columns(len(things), things) + [()] * (numColumns - len(things))

def Slug(text):
    if isinstance(text, bytes):
        text = text.decode('utf-8')
    slug = ''
    for char in text:
        char = unicodedata.normalize('NFD', char)[0]
        if char in ('abcdefghijklmnopqrstuvwxyz' 'ABCDEFGHIJKLMNOPQRSTUVWXYZ' '0123456789'):
            slug += str(char)
        elif char == "ª":
            slug += 'a'
        elif char == "º":
            slug += 'o'
        elif not slug.endswith('__'):
            slug += '_'
    return slug

def selected(value):
    if value:
        return " selected=\"selected\""
    else:
        return ""

def checked(value):
    if value:
        return " checked=\"checked\""
    else:
        return ""

def disabled(value):
    if value:
        return " disabled=\"disabled\""
    else:
        return ""

def EncodeId(decoded):
    decoded = (decoded * 26604337) % 9**8
    digit = 0
    encoded = 0
    for _ in range(8):
        encoded *= 10
        digit = (digit + 1 + decoded % 9) % 10
        encoded += digit
        decoded //= 9
    return encoded

def DecodeId(encoded):
    decoded = 0
    for _ in range(8):
        decoded *= 9
        digit = encoded % 10
        encoded //= 10
        decoded += (digit - encoded % 10 - 1) % 10
    decoded = (decoded * 24059149) % 9**8
    return decoded

class CachedProperty (object):
    def __init__(self, method):
        self.method = method
        self.name = method.__name__

    def __get__(self, inst, cls):
        if not hasattr(inst, '_%s_value' % self.name):
            setattr(inst, '_%s_value' % self.name, self.method(inst))
        return getattr(inst, '_%s_value' % self.name)

def Deprecated(func):
    """This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used."""
    def DeprecationWrapper(*args, **kwargs):
        warnings.warn("Call to deprecated function %s." % func.__name__, category = DeprecationWarning)
        return func(*args, **kwargs)
    DeprecationWrapper.__name__ = func.__name__
    DeprecationWrapper.__dict__.update(func.__dict__)
    return DeprecationWrapper

def AgeInYearsAndMonths(birthDate, referenceDate = None):
    if referenceDate is None:
        from datetime import date
        referenceDate = date.today()
    from .timeutil import GetMonthsElapsed
    monthsElapsed = GetMonthsElapsed(birthDate, referenceDate)
    yearsElapsed, monthsElapsed = monthsElapsed // 12, monthsElapsed % 12
    if yearsElapsed == 0:
        if monthsElapsed == 0:
            return "recém-nascido"
        else:
            return unit("mês")(monthsElapsed)
    elif yearsElapsed < 6:
        if monthsElapsed == 0:
            return unit("ano")(yearsElapsed)
        else:
            return unit("ano")(yearsElapsed) + " e " + unit("mês")(monthsElapsed)
    else:
        return unit("ano")(yearsElapsed)

